@@ -1,6 +1,5 @@
-
 /* ====================================================================
- * Copyright (c) 1995 The Apache Group.  All rights reserved.
+ * Copyright (c) 1995-1997 The Apache Group.  All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
@@ -101,15 +100,18 @@ union block_hdr *block_freelist = NULL;
 
 
 
-/* Get a completely new block from the system pool */
+/* Get a completely new block from the system pool. Note that we rely on
+malloc() to provide aligned memory. */
 
 union block_hdr *malloc_block (int size)
 {
   union block_hdr *blok =
     (union block_hdr *)malloc(size + sizeof(union block_hdr));
 
-  if (blok == NULL) return NULL;
-  
+  if (blok == NULL) {
+      fprintf (stderr, "Ouch!  malloc failed in malloc_block()\n");
+      exit (1);
+  }
   blok->h.next = NULL;
   blok->h.first_avail = (char *)(blok + 1);
   blok->h.endp = size + blok->h.first_avail;
@@ -183,11 +185,8 @@ union block_hdr *new_block (int min_size)
    * on the free list...
    */
 
-  min_size += BLOCK_MINFREE;
-
   while (blok != NULL) {
-    if (min_size <= blok->h.endp - blok->h.first_avail) {
-      
+    if (min_size + BLOCK_MINFREE <= blok->h.endp - blok->h.first_avail) {
       *lastptr = blok->h.next;
       blok->h.next = NULL;
       return blok;
@@ -200,7 +199,8 @@ union block_hdr *new_block (int min_size)
 
   /* Nope. */
 
-  return malloc_block (min_size);
+  min_size += BLOCK_MINFREE;
+  return malloc_block((min_size > BLOCK_MINALLOC) ? min_size : BLOCK_MINALLOC);
 }
 
 
@@ -337,7 +337,7 @@ void *palloc (struct pool *a, int reqsize)
   
   int nclicks = 1 + ((reqsize - 1) / CLICK_SZ);
   int size = nclicks * CLICK_SZ;
-  
+
   /* First, see if we have space in the block most recently
    * allocated to this pool
    */
@@ -346,7 +346,9 @@ void *palloc (struct pool *a, int reqsize)
   char *first_avail = blok->h.first_avail;
   char *new_first_avail;
 
-  if (size <= 0) size = 1;
+  if(reqsize <= 0)
+      return NULL;
+  
   new_first_avail = first_avail + size;
   
   if (new_first_avail <= blok->h.endp) {
@@ -458,27 +460,28 @@ array_header *make_array (pool *p, int nelts, int elt_size)
 void *push_array (array_header *arr)
 {
   if (arr->nelts == arr->nalloc) {
-    char *new_data = pcalloc (arr->pool, arr->nalloc * arr->elt_size * 2);
+    int new_size = (arr->nalloc <= 0) ? 1 : arr->nalloc * 2;
+    char *new_data;
+    
+    new_data = pcalloc (arr->pool, arr->elt_size * new_size);
 
     memcpy (new_data, arr->elts, arr->nalloc * arr->elt_size);
     arr->elts = new_data;
-    arr->nalloc *= 2;
+    arr->nalloc = new_size;
   }
 
   ++arr->nelts;
   return arr->elts + (arr->elt_size * (arr->nelts - 1));
 }
 
-void array_cat (array_header *dst, array_header *src)
+void array_cat (array_header *dst, const array_header *src)
 {
   int elt_size = dst->elt_size;
   
   if (dst->nelts + src->nelts > dst->nalloc) {
-    int new_size = dst->nalloc * 2;
+    int new_size = (dst->nalloc <= 0) ? 1 : dst->nalloc * 2;
     char *new_data;
 
-    if (new_size == 0) ++new_size;
-    
     while (dst->nelts + src->nelts > new_size)
       new_size *= 2;
 
@@ -493,7 +496,7 @@ void array_cat (array_header *dst, array_header *src)
   dst->nelts += src->nelts;
 }
 
-array_header *copy_array (pool *p, array_header *arr)
+array_header *copy_array (pool *p, const array_header *arr)
 {
   array_header *res = make_array (p, arr->nalloc, arr->elt_size);
 
@@ -509,7 +512,7 @@ array_header *copy_array (pool *p, array_header *arr)
  * overhead of the full copy only where it is really needed.
  */
 
-array_header *copy_array_hdr (pool *p, array_header *arr)
+array_header *copy_array_hdr (pool *p, const array_header *arr)
 {
   array_header *res = (array_header *)palloc(p, sizeof(array_header));
 
@@ -526,7 +529,8 @@ array_header *copy_array_hdr (pool *p, array_header *arr)
 /* The above is used here to avoid consing multiple new array bodies... */
 
 array_header *append_arrays (pool *p,
-			     array_header *first, array_header *second)
+			     const array_header *first,
+			     const array_header *second)
 {
   array_header *res = copy_array_hdr (p, first);
 
@@ -544,13 +548,18 @@ table *make_table (pool *p, int nelts) {
     return make_array (p, nelts, sizeof (table_entry));
 }
 
-table *copy_table (pool *p, table *t) {
+table *copy_table (pool *p, const table *t) {
     return copy_array (p, t);
 }
 
+void clear_table (table *t)
+{
+    t->nelts = 0;
+}
+
 array_header *table_elts (table *t) { return t; }
 
-char *table_get (table *t, char *key)
+char *table_get (const table *t, const char *key)
 {
     table_entry *elts = (table_entry *)t->elts;
     int i;
@@ -566,49 +575,54 @@ char *table_get (table *t, char *key)
 
 void table_set (table *t, const char *key, const char *val)
 {
+    register int i, j, k;
     table_entry *elts = (table_entry *)t->elts;
-    int i;
+    int done = 0;
 
     for (i = 0; i < t->nelts; ++i)
-        if (!strcasecmp (elts[i].key, key)) {
-	    elts[i].val = pstrdup (t->pool, val);
-	    return;
+	if (!strcasecmp (elts[i].key, key)) {
+	    if (!done) {
+	        elts[i].val = pstrdup(t->pool, val);
+	        done = 1;
+	    }
+	    else {     /* delete an extraneous element */
+                for (j = i, k = i + 1; k < t->nelts; ++j, ++k) {
+                    elts[j].key = elts[k].key;
+                    elts[j].val = elts[k].val;
+                }
+                --t->nelts;
+	    }
 	}
 
-    elts = (table_entry *)push_array(t);
-    elts->key = pstrdup (t->pool, key);
-    elts->val = pstrdup (t->pool, val);
+    if (!done) {
+        elts = (table_entry *)push_array(t);
+        elts->key = pstrdup (t->pool, key);
+        elts->val = pstrdup (t->pool, val);
+    }
 }
 
-void table_unset( table *t, char *key ) 
+void table_unset( table *t, const char *key ) 
 {
+    register int i, j, k;   
     table_entry *elts = (table_entry *)t->elts;
-    int i;   
-    int j;   
  
     for (i = 0; i < t->nelts; ++i)
         if (!strcasecmp (elts[i].key, key)) {
  
-            /* found the element to skip over
+            /* found an element to skip over
              * there are any number of ways to remove an element from
              * a contiguous block of memory.  I've chosen one that
              * doesn't do a memcpy/bcopy/array_delete, *shrug*...
              */
-            j = i;
-            ++i;
-            for ( ; i < t->nelts; ) {
-                elts[j].key = elts[i].key;
-                elts[j].val = elts[i].val;
-                ++i;
-                ++j;
-            };
+            for (j = i, k = i + 1; k < t->nelts; ++j, ++k) {
+                elts[j].key = elts[k].key;
+                elts[j].val = elts[k].val;
+            }
             --t->nelts;
-
-            return;
         }
 }     
 
-void table_merge (table *t, char *key, char *val)
+void table_merge (table *t, const char *key, const char *val)
 {
     table_entry *elts = (table_entry *)t->elts;
     int i;
@@ -624,7 +638,7 @@ void table_merge (table *t, char *key, char *val)
     elts->val = pstrdup (t->pool, val);
 }
 
-void table_add (table *t, char *key, char *val)
+void table_add (table *t, const char *key, const char *val)
 {
     table_entry *elts = (table_entry *)t->elts;
 
@@ -633,11 +647,56 @@ void table_add (table *t, char *key, char *val)
     elts->val = pstrdup (t->pool, val);
 }
 
-table* overlay_tables (pool *p, table *overlay, table *base)
+table* overlay_tables (pool *p, const table *overlay, const table *base)
 {
     return append_arrays (p, overlay, base);
 }
 
+/* And now for something completely abstract ...
+ *
+ * For each key value given as a vararg:
+ *   run the function pointed to as
+ *     int comp(void *r, char *key, char *value);
+ *   on each valid key-value pair in the table t that matches the vararg key,
+ *   or once for every valid key-value pair if the vararg list is empty,
+ *   until the function returns false (0) or we finish the table.
+ *
+ * Note that we restart the traversal for each vararg, which means that
+ * duplicate varargs will result in multiple executions of the function
+ * for each matching key.  Note also that if the vararg list is empty,
+ * only one traversal will be made and will cut short if comp returns 0.
+ *
+ * Note that the table_get and table_merge functions assume that each key in
+ * the table is unique (i.e., no multiple entries with the same key).  This
+ * function does not make that assumption, since it (unfortunately) isn't
+ * true for some of Apache's tables.
+ *
+ * Note that rec is simply passed-on to the comp function, so that the
+ * caller can pass additional info for the task.
+ */
+void table_do (int (*comp)(void *, const char *, const char *), void *rec,
+               const table *t, ...)
+{
+    va_list vp;
+    char *argp;
+    table_entry *elts = (table_entry *)t->elts;
+    int rv, i;
+  
+    va_start(vp, t);
+
+    argp = va_arg(vp, char *);
+
+    do {
+        for (rv = 1, i = 0; rv && (i < t->nelts); ++i) {
+            if (elts[i].key && (!argp || !strcasecmp(elts[i].key, argp))) {
+                rv = (*comp)(rec, elts[i].key, elts[i].val);
+            }
+        }
+    } while (argp && ((argp = va_arg(vp, char *)) != NULL));
+
+    va_end(vp);
+}
+
 /*****************************************************************
  *
  * Managing generic cleanups.  
@@ -729,25 +788,36 @@ void note_cleanups_for_fd (pool *p, int fd) {
   register_cleanup (p, (void *)fd, fd_cleanup, fd_cleanup);
 }
 
-int popenf(struct pool *a, char *name, int flg, int mode)
+void kill_cleanups_for_fd(pool *p,int fd)
+    {
+    kill_cleanup(p,(void *)fd,fd_cleanup);
+    }
+
+int popenf(pool *a, const char *name, int flg, int mode)
 {
   int fd;
+  int save_errno;
 
   block_alarms();
   fd = open(name, flg, mode);
+  save_errno = errno;
   if (fd >= 0) note_cleanups_for_fd (a, fd);
   unblock_alarms();
+  errno = save_errno;
   return fd;
 }
 
-int pclosef(struct pool *a, int fd)
+int pclosef(pool *a, int fd)
 {
   int res;
+  int save_errno;
   
   block_alarms();
   res = close(fd);
+  save_errno = errno;
   kill_cleanup(a, (void *)fd, fd_cleanup);
   unblock_alarms();
+  errno = save_errno;
   return res;
 }
 
@@ -759,22 +829,35 @@ int pclosef(struct pool *a, int fd)
 static void file_cleanup (void *fpv) { fclose ((FILE *)fpv); }
 static void file_child_cleanup (void *fpv) { close (fileno ((FILE *)fpv)); }
 
-void note_cleanups_for_file (struct pool *p, FILE *fp) {
+void note_cleanups_for_file (pool *p, FILE *fp) {
   register_cleanup (p, (void *)fp, file_cleanup, file_child_cleanup);
 }
 
-FILE *pfopen(struct pool *a, char *name, char *mode)
+FILE *pfopen(pool *a, const char *name, const char *mode)
 {
-  FILE *fd;
+  FILE *fd = NULL;
+  int baseFlag, desc;
 
   block_alarms();
-  fd = fopen(name, mode);
+
+  if (*mode == 'a') {
+    /* Work around faulty implementations of fopen */
+    baseFlag = (*(mode+1) == '+') ? O_RDWR : O_WRONLY;
+    desc = open(name, baseFlag | O_APPEND | O_CREAT,
+		S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH | S_IWOTH);
+    if (desc >= 0) {
+      fd = fdopen(desc, mode);
+    }
+  } else {
+    fd = fopen(name, mode);
+  }
+
   if (fd != NULL) note_cleanups_for_file (a, fd);
   unblock_alarms();
   return fd;
 }
 
-FILE *pfdopen(struct pool *a,int fd,char *mode)
+FILE *pfdopen(pool *a,int fd, const char *mode)
 {
   FILE *f;
 
@@ -787,7 +870,7 @@ FILE *pfdopen(struct pool *a,int fd,char *mode)
 }
 
 
-int pfclose(struct pool *a, FILE *fd)
+int pfclose(pool *a, FILE *fd)
 {
   int res;
   
@@ -798,6 +881,36 @@ int pfclose(struct pool *a, FILE *fd)
   return res;
 }
 
+/*
+ * Here's a pool-based interface to POSIX regex's regcomp().
+ * Note that we return regex_t instead of being passed one.
+ * The reason is that if you use an already-used regex_t structure,
+ * the memory that you've already allocated gets forgotten, and
+ * regfree() doesn't clear it. So we don't allow it.
+ */
+
+static void regex_cleanup (void *preg) { regfree ((regex_t *)preg); }
+
+regex_t *pregcomp(pool *p, const char *pattern, int cflags) {
+    regex_t *preg = palloc(p, sizeof(regex_t));
+
+    if (regcomp(preg, pattern, cflags))
+	return NULL;
+
+    register_cleanup (p, (void *)preg, regex_cleanup, regex_cleanup);
+
+    return preg;
+}
+
+
+void pregfree(pool *p, regex_t *reg)
+{
+    block_alarms();
+    regfree (reg);
+    kill_cleanup (p, (void *)reg, regex_cleanup);
+    unblock_alarms();
+}
+
 /*****************************************************************
  *
  * More grotty system stuff... subprocesses.  Frump.  These don't use
@@ -826,31 +939,38 @@ void note_subprocess (pool *a, int pid, enum kill_conditions how)
   a->subprocesses = new;
 }
 
-int spawn_child (pool *p, void (*func)(void *), void *data,
-		 enum kill_conditions kill_how,
-		 FILE **pipe_in, FILE **pipe_out)
+int spawn_child_err (pool *p, void (*func)(void *), void *data,
+		     enum kill_conditions kill_how,
+		     FILE **pipe_in, FILE **pipe_out, FILE **pipe_err)
 {
   int pid;
   int in_fds[2];
   int out_fds[2];
+  int err_fds[2];
+  int save_errno;
 
   block_alarms();
   
   if (pipe_in && pipe (in_fds) < 0)
   {
+      save_errno = errno;
       unblock_alarms();
+      errno = save_errno;
       return 0;
   }
   
   if (pipe_out && pipe (out_fds) < 0) {
+    save_errno = errno;
     if (pipe_in) {
       close (in_fds[0]); close (in_fds[1]);
     }
     unblock_alarms();
+    errno = save_errno;
     return 0;
   }
 
-  if ((pid = fork()) < 0) {
+  if (pipe_err && pipe (err_fds) < 0) {
+    save_errno = errno;
     if (pipe_in) {
       close (in_fds[0]); close (in_fds[1]);
     }
@@ -858,6 +978,23 @@ int spawn_child (pool *p, void (*func)(void *), void *data,
       close (out_fds[0]); close (out_fds[1]);
     }
     unblock_alarms();
+    errno = save_errno;
+    return 0;
+  }
+
+  if ((pid = fork()) < 0) {
+    save_errno = errno;
+    if (pipe_in) {
+      close (in_fds[0]); close (in_fds[1]);
+    }
+    if (pipe_out) {
+      close (out_fds[0]); close (out_fds[1]);
+    }
+    if (pipe_err) {
+      close (err_fds[0]); close (err_fds[1]);
+    }
+    unblock_alarms();
+    errno = save_errno;
     return 0;
   }
 
@@ -876,6 +1013,12 @@ int spawn_child (pool *p, void (*func)(void *), void *data,
       close (in_fds[0]);
     }
 
+    if (pipe_err) {
+      close (err_fds[0]);
+      dup2 (err_fds[1], STDERR_FILENO);
+      close (err_fds[1]);
+    }
+
     /* HP-UX SIGCHLD fix goes here, if someone will remind me what it is... */
     signal (SIGCHLD, SIG_DFL);	/* Was that it? */
     
@@ -911,6 +1054,18 @@ int spawn_child (pool *p, void (*func)(void *), void *data,
     if (*pipe_in) note_cleanups_for_file (p, *pipe_in);
   }
 
+  if (pipe_err) {
+    close (err_fds[1]);
+#ifdef __EMX__
+    /* Need binary mode set for OS/2. */
+    *pipe_err = fdopen (err_fds[0], "rb");
+#else
+    *pipe_err = fdopen (err_fds[0], "r");
+#endif
+  
+    if (*pipe_err) note_cleanups_for_file (p, *pipe_err);
+  }
+
   unblock_alarms();
   return pid;
 }
@@ -973,98 +1128,3 @@ static void free_proc_chain (struct process_chain *procs)
   }
 }
 
-#ifdef __EMX__
-int spawn_child_os2 (pool *p, void (*func)(void *), void *data,
-         enum kill_conditions kill_how,
-         FILE **pipe_in, FILE **pipe_out, char *buffer, int lenp)
-{
-  int pid;
-  int in_fds[2];
-  int out_fds[2];
-
-  block_alarms();
-  
-  if (pipe_in && pipe (in_fds) < 0)
-  {
-      unblock_alarms();
-      return 0;
-  }
-  
-  if (pipe_out && pipe (out_fds) < 0) {
-    if (pipe_in) {
-      close (in_fds[0]); close (in_fds[1]);
-    }
-    unblock_alarms();
-    return 0;
-  }
-
-  if ((pid = fork()) < 0) {
-    if (pipe_in) {
-      close (in_fds[0]); close (in_fds[1]);
-    }
-    if (pipe_out) {
-      close (out_fds[0]); close (out_fds[1]);
-    }
-    unblock_alarms();
-    return 0;
-  }
-
-  if (!pid) {
-    int stdinpipe[2];
-    /* Due to a limitation of EMX, inheriting socket handles is not
-    allowed so we need to read the input and place it in a pipe and 
-    then pass that handle instead of the socket. */
-    
-    if (lenp > 0) {
-        pipe(stdinpipe);
-        write(stdinpipe[1], buffer, lenp);
-        close(stdinpipe[1]);
-        in_fds[0] = dup(stdinpipe[0]);
-        close(stdinpipe[0]);
-    }
-                                
-    /* Child process */
-    
-    if (pipe_out) {
-      close (out_fds[0]);
-      dup2 (out_fds[1], STDOUT_FILENO);
-      close (out_fds[1]);
-    }
-
-    if (pipe_in) {
-      close (in_fds[1]);
-      dup2 (in_fds[0], STDIN_FILENO);
-      close (in_fds[0]);
-    }
-
-    /* HP-UX SIGCHLD fix goes here, if someone will remind me what it is... */
-    signal (SIGCHLD, SIG_DFL);    /* Was that it? */
-    
-    func (data);
-    exit (0);            /* Should never get here... */
-  }
-
-  /* Parent process */
-
-  note_subprocess (p, pid, kill_how);
-  
-  if (pipe_out) {
-    close (out_fds[1]);
-    /* Need binary mode set for OS/2. */
-    *pipe_out = fdopen (out_fds[0], "rb");
-    
-    if (*pipe_out) note_cleanups_for_file (p, *pipe_out);
-  }
-
-  if (pipe_in) {
-    close (in_fds[0]);
-    /* Need binary mode set for OS/2. */
-    *pipe_in = fdopen (in_fds[1], "wb");
-    
-    if (*pipe_in) note_cleanups_for_file (p, *pipe_in);
-  }
-
-  unblock_alarms();
-  return pid;
-}
-#endif
