@@ -1,5 +1,5 @@
 /* ====================================================================
- * Copyright (c) 1996 The Apache Group.  All rights reserved.
+ * Copyright (c) 1996,1997 The Apache Group.  All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
@@ -50,14 +50,25 @@
  *
  */
 
+#include "conf.h"
+#include "alloc.h"
+#include "buff.h"
+
 #include <errno.h>
 #include <stdio.h>
 #include <stdarg.h>
 #include <string.h>
+#ifndef NO_UNISTD_H
 #include <unistd.h>
+#endif
+#ifndef NO_WRITEV
+#include <sys/types.h>
+#include <sys/uio.h>
+#endif
 
-#include "alloc.h"
-#include "buff.h"
+#ifdef HAVE_BSTRING_H
+#include <bstring.h>		/* for IRIX, FD_SET calls bzero() */
+#endif
 
 #define DEFAULT_BUFSIZE (4096)
 
@@ -88,11 +99,15 @@
 static void
 doerror(BUFF *fb, int err)
 {
+    int errsave = errno;  /* Save errno to prevent overwriting it below */
+
     if (err == B_RD)
 	fb->flags |= B_RDERR;
     else
 	fb->flags |= B_WRERR;
     if (fb->error != NULL) (*fb->error)(fb, err, fb->error_data);
+
+    errno = errsave;
 }
 
 /* Buffering routines */
@@ -112,13 +127,16 @@ bcreate(pool *p, int flags)
     if (flags & B_RD) fb->inbase = palloc(p, fb->bufsiz);
     else fb->inbase = NULL;
 
-    if (flags & B_WR) fb->outbase = palloc(p, fb->bufsiz);
-    else fb->inbase = NULL;
+    /* overallocate so that we can put a chunk trailer of CRLF into this
+     * buffer */
+    if (flags & B_WR) fb->outbase = palloc(p, fb->bufsiz + 2);
+    else fb->outbase = NULL;
 
     fb->inptr = fb->inbase;
 
     fb->incnt = 0;
     fb->outcnt = 0;
+    fb->outchunk = -1;
     fb->error = NULL;
     fb->bytes_sent = 0L;
 
@@ -136,9 +154,6 @@ bpushfd(BUFF *fb, int fd_in, int fd_out)
 {
     fb->fd = fd_out;
     fb->fd_in = fd_in;
-    note_cleanups_for_fd(fb->pool,fb->fd);
-    if(fb->fd != fb->fd_in)
-	note_cleanups_for_fd(fb->pool,fb->fd_in);
 }
 
 int
@@ -171,6 +186,150 @@ bgetopt(BUFF *fb, int optname, void *optval)
     }
 }
 
+/*
+ * start chunked encoding
+ */
+static void
+start_chunk( BUFF *fb )
+{
+    char chunksize[16];	/* Big enough for practically anything */
+    int chunk_header_size;
+
+    if (fb->outchunk != -1) {
+	/* already chunking */
+	return;
+    }
+    if (!(fb->flags & B_WR) || (fb->flags & (B_WRERR|B_EOUT))) {
+	/* unbuffered writes */
+	return;
+    }
+
+    /* we know that the chunk header is going to take at least 3 bytes... */
+    chunk_header_size = ap_snprintf( chunksize, sizeof(chunksize),
+	"%x\015\012", fb->bufsiz - fb->outcnt - 3 );
+    /* we need at least the header_len + at least 1 data byte
+     * remember that we've overallocated fb->outbase so that we can always
+     * fit the two byte CRLF trailer
+     */
+    if( fb->bufsiz - fb->outcnt < chunk_header_size + 1 ) {
+	bflush(fb);
+    }
+    /* assume there's enough space now */
+    memcpy( &fb->outbase[fb->outcnt], chunksize, chunk_header_size );
+    fb->outchunk = fb->outcnt;
+    fb->outcnt += chunk_header_size;
+    fb->outchunk_header_size = chunk_header_size;
+}
+
+
+/*
+ * end a chunk -- tweak the chunk_header from start_chunk, and add a trailer
+ */
+static void
+end_chunk( BUFF *fb )
+{
+    int i;
+
+    if( fb->outchunk == -1 ) {
+	/* not chunking */
+	return;
+    }
+
+    if( fb->outchunk + fb->outchunk_header_size == fb->outcnt ) {
+	/* nothing was written into this chunk, and we can't write a 0 size
+	 * chunk because that signifies EOF, so just erase it
+	 */
+	fb->outcnt = fb->outchunk;
+	fb->outchunk = -1;
+	return;
+    }
+
+    /* we know this will fit because of how we wrote it in start_chunk() */
+    i = ap_snprintf( (char *)&fb->outbase[fb->outchunk],
+	fb->outchunk_header_size,
+	"%x", fb->outcnt - fb->outchunk - fb->outchunk_header_size );
+
+    /* we may have to tack some trailing spaces onto the number we just wrote
+     * in case it was smaller than our estimated size.  We've also written
+     * a \0 into the buffer with ap_snprintf so we might have to put a
+     * \r back in.
+     */
+    i += fb->outchunk;
+    while( fb->outbase[i] != '\015' && fb->outbase[i] != '\012' ) {
+	fb->outbase[i++] = ' ';
+    }
+    if( fb->outbase[i] == '\012' ) {
+	/* we overwrote the \r, so put it back */
+	fb->outbase[i-1] = '\015';
+    }
+
+    /* tack on the trailing CRLF, we've reserved room for this */
+    fb->outbase[fb->outcnt++] = '\015';
+    fb->outbase[fb->outcnt++] = '\012';
+
+    fb->outchunk = -1;
+}
+
+
+/*
+ * Set a flag on (1) or off (0).
+ */
+int bsetflag(BUFF *fb, int flag, int value)
+{
+    if (value) {
+	fb->flags |= flag;
+	if( flag & B_CHUNK ) {
+	    start_chunk(fb);
+	}
+    } else {
+	fb->flags &= ~flag;
+	if( flag & B_CHUNK ) {
+	    end_chunk(fb);
+	}
+    }
+    return value;
+}
+
+
+/*
+ * This is called instead of read() everywhere in here.  It implements
+ * the B_SAFEREAD functionality -- which is to force a flush() if a read()
+ * would block.  It also deals with the EINTR errno result from read().
+ * return code is like read() except EINTR is eliminated.
+ */
+static int
+saferead( BUFF *fb, void *buf, int nbyte )
+{
+    int rv;
+
+    if( fb->flags & B_SAFEREAD ) {
+	fd_set fds;
+	struct timeval tv;
+
+	/* test for a block */
+	do {
+	    FD_ZERO( &fds );
+	    FD_SET( fb->fd_in, &fds );
+	    tv.tv_sec = 0;
+	    tv.tv_usec = 0;
+#ifdef SELECT_NEEDS_CAST
+	    rv = select( fb->fd_in + 1, (int *)&fds, NULL, NULL, &tv );
+#else
+	    rv = select( fb->fd_in + 1, &fds, NULL, NULL, &tv );
+#endif
+	} while( rv < 0 && errno == EINTR );
+	/* treat any error as if it would block as well */
+	if( rv != 1 ) {
+	    bflush(fb);
+	}
+    }
+    do {
+	rv = read( fb->fd_in, buf, nbyte );
+    } while (rv == -1 && errno == EINTR && !(fb->flags & B_EOUT));
+    return( rv );
+}
+
+
 /*
  * Read up to nbyte bytes into buf.
  * If fewer than byte bytes are currently available, then return those.
@@ -187,8 +346,7 @@ bread(BUFF *fb, void *buf, int nbyte)
     if (!(fb->flags & B_RD))
     {
 /* Unbuffered reading */
-	do i = read(fb->fd_in, buf, nbyte);
-	while (i == -1 && errno == EINTR);
+	i = saferead( fb, buf, nbyte );
 	if (i == -1 && errno != EAGAIN) doerror(fb, B_RD);
 	return i;
     }
@@ -216,8 +374,7 @@ bread(BUFF *fb, void *buf, int nbyte)
     if (nbyte >= fb->bufsiz)
     {
 /* read directly into buffer */
-	do i = read(fb->fd_in, buf, nbyte);
-	while (i == -1 && errno == EINTR);
+	i = saferead( fb, buf, nbyte );
 	if (i == -1)
 	{
 	    if (nrd == 0)
@@ -231,8 +388,7 @@ bread(BUFF *fb, void *buf, int nbyte)
     {
 /* read into hold buffer, then memcpy */
 	fb->inptr = fb->inbase;
-	do i = read(fb->fd_in, fb->inptr, fb->bufsiz);
-	while (i == -1 && errno == EINTR);
+	i = saferead( fb, fb->inptr, fb->bufsiz );
 	if (i == -1)
 	{
 	    if (nrd == 0)
@@ -293,8 +449,7 @@ bgets(char *buff, int n, BUFF *fb)
 	    fb->inptr = fb->inbase;
 	    fb->incnt = 0;
 	    if (fb->flags & B_EOF) break;
-	    do i = read(fb->fd_in, fb->inptr, fb->bufsiz);
-	    while (i == -1 && errno == EINTR);
+	    i = saferead( fb, fb->inptr, fb->bufsiz );
 	    if (i == -1)
 	    {
 		buff[ct] = '\0';
@@ -340,6 +495,48 @@ bgets(char *buff, int n, BUFF *fb)
     return ct;
 }
 
+/*
+ * Looks at the stream fb and places the first character into buff
+ * without removing it from the stream buffer.
+ *
+ * Returns 1 on success, zero on end of transmission, or -1 on an error.
+ *
+ */
+int blookc(char *buff, BUFF *fb)
+{
+    int i;
+
+    *buff = '\0';
+    
+    if (!(fb->flags & B_RD)) {   /* Can't do blookc on an unbuffered stream */
+        errno = EINVAL;
+        return -1;
+    }
+    if (fb->flags & B_RDERR) return -1;
+
+    if (fb->incnt == 0) {        /* no characters left in stream buffer */
+        fb->inptr = fb->inbase;
+        if (fb->flags & B_EOF)
+            return 0;
+
+	i = saferead( fb, fb->inptr, fb->bufsiz );
+
+        if (i == -1) {
+            if (errno != EAGAIN)
+                doerror(fb, B_RD);
+            return -1;
+        }
+        if (i == 0) {
+            fb->flags |= B_EOF;
+            return 0; /* EOF */
+        }
+        else fb->incnt = i;
+    }
+
+    *buff = fb->inptr[0];
+    return 1;
+}
+
 /*
  * Skip data until a linefeed character is read
  * Returns 1 on success, 0 if no LF found, or -1 on error
@@ -360,7 +557,7 @@ bskiplf(BUFF *fb)
 
     for (;;)
     {
-	x = memchr(fb->inptr, '\012', fb->incnt);
+	x = (unsigned char *)memchr(fb->inptr, '\012', fb->incnt);
 	if (x != NULL)
 	{
 	    x++;
@@ -372,8 +569,7 @@ bskiplf(BUFF *fb)
 	fb->inptr = fb->inbase;
 	fb->incnt = 0;
 	if (fb->flags & B_EOF) return 0;
-	do i = read(fb->fd_in, fb->inptr, fb->bufsiz);
-	while (i == -1 && errno == EINTR);
+	i = saferead( fb, fb->inptr, fb->bufsiz );
 	if (i == 0) fb->flags |= B_EOF;
 	if (i == -1 && errno != EAGAIN) doerror(fb, B_RD);
 	if (i == 0 || i == -1) return i;
@@ -408,10 +604,121 @@ bfilbuf(BUFF *fb)
     else return buf[0];
 }
 
+
+/*
+ * When doing chunked encodings we really have to write everything in the
+ * chunk before proceeding onto anything else.  This routine either writes
+ * nbytes and returns 0 or returns -1 indicating a failure.
+ *
+ * This is *seriously broken* if used on a non-blocking fd.  It will poll.
+ */
+static int
+write_it_all(BUFF *fb, const void *buf, int nbyte)
+{
+    int i;
+
+    if (fb->flags & (B_WRERR|B_EOUT))
+	return -1;
+
+    while (nbyte > 0) {
+	i = write(fb->fd, buf, nbyte);
+	if (i < 0) {
+	    if (errno != EAGAIN && errno != EINTR) {
+		return -1;
+	    }
+	}
+	else {
+	    nbyte -= i;
+	    buf = i + (const char *)buf;
+	}
+	if (fb->flags & B_EOUT)
+	    return -1;
+    }
+    return 0;
+}
+
+
+/*
+ * A hook to write() that deals with chunking. This is really a protocol-
+ * level issue, but we deal with it here because it's simpler; this is
+ * an interim solution pending a complete rewrite of all this stuff in
+ * 2.0, using something like sfio stacked disciplines or BSD's funopen().
+ */
+static int
+bcwrite(BUFF *fb, const void *buf, int nbyte)
+{
+    char chunksize[16];	/* Big enough for practically anything */
+#ifndef NO_WRITEV
+    struct iovec vec[3];
+    int i, rv;
+#endif
+
+    if (fb->flags & (B_WRERR|B_EOUT))
+	return -1;
+
+    if (!(fb->flags & B_CHUNK))
+	return write(fb->fd, buf, nbyte);
+
+#ifdef NO_WRITEV
+    /* without writev() this has poor performance, too bad */
+
+    ap_snprintf(chunksize, sizeof(chunksize), "%x\015\012", nbyte);
+    if (write_it_all(fb, chunksize, strlen(chunksize)) == -1)
+	return -1;
+    if (write_it_all(fb, buf, nbyte) == -1)
+	return -1;
+    if (write_it_all(fb, "\015\012", 2) == -1)
+	return -1;
+    return nbyte;
+#else
+
+#define NVEC	(sizeof(vec)/sizeof(vec[0]))
+
+    vec[0].iov_base = chunksize;
+    vec[0].iov_len = ap_snprintf(chunksize, sizeof(chunksize), "%x\015\012",
+	nbyte);
+    vec[1].iov_base = (void *)buf;	/* cast is to avoid const warning */
+    vec[1].iov_len = nbyte;
+    vec[2].iov_base = "\r\n";
+    vec[2].iov_len = 2;
+    /* while it's nice an easy to build the vector and crud, it's painful
+     * to deal with a partial writev()
+     */
+    for( i = 0; i < NVEC; ) {
+	do rv = writev( fb->fd, &vec[i], NVEC - i );
+	while (rv == -1 && errno == EINTR && !(fb->flags & B_EOUT));
+	if (rv == -1)
+	    return -1;
+	/* recalculate vec to deal with partial writes */
+	while (rv > 0) {
+	    if( rv <= vec[i].iov_len ) {
+		vec[i].iov_base = (char *)vec[i].iov_base + rv;
+		vec[i].iov_len -= rv;
+		rv = 0;
+		if( vec[i].iov_len == 0 ) {
+		    ++i;
+		}
+	    } else {
+		rv -= vec[i].iov_len;
+		++i;
+	    }
+	}
+	if (fb->flags & B_EOUT)
+	    return -1;
+    }
+    /* if we got here, we wrote it all */
+    return nbyte;
+#undef NVEC
+#endif
+}
+
+
 /*
  * Write nbyte bytes.
  * Only returns fewer than nbyte if an error ocurred.
  * Returns -1 if no bytes were written before the error ocurred.
+ * It is worth noting that if an error occurs, the buffer is in an unknown
+ * state.
  */
 int
 bwrite(BUFF *fb, const void *buf, int nbyte)
@@ -423,17 +730,24 @@ bwrite(BUFF *fb, const void *buf, int nbyte)
 
     if (!(fb->flags & B_WR))
     {
-/* unbuffered write */
-	do i = write(fb->fd, buf, nbyte);
-	while (i == -1 && errno == EINTR);
-	if (i > 0) fb->bytes_sent += i;
-	if (i == 0)
-	{
-	    i = -1;  /* return of 0 means non-blocking */
+/* unbuffered write -- have to use bcwrite since we aren't taking care
+ * of chunking any other way */
+	do i = bcwrite(fb, buf, nbyte);
+	while (i == -1 && errno == EINTR && !(fb->flags & B_EOUT));
+	if (i == 0) {  /* return of 0 means non-blocking */
 	    errno = EAGAIN;
+	    return -1;
 	}
-	if (i == -1 && errno != EAGAIN) doerror(fb, B_WR);
-	return i;
+	else if (i < 0) {
+	    if (errno != EAGAIN)
+	        doerror(fb, B_WR);
+	    return -1;
+	}
+	fb->bytes_sent += i;
+	if (fb->flags & B_EOUT)
+	    return -1;
+	else
+	    return i;
     }
 
 /*
@@ -457,67 +771,75 @@ bwrite(BUFF *fb, const void *buf, int nbyte)
 	}
 
 /* the buffer must be full */
-	do i = write(fb->fd, fb->outbase, fb->bufsiz);
-	while (i == -1 && errno == EINTR);
-	if (i > 0) fb->bytes_sent += i;
-	if (i == 0)
-	{
-	    i = -1;  /* return of 0 means non-blocking */
-	    errno = EAGAIN;
+	if (fb->flags & B_CHUNK) {
+	    end_chunk(fb);
+	    /* it is just too painful to try to re-cram the buffer while
+	     * chunking
+	     */
+	    i = (write_it_all(fb, fb->outbase, fb->outcnt) == -1) ?
+	            -1 : fb->outcnt;
 	}
-	if (i == -1)
-	{
-	    if (nwr == 0)
-	    {
+	else {
+	    do i = write(fb->fd, fb->outbase, fb->outcnt);
+	    while (i == -1 && errno == EINTR && !(fb->flags & B_EOUT));
+	}
+	if (i <= 0) {
+	    if (i == 0) /* return of 0 means non-blocking */
+	        errno = EAGAIN;
+	    if (nwr == 0) {
 		if (errno != EAGAIN) doerror(fb, B_WR);
 		return -1;
 	    }
 	    else return nwr;
 	}
+	fb->bytes_sent += i;
 
-/*
- * we should have written all the data, however if the fd was in a
- * strange (non-blocking) mode, then we might not have done so.
- */
-	if (i < fb->bufsiz)
+	/* deal with a partial write */
+	if (i < fb->outcnt)
 	{
-	    int j, n=fb->bufsiz;
+	    int j, n=fb->outcnt;
 	    unsigned char *x=fb->outbase;
 	    for (j=i; j < n; j++) x[j-i] = x[j];
-	    fb->outcnt = fb->bufsiz - i;
-	} else
+	    fb->outcnt -= i;
+	}
+	else
 	    fb->outcnt = 0;
+
+	if (fb->flags & B_EOUT)
+	    return -1;
     }
 /* we have emptied the file buffer. Now try to write the data from the
- * original buffer until there is less than bufsiz left
+ * original buffer until there is less than bufsiz left.  Note that we
+ * use bcwrite() to do this for us, it will do the chunking so that
+ * we don't have to dink around building a chunk in our own buffer.
  */
-    while (nbyte > fb->bufsiz)
+    while (nbyte >= fb->bufsiz)
     {
-	do i = write(fb->fd, buf, nbyte);
-	while (i == -1 && errno == EINTR);
-	if (i > 0) fb->bytes_sent += i;
-	if (i == 0)
-	{
-	    i = -1;  /* return of 0 means non-blocking */
-	    errno = EAGAIN;
-	}
-	if (i == -1)
-	{
-	    if (nwr == 0)
-	    {
+	do i = bcwrite(fb, buf, nbyte);
+	while (i == -1 && errno == EINTR && !(fb->flags & B_EOUT));
+	if (i <= 0) {
+	    if (i == 0) /* return of 0 means non-blocking */
+	        errno = EAGAIN;
+	    if (nwr == 0) {
 		if (errno != EAGAIN) doerror(fb, B_WR);
 		return -1;
 	    }
 	    else return nwr;
 	}
+	fb->bytes_sent += i;
 
 	buf = i + (const char *)buf;
 	nwr += i;
 	nbyte -= i;
+
+	if (fb->flags & B_EOUT)
+	    return -1;
     }
 /* copy what's left to the file buffer */
-    if (nbyte > 0) memcpy(fb->outbase, buf, nbyte);
-    fb->outcnt = nbyte;
+    fb->outcnt = 0;
+    if( fb->flags & B_CHUNK ) start_chunk( fb );
+    if (nbyte > 0) memcpy(fb->outbase + fb->outcnt, buf, nbyte);
+    fb->outcnt += nbyte;
     nwr += nbyte;
     return nwr;
 }
@@ -529,34 +851,33 @@ bwrite(BUFF *fb, const void *buf, int nbyte)
 int
 bflush(BUFF *fb)
 {
-    int i, j;
+    int i;
 
     if (!(fb->flags & B_WR) || (fb->flags & B_EOUT)) return 0;
 
     if (fb->flags & B_WRERR) return -1;
     
+    if (fb->flags & B_CHUNK) end_chunk(fb);
+
     while (fb->outcnt > 0)
     {
-/* the buffer must be full */
-	j = fb->outcnt;
+	/* the buffer must be full */
 	do i = write(fb->fd, fb->outbase, fb->outcnt);
-	while (i == -1 && errno == EINTR);
-	if (i > 0) fb->bytes_sent += i;
-	if (i == 0)
-	{
+	while (i == -1 && errno == EINTR && !(fb->flags & B_EOUT));
+	if (i == 0) {
 	    errno = EAGAIN;
 	    return -1;  /* return of 0 means non-blocking */
 	}
-	if (i == -1)
-	{
+	else if (i < 0) {
 	    if (errno != EAGAIN) doerror(fb, B_WR);
 	    return -1;
 	}
+	fb->bytes_sent += i;
 
-/*
- * we should have written all the data, however if the fd was in a
- * strange (non-blocking) mode, then we might not have done so.
- */
+	/*
+ 	 * We should have written all the data, but if the fd was in a
+ 	 * strange (non-blocking) mode, then we might not have done so.
+ 	 */
 	if (i < fb->outcnt)
 	{
 	    int j, n=fb->outcnt;
@@ -564,6 +885,12 @@ bflush(BUFF *fb)
 	    for (j=i; j < n; j++) x[j-i] = x[j];
 	}
 	fb->outcnt -= i;
+
+	/* If a soft timeout occurs while flushing, the handler should
+	 * have set the buffer flag B_EOUT.
+	 */
+	if (fb->flags & B_EOUT)
+	    return -1;
     }
     return 0;
 }
