@@ -1,6 +1,5 @@
-
 /* ====================================================================
- * Copyright (c) 1995 The Apache Group.  All rights reserved.
+ * Copyright (c) 1995-1997 The Apache Group.  All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
@@ -51,70 +50,113 @@
  *
  */
 
-
 /*
  * conf.h: system-dependant #defines and includes...
  * See README for a listing of what they mean
  */
 
-#ifndef QNX
+#if !defined(QNX) && !defined(MPE)
 #include <sys/param.h>
 #endif
 
 /* Define one of these according to your system. */
-#if defined(SUNOS4)
-#define HAS_GMTOFF
-#define HAVE_RESOURCE 1
+#if defined(MPE)
+#include <sys/times.h>
+#define JMP_BUF sigjmp_buf
+#define NO_SETSID
+#define NO_KILLPG
+#define NO_WRITEV
+#define NEED_INITGROUPS
+#define NEED_STRCASECMP
+#define NEED_STRDUP
+#define NEED_STRNCASECMP
+extern void GETPRIVMODE();
+extern void GETUSERMODE();
+extern char *inet_ntoa();
+
+#elif defined(SUNOS4)
+#define HAVE_GMTOFF
+#define HAVE_SYS_RESOURCE_H
 #undef NO_KILLPG
 #undef NO_SETSID
-char *crypt(char *pw, char *salt);
+char *crypt(const char *pw, const char *salt);
+char *mktemp(char *template);
 #define JMP_BUF sigjmp_buf
 #define HAVE_MMAP
 #include <sys/time.h>     
 #define NEED_STRERROR
+typedef int rlim_t;
+#define memmove(a,b,c) bcopy(b,a,c)
+#define NO_LINGCLOSE
 
 #elif defined(SOLARIS2)
-#undef HAS_GMTOFF
+#undef HAVE_GMTOFF
 #define NO_KILLPG
 #undef NO_SETSID
-#define HAVE_RESOURCE 1
+#define HAVE_SYS_RESOURCE_H
 #define bzero(a,b) memset(a,0,b)
-#define getwd(d) getcwd(d,MAX_STRING_LEN)
 #define JMP_BUF sigjmp_buf
-#define FCNTL_SERIALIZED_ACCEPT
+#define USE_FCNTL_SERIALIZED_ACCEPT
 #define HAVE_MMAP
 #define HAVE_CRYPT_H
+int gethostname(char *name, int namelen);
 
 #elif defined(IRIX)
-#undef HAS_GMTOFF
-#undef NO_KILLPG
+#undef HAVE_GMTOFF
+/* IRIX has killpg, but it's only in _BSD_COMPAT, so don't use it in case
+ * there's some weird conflict with non-BSD signals */
+#define NO_KILLPG
 #undef NO_SETSID
 #define JMP_BUF sigjmp_buf
-#define FCNTL_SERIALIZED_ACCEPT
+#define USE_FCNTL_SERIALIZED_ACCEPT
 #define HAVE_SHMGET
 #define HAVE_CRYPT_H
- 
-#elif defined(HPUX)
-#define HAVE_RESOURCE 1
-#undef HAS_GMTOFF
+#define NO_LONG_DOUBLE
+#define HAVE_BSTRING_H
+#define NO_LINGCLOSE
+
+#elif defined(HIUX)
+#define HAVE_SYS_RESOURCE_H
+#undef HAVE_GMTOFF
+#define NO_KILLPG
+#undef NO_SETSID
+#ifndef _HIUX_SOURCE
+#define _HIUX_SOURCE
+#endif
+#define JMP_BUF sigjmp_buf
+#define HAVE_SHMGET
+#define SELECT_NEEDS_CAST
+
+#elif defined(HPUX) || defined(HPUX10)
+#define HAVE_SYS_RESOURCE_H
+#undef HAVE_GMTOFF
 #define NO_KILLPG
 #undef NO_SETSID
 #ifndef _HPUX_SOURCE
 #define _HPUX_SOURCE
 #endif
-#define getwd(d) getcwd(d,MAX_STRING_LEN)
 #define JMP_BUF sigjmp_buf
-#define HAVE_MMAP
+#define HAVE_SHMGET
+#ifndef HPUX10
+#define SELECT_NEEDS_CAST
+typedef int rlim_t;
+#endif
 
 #elif defined(AIX)
-#undef HAS_GMTOFF
+#undef HAVE_GMTOFF
 #undef NO_KILLPG
 #undef NO_SETSID
-#define NEED_SELECT_H
+#define HAVE_SYS_SELECT_H
 #define JMP_BUF sigjmp_buf
+#ifndef __ps2__
+#define HAVE_MMAP
+#define DEFAULT_GROUP "nobody"
+#endif
+#define DEFAULT_USER "nobody"
+typedef int rlim_t;
 
 #elif defined(ULTRIX)
-#define HAS_GMTOFF
+#define HAVE_GMTOFF
 #undef NO_KILLPG
 #undef NO_SETSID
 #define ULTRIX_BRAIN_DEATH
@@ -126,27 +168,39 @@ char *crypt(char *pw, char *salt);
 #define JMP_BUF sigjmp_buf
 
 #elif defined(OSF1)
-#define HAS_GMTOFF
+#define HAVE_GMTOFF
 #undef NO_KILLPG
 #undef NO_SETSID
 #define JMP_BUF sigjmp_buf
 #define HAVE_MMAP
 #define HAVE_CRYPT_H
+#define NO_LONG_DOUBLE
+
+#elif defined(PARAGON)
+#define HAVE_GMTOFF
+#undef NO_KILLPG
+#undef NO_SETSID
+#define JMP_BUF sigjmp_buf
+#define HAVE_MMAP
+#define HAVE_CRYPT_H
+#define NO_LONG_DOUBLE
+typedef int rlim_t;
 
 #elif defined(SEQUENT)
-#define HAS_GMTOFF
+#define HAVE_GMTOFF
 #undef NO_KILLPG
 #define NO_SETSID
 #define NEED_STRDUP
 #define tolower(c) (isupper(c) ? tolower(c) : c)
 
 #elif defined(NEXT)
-#include <libc.h>
 typedef unsigned short mode_t;
-#define HAS_GMTOFF
+#define HAVE_GMTOFF
 #undef NO_KILLPG
 #define NO_SETSID
 #define NEED_STRDUP
+#define NO_LINGCLOSE
+#define NO_UNISTD_H
 #undef _POSIX_SOURCE
 #ifndef FD_CLOEXEC
 #define FD_CLOEXEC 1
@@ -178,39 +232,58 @@ typedef unsigned short mode_t;
 #ifndef S_IWUSR
 #define S_IWUSR S_IWRITE
 #endif
+#ifndef S_IWGRP
+#define S_IWGRP	000020
+#endif
+#ifndef S_IWOTH
+#define S_IWOTH 000002
+#ifndef rlim_t
+typedef int rlim_t;
+#endif
+typedef u_long  n_long;
+#endif
+
 #define STDIN_FILENO  0
 #define STDOUT_FILENO 1
 #define STDERR_FILENO 2
-#define waitpid(a,b,c) wait4(a,(union wait *)b,c,NULL)
+#define waitpid(a,b,c) wait4((a) == -1 ? 0 : (a),(union wait *)(b),c,NULL)
 typedef int pid_t;
 #define JMP_BUF jmp_buf
+#define USE_LONGJMP
 #define NO_USE_SIGACTION
 
 #elif defined(LINUX)
-#undef HAS_GMTOFF
+#if LINUX > 1
+#define HAVE_SHMGET
+#define HAVE_SYS_RESOURCE_H
+typedef int rlim_t;
+#endif
+#define USE_FCNTL_SERIALIZED_ACCEPT
+#undef HAVE_GMTOFF
 #undef NO_KILLPG
 #undef NO_SETSID
 #undef NEED_STRDUP
 #define JMP_BUF sigjmp_buf
-#define FCNTL_SERIALIZED_ACCEPT
 #include <sys/time.h>     
 
 #elif defined(SCO)
-#undef HAS_GMTOFF
+#undef HAVE_GMTOFF
 #undef NO_KILLPG
 #undef NO_SETSID
 #define NEED_INITGROUPS
+#define NO_WRITEV
 #define JMP_BUF sigjmp_buf
 #define SIGURG SIGUSR1 /* but note, this signal will be sent to a process group if enabled (for OOB data). It is not currently enabled. */
-#define getwd(d) getcwd(d,MAX_STRING_LEN)
+#include <sys/time.h>     
 
 #elif defined(SCO5)
 
 #define JMP_BUF sigjmp_buf
 #define SIGURG SIGUSR1
-#define NEED_SELECT_H
-#define FCNTL_SERIALIZED_ACCEPT
+#define HAVE_SYS_SELECT_H
+#define USE_FCNTL_SERIALIZED_ACCEPT
 #define HAVE_MMAP
+#define HAVE_SYS_RESOURCE_H
 #define SecureWare
 
 /* Although SCO 5 defines these in <strings.h> (note the "s") they don't have
@@ -219,23 +292,41 @@ extern int strcasecmp(const char *,const char *);
 extern int strncasecmp(const char *,const char *,unsigned);
 
 #elif defined(CONVEXOS)
-#define HAS_GMTOFF
+#define HAVE_GMTOFF
 #define NEED_STRDUP
-#define getwd(d) getcwd(d,MAX_STRING_LEN)
 
 #elif defined(AUX)
-#undef HAS_GMTOFF
+/* These are to let -Wall compile more cleanly */
+extern int strcasecmp(const char *, const char *);
+extern int strncasecmp(const char *,const char *,unsigned);
+extern int set42sig(), getopt(), getpeername(), bzero();
+extern int listen(), bind(), socket(), getsockname();
+extern int accept(), gethostname(), connect(), lstat();
+extern int select(), killpg(), shutdown();
+extern int initgroups(), setsockopt();
+extern char *shmat();
+extern int  shmctl();
+extern int  shmget();
+extern char *sbrk();
+extern char *crypt();
+#include <sys/time.h>
+#undef HAVE_GMTOFF
 #undef NO_KILLPG
 #undef NO_SETSID
 #define NEED_STRDUP
 #define JMP_BUF sigjmp_buf
 /* fcntl() locking is expensive with NFS */
-#undef FLOCK_SERIALIZED_ACCEPT
+#undef USE_FLOCK_SERIALIZED_ACCEPT
 #define HAVE_SHMGET
 #define MOVEBREAK		0x4000000
-/* These are to let -Wall compile more cleanly */
-extern int strcasecmp(const char *, const char *);
-extern int strncasecmp(const char *,const char *,unsigned);
+/*
+ * NOTE: If when you run Apache under A/UX and you get a warning
+ * that httpd couldn't move break, then the above value for
+ * MOVEBREAK (64megs) is too large for your setup. Try reducing
+ * to 0x2000000 which is still PLENTY of space. I doubt if
+ * even on heavy systems sbrk() would be called at all...
+ */
+#define NO_LINGCLOSE
 
 #elif defined(SVR4)
 #define NO_KILLPG
@@ -245,9 +336,26 @@ extern int strncasecmp(const char *,const char *,unsigned);
 #define NEED_STRNCASECMP
 #define bzero(a,b) memset(a,0,b)
 #define JMP_BUF sigjmp_buf
-#define getwd(d) getcwd(d,MAX_STRING_LEN)
 /* A lot of SVR4 systems need this */
-#define FCNTL_SERIALIZED_ACCEPT
+#define USE_FCNTL_SERIALIZED_ACCEPT
+
+#elif defined(UW)
+#define NO_LINGCLOSE
+#define NO_KILLPG
+#undef  NO_SETSID
+#undef NEED_STRDUP
+#define NEED_STRCASECMP
+#define NEED_STRNCASECMP
+#define bzero(a,b) memset(a,0,b)
+#define JMP_BUF sigjmp_buf
+#define HAVE_RESOURCE
+#define HAVE_MMAP
+#define HAVE_SHMGET
+#define HAVE_CRYPT_H
+#define HAVE_SYS_SELECT_H
+#define HAVE_SYS_RESOURCE_H
+#include <sys/time.h>
+#define _POSIX_SOURCE
 
 #elif defined(DGUX)
 #define NO_KILLPG
@@ -257,18 +365,20 @@ extern int strncasecmp(const char *,const char *,unsigned);
 #define NEED_STRNCASECMP
 #define bzero(a,b) memset(a,0,b)
 #define JMP_BUF sigjmp_buf
-#define getwd(d) getcwd(d,MAX_STRING_LEN)
 /* A lot of SVR4 systems need this */
-#define FCNTL_SERIALIZED_ACCEPT
+#define USE_FCNTL_SERIALIZED_ACCEPT
 
-#elif defined(__NetBSD__)
-#define HAS_GMTOFF
+#elif defined(__NetBSD__) || defined(__OpenBSD__)
+#define HAVE_SYS_RESOURCE_H
+#define HAVE_GMTOFF
 #undef NO_KILLPG
 #undef NO_SETSID
 #define JMP_BUF sigjmp_buf
+#define DEFAULT_USER "nobody"
+#define DEFAULT_GROUP "nogroup"
 
 #elif defined(UTS21)
-#undef HAS_GMTOFF
+#undef HAVE_GMTOFF
 #undef NO_KILLPG
 #define NO_SETSID
 #define NEED_WAITPID
@@ -279,34 +389,68 @@ extern int strncasecmp(const char *,const char *,unsigned);
 #include <sys/types.h>
 
 #elif defined(APOLLO)
-#undef HAS_GMTOFF
+#undef HAVE_GMTOFF
 #undef NO_KILLPG
 #undef NO_SETSID
-#define timezone	_bky_timezone
 
 #elif defined(__FreeBSD__) || defined(__bsdi__)
-#define HAS_GMTOFF
+#if defined(__FreeBSD__)
+#include <osreldate.h>
+#endif
+#define HAVE_SYS_RESOURCE_H
+#define HAVE_GMTOFF
 #undef NO_KILLPG
 #undef NO_SETSID
 #define JMP_BUF sigjmp_buf
 #define HAVE_MMAP
+#define DEFAULT_USER "nobody"
+#define DEFAULT_GROUP "nogroup"
+#if defined(__bsdi__) || \
+(defined(__FreeBSD_version) && (__FreeBSD_version < 220000))
+typedef quad_t rlim_t;
+#endif
 
 #elif defined(QNX)
+#ifndef crypt
+char *crypt(const char *pw, const char *salt);
+#endif
+#ifndef initgroups
+int initgroups (char *, int);     
+#endif
+#ifndef strncasecmp
+#define strncasecmp strnicmp
+#endif
 #undef NO_KILLPG
 #undef NO_SETSID
 #define NEED_INITGROUPS
 #define NEED_SELECT_H
+#define NEED_PROCESS_H
+#define HAVE_SYS_SELECT_H
+#include <unix.h>
 #define JMP_BUF sigjmp_buf
 
 #elif defined(LYNXOS)
 #undef NO_KILLPG
 #undef NO_SETSID
-#define NO_TIMEZONE
 #define NEED_STRCASECMP
 #define NEED_STRNCASECMP
 #define NEED_INITGROUPS
 #define JMP_BUF jmp_buf
 
+#elif defined(UXPDS)
+#undef NEED_STRCASECMP
+#undef NEED_STRNCASECMP
+#undef NEED_STRDUP
+#undef HAVE_GMTOFF
+#define NO_KILLPG
+#undef NO_SETSID
+#define HAVE_RESOURCE 1
+#define bzero(a,b) memset(a,0,b)
+#define JMP_BUF sigjmp_buf
+#define USE_FCNTL_SERIALIZED_ACCEPT
+#define HAVE_MMAP
+#define HAVE_CRYPT_H
+ 
 #elif defined(__EMX__)
 /* Defines required for EMX OS/2 port. */
 #define JMP_BUF sigjmp_buf
@@ -316,13 +460,52 @@ extern int strncasecmp(const char *,const char *,unsigned);
 #define NO_SETSID
 /* Add some drive name support */
 #define chdir _chdir2
+#include <sys/time.h>     
+#define MAXSOCKETS 4096
+#define HAVE_MMAP
+    
+#elif defined(__MACHTEN__)
+typedef int rlim_t;
+#define JMP_BUF sigjmp_buf
+#undef NO_KILLPG
+#define NO_SETSID
+#define HAVE_GMTOFF
+#ifndef __MACHTEN_PPC__
+#ifndef __MACHTEN_68K__
+#define __MACHTEN_68K__
+#endif
+#define USE_FLOCK_SERIALIZED_ACCEPT
+#define NO_USE_SIGACTION
+#define USE_LONGJMP
+#undef NEED_STRDUP
+#else
+#define HAVE_SHMGET
+#define USE_FCNTL_SERIALIZED_ACCEPT
+#endif
+
+/* Convex OS v11 */
+#elif defined(CONVEXOS11)
+#define NO_TIMEZONE
+#include <stdio.h>
+#include <sys/types.h>
+#define JMP_BUF jmp_buf
+typedef int rlim_t;
+
+#elif defined(ISC)
+#include <net/errno.h>     
+#define NO_KILLPG
+#undef NO_SETSID
+#define HAVE_SHMGET
+#define SIGURG SIGUSR1
+#define JMP_BUF sigjmp_buf
+#define USE_FCNTL_SERIALIZED_ACCEPT
 
 /* Unknown system - Edit these to match */
 #else
 #ifdef BSD
-#define HAS_GMTOFF
+#define HAVE_GMTOFF
 #else
-#undef HAS_GMTOFF
+#undef HAVE_GMTOFF
 #endif
 /* NO_KILLPG is set on systems that don't have killpg */
 #undef NO_KILLPG
@@ -333,13 +516,11 @@ extern int strncasecmp(const char *,const char *,unsigned);
 #endif
 
 /* Do we have sys/resource.h; assume that BSD does. */
-#ifndef HAVE_RESOURCE
+#ifndef HAVE_SYS_RESOURCE_H
 #ifdef BSD
-#define HAVE_RESOURCE 1
-#else
-#define HAVE_RESOURCE 0
+#define HAVE_SYS_RESOURCE_H
 #endif
-#endif /* HAVE_RESOURCE */
+#endif /* HAVE_SYS_RESOURCE_H */
 
 /*
  * The particular directory style your system supports. If you have dirent.h
@@ -350,6 +531,19 @@ extern int strncasecmp(const char *,const char *,unsigned);
  */
 
 #include <sys/types.h>
+#include <stdarg.h>
+/*
+ * We use snprintf() to avoid overflows, but we include
+ * our own version (ap_snprintf). Allow for people to use their
+ * snprintf() if they want
+ */
+#ifdef HAVE_SNPRINTF
+#define ap_snprintf     snprintf
+#define ap_vsnprintf    vsnprintf
+#else
+int ap_snprintf(char *buf, size_t len, const char *format,...);
+int ap_vsnprintf(char *buf, size_t len, const char *format, va_list ap);
+#endif
 
 #if !defined(NEXT) && !defined(CONVEXOS)
 #include <dirent.h>
@@ -363,16 +557,20 @@ extern int strncasecmp(const char *,const char *,unsigned);
 #include <stdlib.h>
 #include <string.h>
 #include <sys/stat.h>
+#ifndef MPE
 #include <sys/file.h>
+#endif
 #include <sys/socket.h>
-#ifdef NEED_SELECT_H
+#ifdef HAVE_SYS_SELECT_H
 #include <sys/select.h>
 #endif
 #include <ctype.h>
 #include <netinet/in.h>
 #include <netdb.h>
 #include <sys/ioctl.h>
+#ifndef MPE
 #include <arpa/inet.h>  /* for inet_ntoa */
+#endif
 #include <time.h>  /* for ctime */
 #include <signal.h>
 #include <errno.h>
@@ -381,11 +579,16 @@ extern int strncasecmp(const char *,const char *,unsigned);
 #include <grp.h>
 #include <fcntl.h>
 #include <limits.h>
-#ifndef QNX
+#if !defined(QNX) && !defined(CONVEXOS11) && !defined(NEXT)
 #include <memory.h>
 #endif
+#ifdef NEED_PROCESS_H
+#include <process.h>
+#endif
 
-#if HAVE_RESOURCE
+#include <regex.h>
+
+#ifdef HAVE_SYS_RESOURCE_H
 #include <sys/resource.h>
 #ifdef SUNOS4
 int getrlimit( int, struct rlimit *);
@@ -393,8 +596,11 @@ int setrlimit( int, struct rlimit *);
 #endif
 #endif
 #ifdef HAVE_MMAP
+#ifndef __EMX__
+/* This file is not needed for OS/2 */
 #include <sys/mman.h>
 #endif
+#endif
 #if !defined(MAP_ANON) && defined(MAP_ANONYMOUS)
 #define MAP_ANON MAP_ANONYMOUS
 #endif
@@ -416,8 +622,45 @@ int setrlimit( int, struct rlimit *);
 #endif
 
 #ifndef S_ISLNK
+#ifndef __EMX__
+/* Don't define this for OS/2 */
 #define S_ISLNK(m) (((m) & S_IFMT) == S_IFLNK)
 #endif
+#endif
+
+#ifndef INADDR_NONE
+#define INADDR_NONE ((unsigned long) -1)
+#endif
+
+/*
+ * Replace signal function with sigaction equivalent
+ */
+#ifndef NO_USE_SIGACTION
+typedef void Sigfunc(int);
+
+#if defined(SIG_IGN) && !defined(SIG_ERR)
+#define SIG_ERR ((Sigfunc *)-1)
+#endif
+
+/*
+ * For some strange reason, QNX defines signal to signal. Eliminate it.
+ */
+#ifdef signal
+#undef signal
+#endif
+#define signal(s,f)	ap_signal(s,f)
+Sigfunc *signal(int signo, Sigfunc *func);
+#endif
+
+#include <setjmp.h>
+
+#if defined(USE_LONGJMP)
+#define ap_longjmp(x, y)        longjmp((x), (y))
+#define ap_setjmp(x)            setjmp(x)
+#else
+#define ap_longjmp(x, y)        siglongjmp((x), (y))
+#define ap_setjmp(x)            sigsetjmp((x), 1)
+#endif
 
 /* Finding offsets of elements within structures.
  * Taken from the X code... they've sweated portability of this stuff
@@ -487,7 +730,6 @@ int wait3 (int *, int, void*);	/* Close enough for us... */
 int lstat (const char *, struct stat *);
 int stat (const char *, struct stat *);     
 int flock (int, int);
-int getwd (char *);
 #ifndef NO_KILLPG
 int killpg(int, int);
 #endif
@@ -505,7 +747,6 @@ int gethostname (char *name, int namelen);
 void syslog (int, char *, ...);
 char *mktemp (char *);
      
-#include <stdarg.h>
 long vfprintf (FILE *, char *, va_list);
      
 #endif
