@@ -1,6 +1,5 @@
-
 /* ====================================================================
- * Copyright (c) 1995 The Apache Group.  All rights reserved.
+ * Copyright (c) 1995-1997 The Apache Group.  All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
@@ -51,7 +50,6 @@
  *
  */
 
-
 /* 
  * Process config --- what the process ITSELF is doing
  */
@@ -60,12 +58,17 @@ extern int standalone;
 extern uid_t user_id;
 extern char *user_name;
 extern gid_t group_id;
+#ifdef MULTIPLE_GROUPS
+extern gid_t group_id_list[NGROUPS_MAX];
+#endif
 extern int max_requests_per_child;
 extern struct in_addr bind_address;
+extern listen_rec *listeners;
 extern int daemons_to_start;
 extern int daemons_min_free;
 extern int daemons_max_free;
 extern int daemons_limit;
+extern int suexec_enabled;
 
 extern char *pid_fname;
 extern char *scoreboard_fname;
