@@ -1,6 +1,5 @@
-
 /* ====================================================================
- * Copyright (c) 1995 The Apache Group.  All rights reserved.
+ * Copyright (c) 1995-1997 The Apache Group.  All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
@@ -51,17 +50,16 @@
  *
  */
 
-
 /*
  * http_config.c: once was auxillary functions for reading httpd's config
  * file and converting filenames into a namespace
  *
  * Rob McCool 
  * 
- * Wall-to-wall rewrite for Shambhala... commands which are part of the
+ * Wall-to-wall rewrite for Apache... commands which are part of the
  * server core can now be found next door in "http_core.c".  Now contains
  * general command loop, and functions which do bookkeeping for the new
- * Shambhala config stuff (modules and configuration vectors).
+ * Apache config stuff (modules and configuration vectors).
  *
  * rst
  *
@@ -75,15 +73,20 @@
 #include "http_log.h"		/* for errors in parse_htaccess */
 #include "http_request.h"	/* for default_handler (see invoke_handler) */
 #include "http_conf_globals.h"	/* Sigh... */
+#include "explain.h"
+
+DEF_Explain
 
 /****************************************************************
  *
  * We begin with the functions which deal with the linked list
- * of modules which control just about all of server operation in
- * Shambhala.
+ * of modules which control just about all of the server operation.
  */
 
+/* num_modules is the number of currently active modules.  */
 static int num_modules = 0;    
+/* total_modules is the number of modules linked in.  */
+static int total_modules = 0;
 module *top_module = NULL;
     
 typedef int (*handler)(request_rec *);
@@ -119,14 +122,14 @@ set_module_config (void *conf_vector, module *m, void *val)
 void *
 create_empty_config (pool *p)
 {
-   void **conf_vector = (void **)pcalloc(p, sizeof(void*) * num_modules);
+   void **conf_vector = (void **)pcalloc(p, sizeof(void*) * total_modules);
    return (void *)conf_vector;
 }
 
 void *
 create_default_per_dir_config (pool *p)
 {
-   void **conf_vector = (void **)pcalloc(p, sizeof(void*) * (num_modules+DYNAMIC_MODULE_LIMIT));
+   void **conf_vector = (void **)pcalloc(p, sizeof(void*) * (total_modules+DYNAMIC_MODULE_LIMIT));
    module *modp;
 
    for (modp = top_module; modp; modp = modp->next) {
@@ -141,7 +144,7 @@ create_default_per_dir_config (pool *p)
 void *
 merge_per_dir_configs (pool *p, void *base, void *new)
 {
-   void **conf_vector = (void **)pcalloc(p, sizeof(void*) * num_modules);
+   void **conf_vector = (void **)pcalloc(p, sizeof(void*) * total_modules);
    void **base_vector = (void **) base;
    void **new_vector = (void **) new;
    module *modp;
@@ -162,7 +165,7 @@ merge_per_dir_configs (pool *p, void *base, void *new)
 void *
 create_server_config (pool *p, server_rec *s)
 {
-   void **conf_vector = (void **)pcalloc(p, sizeof(void*) * (num_modules+DYNAMIC_MODULE_LIMIT));
+   void **conf_vector = (void **)pcalloc(p, sizeof(void*) * (total_modules+DYNAMIC_MODULE_LIMIT));
    module *modp;
 
    for (modp = top_module; modp; modp = modp->next) {
@@ -206,6 +209,40 @@ void *create_per_dir_config (pool *p) {
     return create_empty_config (p);
 }
 
+#ifdef EXPLAIN
+
+struct
+    {
+    int offset;
+    char *method;
+    } aMethods[]=
+    {
+#define m(meth)	{ XtOffsetOf(module,meth),#meth }
+    m(translate_handler),
+    m(check_user_id),
+    m(auth_checker),
+    m(type_checker),
+    m(fixer_upper),
+    m(logger),
+    { -1,"?" },
+#undef m
+    };
+
+char *ShowMethod(module *modp,int offset)
+    {
+    int n;
+    static char buf[200];
+
+    for(n=0 ; aMethods[n].offset >= 0 ; ++n)
+	if(aMethods[n].offset == offset)
+	    break;
+    ap_snprintf(buf, sizeof(buf), "%s:%s",modp->name,aMethods[n].method);
+    return buf;
+    }
+#else
+#define ShowMethod(modp,offset)
+#endif
+
 /****************************************************************
  *
  * Dispatch through the modules to find handlers for various phases
@@ -221,8 +258,12 @@ run_method (request_rec *r, int offset, int run_all)
        handler mod_handler = *(handler *)(offset + (char *)(modp));
 
        if (mod_handler) {
-	   int result = (*mod_handler)(r);
+           int result;
 
+           Explain1("Run %s",ShowMethod(modp,offset));
+	   result = (*mod_handler)(r);
+
+	   Explain2("%s returned %d",ShowMethod(modp,offset),result);
 	   if (result != DECLINED && (!run_all || result != OK))
 	       return result;
        }
@@ -251,6 +292,10 @@ int log_transaction (request_rec *r) {
    return run_method (r, XtOffsetOf (module, logger), 1);
 }
 
+int header_parse (request_rec *r) {
+    return run_method (r, XtOffsetOf (module, header_parser), 1);
+}
+
 /* Auth stuff --- anything that defines one of these will presumably
  * want to define something for the other.  Note that check_auth is
  * separate from check_access to make catching some config errors easier.
@@ -269,7 +314,13 @@ int invoke_handler (request_rec *r)
    module *modp;
    handler_rec *handp;
    char *content_type = r->content_type ? r->content_type : default_type (r);
-   char *handler = r->handler ? r->handler : content_type;
+   char *handler, *p; 
+
+   if ((p = strchr(content_type, ';')) != NULL) {  /* MIME type arguments */
+       while (p > content_type && p[-1] == ' ') --p; /* strip trailing spaces */
+       content_type = pstrndup(r->pool, content_type, p - content_type);
+   }
+   handler = r->handler ? r->handler : content_type;
   
    /* Pass one --- direct matches */
    
@@ -321,20 +372,95 @@ void add_module (module *m)
      * teased out...
      */
 
-    m->next = top_module;
-    top_module = m;
-    m->module_index = num_modules++;
+    /* At some point, we may want to offer back-compatibility for
+     * loading modules that are for older versions of Apache. For now,
+     * though, we don't.
+     */
+
+    if (m->version != MODULE_MAGIC_NUMBER) {
+	fprintf(stderr, "httpd: module \"%s\" is not compatible with this "
+		        "version of Apache.\n", m->name);
+	fprintf(stderr, "Please contact the author for the correct version.\n");
+	exit(1);
+    }
+
+    if (m->next == NULL) {
+        m->next = top_module;
+	top_module = m;
+    }
+    if (m->module_index == -1) {
+	m->module_index = num_modules++;
+    }
 }
 
 void setup_prelinked_modules ()
 {
-    extern module *prelinked_modules[];
-    module **m = prelinked_modules;
+    extern module *prelinked_modules[], *preloaded_modules[];
+    module **m;
+
+    /* First, set all module indices, and init total_modules.  */
+    total_modules = 0;
+    for (m = preloaded_modules; *m; ++m, ++total_modules) {
+        (*m)->module_index = total_modules;
+    }
+
+    for (m = prelinked_modules; *m; ++m) {
+        add_module (*m);
+    }
+}
+
+const char *find_module_name (module *m)
+{
+    return m->name;
+}
+
+module *find_linked_module (const char *name)
+{
+    module *modp;
+
+    for (modp = top_module; modp; modp = modp->next) {
+        if (strcmp(modp->name, name) == 0)
+            return modp;
+    }
+    return NULL;
+}
+
+/* Add a named module.  Returns 1 if module found, 0 otherwise.  */
+int add_named_module (const char *name)
+{
+    extern module *preloaded_modules[];
+    module *modp;
+    int i = 0;
+
+    for (modp = preloaded_modules[i]; modp; modp = preloaded_modules[++i]) {
+        if (strcmp(modp->name, name) == 0) {
+	    /* Only add modules that are not already enabled.  */
+	    if (modp->next == NULL) {
+	        add_module(modp);
+	    }
+	    return 1;
+	}
+    }
+
+    return 0;
+}
+
+/* Clear the internal list of modules, in preparation for starting over. */
+void clear_module_list ()
+{
+    module **m = &top_module;
+    module **next_m;
 
     while (*m) {
-        add_module (*m);
-	++m;
+	next_m = &((*m)->next);
+	*m = NULL;
+	m = next_m;
     }
+
+    num_modules = 0;
+
+    /* This is required; so we add it always.  */
+    add_named_module ("http_core.c");
 }
 
 /*****************************************************************
@@ -346,14 +472,17 @@ void setup_prelinked_modules ()
  * invoking the function...
  */
 
-char *invoke_cmd(command_rec *cmd, cmd_parms *parms, void *mconfig, char *args)
+const char *invoke_cmd(const command_rec *cmd, cmd_parms *parms, void *mconfig,
+		 const char *args)
 {
-    char *w, *w2, *errmsg;
+    char *w, *w2, *w3;
+    const char *errmsg;
 
     if ((parms->override & cmd->req_override) == 0)
         return pstrcat (parms->pool, cmd->name, " not allowed here", NULL);
     
     parms->info = cmd->cmd_data;
+    parms->cmd = cmd;
     
     switch (cmd->args_how) {
     case RAW_ARGS:
@@ -386,6 +515,65 @@ char *invoke_cmd(command_rec *cmd, cmd_parms *parms, void *mconfig, char *args)
 
 	return (*cmd->func) (parms, mconfig, w, w2);
 	
+    case TAKE12:
+
+	w = getword_conf (parms->pool, &args);
+	w2 = getword_conf (parms->pool, &args);
+	
+	if (*w == '\0' || *args != 0) 
+	    return pstrcat (parms->pool, cmd->name, " takes 1-2 arguments",
+			    cmd->errmsg ? ", " : NULL, cmd->errmsg, NULL);
+
+	return (*cmd->func) (parms, mconfig, w, *w2 ? w2 : NULL);
+	
+    case TAKE3:
+
+	w = getword_conf (parms->pool, &args);
+	w2 = getword_conf (parms->pool, &args);
+	w3 = getword_conf (parms->pool, &args);
+	
+	if (*w == '\0' || *w2 == '\0' || *w3 == '\0' || *args != 0) 
+	    return pstrcat (parms->pool, cmd->name, " takes three arguments",
+			    cmd->errmsg ? ", " : NULL, cmd->errmsg, NULL);
+
+	return (*cmd->func) (parms, mconfig, w, w2, w3);
+	
+    case TAKE23:
+
+	w = getword_conf (parms->pool, &args);
+	w2 = getword_conf (parms->pool, &args);
+	w3 = *args ? getword_conf (parms->pool, &args) : NULL;
+	
+	if (*w == '\0' || *w2 == '\0' || *args != 0) 
+	    return pstrcat (parms->pool, cmd->name, " takes two or three arguments",
+			    cmd->errmsg ? ", " : NULL, cmd->errmsg, NULL);
+
+	return (*cmd->func) (parms, mconfig, w, w2, w3);
+	
+    case TAKE123:
+
+	w = getword_conf (parms->pool, &args);
+	w2 = *args ? getword_conf (parms->pool, &args) : NULL;
+	w3 = *args ? getword_conf (parms->pool, &args) : NULL;
+	
+	if (*w == '\0' || *args != 0) 
+	    return pstrcat (parms->pool, cmd->name, " takes one, two or three arguments",
+			    cmd->errmsg ? ", " : NULL, cmd->errmsg, NULL);
+
+	return (*cmd->func) (parms, mconfig, w, w2, w3);
+	
+    case TAKE13:
+
+	w = getword_conf (parms->pool, &args);
+	w2 = *args ? getword_conf (parms->pool, &args) : NULL;
+	w3 = *args ? getword_conf (parms->pool, &args) : NULL;
+	
+	if (*w == '\0' || (*w2 && !w3) || *args != 0) 
+	    return pstrcat (parms->pool, cmd->name, " takes one or three arguments",
+			    cmd->errmsg ? ", " : NULL, cmd->errmsg, NULL);
+
+	return (*cmd->func) (parms, mconfig, w, w2, w3);
+	
     case ITERATE:
 
 	while (*(w = getword_conf (parms->pool, &args)) != '\0')
@@ -414,7 +602,7 @@ char *invoke_cmd(command_rec *cmd, cmd_parms *parms, void *mconfig, char *args)
 
 	w = getword_conf (parms->pool, &args);
 
-	if (*w == '\0' || ((!strcasecmp(w, "on")) && (!strcasecmp (w, "off"))))
+	if (*w == '\0' || (strcasecmp(w, "on") && strcasecmp (w, "off")))
 	    return pstrcat (parms->pool, cmd->name, " must be On or Off",
 			    NULL);
 
@@ -428,7 +616,7 @@ char *invoke_cmd(command_rec *cmd, cmd_parms *parms, void *mconfig, char *args)
     }
 }
 
-command_rec *find_command (char *name, command_rec *cmds)
+const command_rec *find_command (const char *name, const command_rec *cmds)
 {
     while (cmds->name) 
         if (!strcasecmp (name, cmds->name))
@@ -439,12 +627,12 @@ command_rec *find_command (char *name, command_rec *cmds)
     return NULL;
 }
     
-command_rec *find_command_in_modules (char *cmd_name, module **mod)
+const command_rec *find_command_in_modules (const char *cmd_name, module **mod)
 {
-   command_rec *cmdp;
+   const command_rec *cmdp;
    module *modp;
 
-   for (modp = top_module; modp; modp = modp->next) 
+   for (modp = *mod; modp; modp = modp->next) 
        if (modp->cmds && (cmdp = find_command (cmd_name, modp->cmds))) {
 	   *mod = modp;
 	   return cmdp;
@@ -453,11 +641,11 @@ command_rec *find_command_in_modules (char *cmd_name, module **mod)
    return NULL;
 }
 
-char *handle_command (cmd_parms *parms, void *config, char *l)
+const char *handle_command (cmd_parms *parms, void *config, const char *l)
 {
-    char *args, *cmd_name;
-    command_rec *cmd;
-    module *mod;
+    const char *args, *cmd_name, *retval;
+    const command_rec *cmd;
+    module *mod = top_module;
 
     ++parms->config_line;
     if((l[0] == '#') || (!l[0])) return NULL;
@@ -466,33 +654,40 @@ char *handle_command (cmd_parms *parms, void *config, char *l)
     cmd_name = getword_conf (parms->temp_pool, &args);
     if (*cmd_name == '\0') return NULL;
 	
-    if (!(cmd = find_command_in_modules (cmd_name, &mod))) {
-	return pstrcat (parms->pool, "Invalid command ", cmd_name, NULL);
-    }
-    else {
-	void *mconfig = get_module_config (config, mod);
-	void *sconfig = get_module_config (parms->server->module_config, mod);
-	      
-	if (!mconfig && mod->create_dir_config) {
-	    mconfig = (*mod->create_dir_config) (parms->pool, parms->path);
-	    set_module_config (config, mod, mconfig);
+    do {
+	if (!(cmd = find_command_in_modules (cmd_name, &mod))) {
+	    return pstrcat (parms->pool, "Invalid command ", cmd_name, NULL);
 	}
+	else {
+	    void *mconfig = get_module_config (config, mod);
+	    void *sconfig =
+		get_module_config (parms->server->module_config, mod);
 	    
-	if (!sconfig && mod->create_server_config) {
-	    sconfig = (*mod->create_server_config)(parms->pool, parms->server);
-	    set_module_config (parms->server->module_config, mod, sconfig);
+	    if (!mconfig && mod->create_dir_config) {
+		mconfig = (*mod->create_dir_config) (parms->pool, parms->path);
+		set_module_config (config, mod, mconfig);
+	    }
+	    
+	    if (!sconfig && mod->create_server_config) {
+		sconfig =
+		    (*mod->create_server_config)(parms->pool, parms->server);
+		set_module_config (parms->server->module_config, mod, sconfig);
+	    }
+	    
+	    retval = invoke_cmd (cmd, parms, mconfig, args);
+	    mod = mod->next;	/* Next time around, skip this one */
 	}
-	
-	return invoke_cmd (cmd, parms, mconfig, args);
-    }
+    } while (retval && !strcmp(retval, DECLINE_CMD));
+
+    return retval;
 }
 
-char *srm_command_loop (cmd_parms *parms, void *config)
+const char *srm_command_loop (cmd_parms *parms, void *config)
 {
     char l[MAX_STRING_LEN];
     
     while (!(cfg_getline (l, MAX_STRING_LEN, parms->infile))) {
-	char *errmsg = handle_command (parms, config, l);
+	const char *errmsg = handle_command (parms, config, l);
 	if (errmsg) return errmsg;
     }
 
@@ -503,7 +698,7 @@ char *srm_command_loop (cmd_parms *parms, void *config)
  * Generic command functions...
  */
 
-char *set_string_slot (cmd_parms *cmd, char *struct_ptr, char *arg)
+const char *set_string_slot (cmd_parms *cmd, char *struct_ptr, char *arg)
 {
     /* This one's pretty generic... */
   
@@ -512,6 +707,15 @@ char *set_string_slot (cmd_parms *cmd, char *struct_ptr, char *arg)
     return NULL;
 }
 
+const char *set_flag_slot (cmd_parms *cmd, char *struct_ptr, int arg)
+{
+    /* This one's pretty generic too... */
+  
+    int offset = (int)cmd->info; 
+    *(int *)(struct_ptr + offset) = arg ? 1 : 0;
+    return NULL;
+}
+
 /*****************************************************************
  *
  * Reading whole config files...
@@ -533,10 +737,17 @@ char *server_root_relative (pool *p, char *file)
 void process_resource_config(server_rec *s, char *fname, pool *p, pool *ptemp)
 {
     FILE *cfg;
-    char *errmsg;
+    const char *errmsg;
     cmd_parms parms;
+    struct stat finfo;
     
     fname = server_root_relative (p, fname);
+
+    if (!(strcmp(fname, server_root_relative(p, RESOURCE_CONFIG_FILE))) ||
+	!(strcmp(fname, server_root_relative(p, ACCESS_CONFIG_FILE)))) {
+	if (stat(fname, &finfo) == -1)
+	    return;
+    }
     
     /* GCC's initialization extensions are soooo nice here... */
     
@@ -548,9 +759,9 @@ void process_resource_config(server_rec *s, char *fname, pool *p, pool *ptemp)
     parms.override = (RSRC_CONF|OR_ALL)&~(OR_AUTHCFG|OR_LIMIT);
     
     if(!(cfg = fopen(fname, "r"))) {
+        perror("fopen");
         fprintf(stderr,"httpd: could not open document config file %s\n",
                 fname);
-        perror("fopen");
         exit(1);
     } 
 
@@ -574,7 +785,7 @@ int parse_htaccess(void **result, request_rec *r, int override,
 {
     FILE *f;
     cmd_parms parms;
-    char *errmsg;
+    const char *errmsg;
     const struct htaccess_result *cache;
     struct htaccess_result *new;
     void *dc;
@@ -631,9 +842,88 @@ int parse_htaccess(void **result, request_rec *r, int override,
  * are with the command table in http_core.c.
  */
 
-server_rec *init_virtual_host (pool *p, char *hostname)
+/*
+ * Parses a host of the form <address>[:port]
+ * paddr is used to create a list in the order of input
+ * **paddr is the ->next pointer of the last entry (or s->addrs)
+ * *paddr is the variable used to keep track of **paddr between calls
+ * port is the default port to assume
+ */
+static void get_addresses (pool *p, char *w, server_addr_rec ***paddr, unsigned port)
+{
+    struct hostent *hep;
+    unsigned long my_addr;
+    server_addr_rec *sar;
+    char *t;
+    int i, is_an_ip_addr;
+
+    if( *w == 0 ) return;
+
+    t = strchr(w, ':');
+    if (t) {
+	if( strcmp(t+1,"*") == 0 ) {
+	    port = 0;
+	} else if( (i = atoi(t+1)) ) {
+	    port = i;
+	} else {
+	    fprintf( stderr, "Port must be numeric\n" );
+	}
+	*t = 0;
+    }
+
+    is_an_ip_addr = 0;
+    if (strcmp(w, "*") == 0) {
+	my_addr = htonl(INADDR_ANY);
+	is_an_ip_addr = 1;
+    } else if( strcmp(w, "_default_") == 0
+	    || strcmp(w, "255.255.255.255") == 0 ) {
+	my_addr = DEFAULT_VHOST_ADDR;
+	is_an_ip_addr = 1;
+    } else if(
+#ifdef DGUX
+	    ( my_addr = inet_network(w) )
+#else
+	    ( my_addr = inet_addr(w) )
+#endif
+	    != INADDR_NONE ) {
+	is_an_ip_addr = 1;
+    }
+    if( is_an_ip_addr ) {
+	sar = pcalloc( p, sizeof( server_addr_rec ) );
+	**paddr = sar;
+	*paddr = &sar->next;
+	sar->host_addr.s_addr = my_addr;
+	sar->host_port = port;
+	sar->virthost = pstrdup(p, w);
+	if (t != NULL) *t = ':';
+	return;
+    }
+
+    hep = gethostbyname(w);
+
+    if ((!hep) || (hep->h_addrtype != AF_INET || !hep->h_addr_list[0])) {
+	fprintf (stderr, "Cannot resolve host name %s --- ignoring!\n", w);
+	if (t != NULL) *t = ':';
+	return;
+    }
+
+    for( i = 0; hep->h_addr_list[i]; ++i ) {
+	sar = pcalloc( p, sizeof( server_addr_rec ) );
+	**paddr = sar;
+	*paddr = &sar->next;
+	sar->host_addr = *(struct in_addr *)hep->h_addr_list[i];
+	sar->host_port = port;
+	sar->virthost = pstrdup(p, w);
+    }
+
+    if (t != NULL) *t = ':';
+}
+
+server_rec *init_virtual_host (pool *p, const char *hostname,
+				server_rec *main_server)
 {
     server_rec *s = (server_rec *)pcalloc (p, sizeof (server_rec));
+    server_addr_rec **addrs;
 
 #ifdef RLIMIT_NOFILE
     struct rlimit limits;
@@ -641,8 +931,10 @@ server_rec *init_virtual_host (pool *p, char *hostname)
     getrlimit ( RLIMIT_NOFILE, &limits );
     if ( limits.rlim_cur < limits.rlim_max ) {
       limits.rlim_cur += 2;
-      if ( setrlimit ( RLIMIT_NOFILE, &limits ) < 0 )
+      if ( setrlimit ( RLIMIT_NOFILE, &limits ) < 0 ) {
+	perror ("setrlimit(RLIMIT_NOFILE)");
 	fprintf (stderr, "Cannot exceed hard limit for open files");
+      }
     }
 #endif
 
@@ -654,17 +946,35 @@ server_rec *init_virtual_host (pool *p, char *hostname)
     s->timeout = 0;
     s->keep_alive_timeout = 0;
     s->keep_alive = -1;
-    s->host_addr.s_addr = get_virthost_addr (hostname, &s->host_port);
-    s->port = s->host_port;  /* set them the same, by default */
+    s->keep_alive_max = -1;
+    s->error_log = main_server->error_log;
+    /* start the list of addreses */
+    addrs = &s->addrs;
+    while( hostname[0] ) {
+	get_addresses( p, getword_conf( p, &hostname ), &addrs,
+	    main_server->port );
+    }
+    /* terminate the list */
+    *addrs = NULL;
+    if( s->addrs ) {
+	if (s->addrs->host_port) {
+	    s->port = s->addrs->host_port;  /* set them the same, by default */
+	} else {
+	    /* otherwise we get a port of 0 on redirects */
+	    s->port = main_server->port;
+	}
+    }
     s->next = NULL;
 
     s->is_virtual = 1;
-    s->virthost = pstrdup(p, hostname);
     s->names = NULL;
 
     s->module_config = create_empty_config (p);
     s->lookup_defaults = create_per_dir_config (p);
     
+    s->server_uid = user_id;
+    s->server_gid = group_id;
+
     return s;
 }
 
@@ -685,9 +995,6 @@ void fixup_virtual_hosts (pool *p, server_rec *main_server)
 	    merge_per_dir_configs (p, main_server->lookup_defaults,
 				   virt->lookup_defaults);
 
-	if (virt->port == 0)
-	    virt->port = main_server->port;
-
 	if (virt->server_admin == NULL)
 	    virt->server_admin = main_server->server_admin;
 
@@ -705,6 +1012,12 @@ void fixup_virtual_hosts (pool *p, server_rec *main_server)
 
 	if (virt->keep_alive == -1)
 	    virt->keep_alive = main_server->keep_alive;
+
+	if (virt->keep_alive_max == -1)
+	    virt->keep_alive_max = main_server->keep_alive_max;
+
+	if (virt->send_buffer_size == 0)
+		virt->send_buffer_size = main_server->send_buffer_size;
     }
 }
 
@@ -740,21 +1053,24 @@ server_rec *init_server_config(pool *p)
     s->server_admin = DEFAULT_ADMIN;
     s->server_hostname = NULL; 
     s->error_fname = DEFAULT_ERRORLOG;
+    s->error_log = stderr;
     s->srm_confname = RESOURCE_CONFIG_FILE;
     s->access_confname = ACCESS_CONFIG_FILE;
     s->timeout = DEFAULT_TIMEOUT;
     s->keep_alive_timeout = DEFAULT_KEEPALIVE_TIMEOUT;
-    s->keep_alive = DEFAULT_KEEPALIVE;
+    s->keep_alive_max = DEFAULT_KEEPALIVE;
+    s->keep_alive = 1;
     s->next = NULL;
-    s->host_addr.s_addr = htonl (INADDR_ANY); /* NOT virtual host;
+    s->addrs = pcalloc(p, sizeof (server_addr_rec));
+    s->addrs->host_addr.s_addr = htonl (INADDR_ANY); /* NOT virtual host;
 					       * don't match any real network
 					       * interface.
 					       */
-    s->host_port = 0; /* matches any port */
+    s->addrs->host_port = 0; /* matches any port */
 
     s->module_config = create_server_config (p, s);
     s->lookup_defaults = create_default_per_dir_config (p);
-    
+
     return s;
 }
 
@@ -780,3 +1096,103 @@ server_rec *read_config(pool *p, pool *ptemp, char *confname)
     return s;
 }
 
+/********************************************************************
+ * Configuration directives are restricted in terms of where they may
+ * appear in the main configuration files and/or .htaccess files according
+ * to the bitmask req_override in the command_rec structure.
+ * If any of the overrides set in req_override are also allowed in the
+ * context in which the command is read, then the command is allowed.
+ * The context is determined as follows:
+ *
+ *    inside *.conf --> override = (RSRC_CONF|OR_ALL)&~(OR_AUTHCFG|OR_LIMIT);
+ *    within <Directory> or <Location> --> override = OR_ALL|ACCESS_CONF;
+ *    within .htaccess --> override = AllowOverride for current directory;
+ *
+ * the result is, well, a rather confusing set of possibilities for when
+ * a particular directive is allowed to be used.  This procedure prints
+ * in English where the given (pc) directive can be used.
+ */
+void show_overrides(command_rec *pc, module *pm)
+{
+    int n = 0;
+    
+    printf("\tAllowed in *.conf ");
+    if ((pc->req_override & (OR_OPTIONS|OR_FILEINFO|OR_INDEXES)) ||
+        ((pc->req_override & RSRC_CONF) &&
+         ((pc->req_override & (ACCESS_CONF|OR_AUTHCFG|OR_LIMIT)))))
+        printf("anywhere");
+    else if (pc->req_override & RSRC_CONF)
+        printf("only outside <Directory> or <Location>");
+    else 
+        printf("only inside <Directory> or <Location>");
+
+    /* Warn if the directive is allowed inside <Directory> or .htaccess
+     * but module doesn't support per-dir configuration */
+
+    if ((pc->req_override & (OR_ALL|ACCESS_CONF)) && !pm->create_dir_config)
+        printf(" [no per-dir config]");
+
+    if (pc->req_override & OR_ALL) {
+        printf(" and in .htaccess\n\twhen AllowOverride");
+
+        if ((pc->req_override & OR_ALL) == OR_ALL)
+            printf(" isn't None");
+        else {
+            printf(" includes ");
+
+            if (pc->req_override & OR_AUTHCFG) {
+                if (n++) printf(" or ");
+                printf("AuthConfig");
+            }
+            if (pc->req_override & OR_LIMIT) {
+                if (n++) printf(" or ");
+                printf("Limit");
+            }
+            if (pc->req_override & OR_OPTIONS) {
+                if (n++) printf(" or ");
+                printf("Options");
+            }
+            if (pc->req_override & OR_FILEINFO) {
+                if (n++) printf(" or ");
+                printf("FileInfo");
+            }
+            if (pc->req_override & OR_INDEXES) {
+                if (n++) printf(" or ");
+                printf("Indexes");
+            }
+        }
+    }
+    printf("\n");
+}
+
+/* Show the preloaded configuration directives, the help string explaining
+ * the directive arguments, in what module they are handled, and in
+ * what parts of the configuration they are allowed.  Used for httpd -h.
+ */
+void show_directives()
+{
+    extern module *preloaded_modules[];
+    command_rec *pc;
+    int n;
+    
+    for (n = 0; preloaded_modules[n]; ++n)
+        for (pc = preloaded_modules[n]->cmds; pc && pc->name; ++pc) {
+            printf("%s\n", pc->name);
+            if (pc->errmsg)
+                printf("\t%s\n", pc->errmsg);
+            printf("\t%s\n", preloaded_modules[n]->name);
+            show_overrides(pc, preloaded_modules[n]);
+        }
+}
+
+/* Show the preloaded module names.  Used for httpd -l. */
+void show_modules()
+{
+    extern module *preloaded_modules[];
+    int n;
+ 
+    printf ("Compiled-in modules:\n");
+    for (n = 0; preloaded_modules[n]; ++n)
+        printf ("  %s\n", preloaded_modules[n]->name);
+}
+
