@@ -1,6 +1,5 @@
-
 /* ====================================================================
- * Copyright (c) 1995 The Apache Group.  All rights reserved.
+ * Copyright (c) 1995-1997 The Apache Group.  All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
@@ -51,7 +50,6 @@
  *
  */
 
-
 /*
  * The central data structures around here...
  */
@@ -69,12 +67,17 @@ enum cmd_how {
 				 * (e.g., AddIcon)
 				 */
   FLAG,				/* One of 'On' or 'Off' */
-  NO_ARGS			/* No args at all, e.g. </Directory> */
+  NO_ARGS,			/* No args at all, e.g. </Directory> */
+  TAKE12,			/* one or two arguments */
+  TAKE3,			/* three arguments only */
+  TAKE23,			/* two or three arguments */
+  TAKE123,			/* one, two or three arguments */
+  TAKE13			/* one or three arguments */
 };
 
 typedef struct command_struct {
   char *name;			/* Name of this command */
-  char *(*func)();		/* Function invoked */
+  const char *(*func)();	/* Function invoked */
   void *cmd_data;		/* Extra data, for functions which
 				 * implement multiple commands...
 				 */
@@ -86,8 +89,22 @@ typedef struct command_struct {
   char *errmsg;			/* 'usage' message, in case of syntax errors */
 } command_rec;
 
-/* Values for the req_override slot */
-
+/* The allowed locations for a configuration directive are the union of
+ * those indicated by each set bit in the req_override mask.
+ *
+ * (req_override & RSRC_CONF)   => *.conf outside <Directory> or <Location>
+ * (req_override & ACCESS_CONF) => *.conf inside <Directory> or <Location>
+ * (req_override & OR_AUTHCFG)  => *.conf inside <Directory> or <Location>
+ *                                 and .htaccess when AllowOverride AuthConfig
+ * (req_override & OR_LIMIT)    => *.conf inside <Directory> or <Location>
+ *                                 and .htaccess when AllowOverride Limit
+ * (req_override & OR_OPTIONS)  => *.conf anywhere
+ *                                 and .htaccess when AllowOverride Options
+ * (req_override & OR_FILEINFO) => *.conf anywhere
+ *                                 and .htaccess when AllowOverride FileInfo
+ * (req_override & OR_INDEXES)  => *.conf anywhere
+ *                                 and .htaccess when AllowOverride Indexes
+ */
 #define OR_NONE 0
 #define OR_LIMIT 1
 #define OR_OPTIONS 2
@@ -95,10 +112,17 @@ typedef struct command_struct {
 #define OR_AUTHCFG 8
 #define OR_INDEXES 16
 #define OR_UNSET 32
-#define ACCESS_CONF 64		/* command valid in access.conf */
-#define RSRC_CONF 128		/* command valid in srm.conf */
+#define ACCESS_CONF 64
+#define RSRC_CONF 128
 #define OR_ALL (OR_LIMIT|OR_OPTIONS|OR_FILEINFO|OR_AUTHCFG|OR_INDEXES)
 
+/* This can be returned by a function if they don't wish to handle
+ * a command. Make it something not likely someone will actually use
+ * as an error code.
+ */
+
+#define DECLINE_CMD "\a\b"
+
 /*
  * This structure is passed to a command which is being invoked,
  * to carry a large variety of miscellaneous data which is all of
@@ -123,13 +147,14 @@ typedef struct {
     char *path;			/* If configuring for a directory,
 				 * pathname of that directory.
 				 */
+    const command_rec *cmd;	/* configuration command */
 } cmd_parms;
 
 /* This structure records the existence of handlers in a module... */
 
 typedef struct {
     char *content_type;
-    int (*handler) ();
+    int (*handler) (request_rec *);
 } handler_rec;
 
 /*
@@ -146,6 +171,9 @@ typedef struct module_struct {
     int module_index;		/* Index to this modules structures in
 				 * config vectors.
 				 */
+
+    const char *name;
+
     struct module_struct *next;
 
 #ifdef ULTRIX_BRAIN_DEATH
@@ -188,6 +216,7 @@ typedef struct module_struct {
     int (*type_checker)(request_rec *);
     int (*fixer_upper)(request_rec *);
     int (*logger)(request_rec *);
+    int (*header_parser)(request_rec *);
 } module;
 
 /* Initializer for the first few module slots, which are only
@@ -197,8 +226,8 @@ typedef struct module_struct {
  * handle it back-compatibly, or at least signal an error).
  */
 
-#define MODULE_MAGIC_NUMBER 19960526
-#define STANDARD_MODULE_STUFF MODULE_MAGIC_NUMBER, 0, NULL
+#define MODULE_MAGIC_NUMBER 19970526
+#define STANDARD_MODULE_STUFF MODULE_MAGIC_NUMBER, -1, __FILE__, NULL
 
 /* Generic accessors for other modules to get at their own module-specific
  * data
@@ -209,7 +238,8 @@ void set_module_config (void *conf_vector, module *m, void *val);
      
 /* Generic command handling function... */
 
-char *set_string_slot (cmd_parms *, char *, char *);
+const char *set_string_slot (cmd_parms *, char *, char *);
+const char *set_flag_slot (cmd_parms *, char *, int);
 
 /* For modules which need to read config files, open logs, etc. ...
  * this returns the fname argument if it begins with '/'; otherwise
@@ -221,6 +251,10 @@ char *server_root_relative (pool *p, char *fname);
 /* Finally, the hook for dynamically loading modules in... */
 
 void add_module (module *m);
+int add_named_module (const char *name);
+void clear_module_list ();
+const char *find_module_name (module *m);
+module *find_linked_module (const char *name);
 
 #ifdef CORE_PRIVATE
 
@@ -228,6 +262,8 @@ void add_module (module *m);
 
 server_rec *read_config (pool *conf_pool, pool *temp_pool, char *config_name);
 void setup_prelinked_modules();
+void show_directives();
+void show_modules();
 
 /* For http_request.c... */
 
@@ -239,9 +275,9 @@ void *merge_per_dir_configs (pool *p, void *base, void *new);
 
 int parse_htaccess(void **result, request_rec *r, int override,
 		   char *path, char *file);
-char *srm_command_loop (cmd_parms *parms, void *config);
+const char *srm_command_loop (cmd_parms *parms, void *config);
 
-server_rec *init_virtual_host (pool *p, char *hostname);
+server_rec *init_virtual_host (pool *p, const char *hostname, server_rec *main_server);
 int is_virtual_server (server_rec *);
 void process_resource_config(server_rec *s, char *fname, pool *p, pool *ptemp);
 
@@ -256,5 +292,6 @@ int find_types (request_rec *);	/* identify MIME type */
 int run_fixups (request_rec *);	/* poke around for other metainfo, etc.... */
 int invoke_handler (request_rec *);     
 int log_transaction (request_rec *r);
-     
+int header_parse (request_rec *);
+
 #endif
