@@ -1,6 +1,5 @@
-
 /* ====================================================================
- * Copyright (c) 1995 The Apache Group.  All rights reserved.
+ * Copyright (c) 1995-1997 The Apache Group.  All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
@@ -51,13 +50,12 @@
  *
  */
 
-
 /*
  * http_request.c: functions to get and process requests
  * 
  * Rob McCool 3/21/93
  *
- * Thoroughly revamped by rst for Shambhala.  NB this file reads
+ * Thoroughly revamped by rst for Apache.  NB this file reads
  * best from the bottom up.
  * 
  */
@@ -127,11 +125,11 @@ int check_symlinks (char *d, int opts)
 
     /* OK, it's a symlink.  May still be OK with OPT_SYM_OWNER */
     
-    if (!(opts & OPT_SYM_OWNER)) return FORBIDDEN;
+    if (!(opts & OPT_SYM_OWNER)) return HTTP_FORBIDDEN;
 	
-    if (stat (d, &fi) < 0) return FORBIDDEN;
+    if (stat (d, &fi) < 0) return HTTP_FORBIDDEN;
     
-    return (fi.st_uid == lfi.st_uid) ? OK : FORBIDDEN;
+    return (fi.st_uid == lfi.st_uid) ? OK : HTTP_FORBIDDEN;
 
 #endif    
 }
@@ -139,7 +137,7 @@ int check_symlinks (char *d, int opts)
 /* Dealing with the file system to get PATH_INFO
  */
 
-void get_path_info(request_rec *r)
+int get_path_info(request_rec *r)
 {
     char *cp;
     char *path = r->filename;
@@ -157,7 +155,10 @@ void get_path_info(request_rec *r)
 	/* See if the pathname ending here exists... */
       
 	*cp = '\0';
+
+	errno = 0;
 	rv = stat(path, &r->finfo);
+
 	if (cp != end) *cp = '/';
       
 	if (!rv) {
@@ -174,9 +175,29 @@ void get_path_info(request_rec *r)
 	
 	    r->path_info = pstrdup (r->pool, cp);
 	    *cp = '\0';
-	    return;
+	    return OK;
 	}
+#if defined(ENOENT) && defined(ENOTDIR)
+	else if (errno == ENOENT || errno == ENOTDIR) {
+#else
+#error ENOENT || ENOTDIR not defined -- check the comment below this line in the source for details
+	/*
+	 * If ENOENT || ENOTDIR is not defined in one of the your OS's
+	 * include files, Apache does not know how to check to see why
+	 * the stat() of the index file failed; there are cases where
+	 * it can fail even though the file exists.  This means
+	 * that it is possible for someone to get a directory
+	 * listing of a directory even though there is an index
+	 * (eg. index.html) file in it.  If you do not have a
+	 * problem with this, delete the above #error line and
+	 * start the compile again.  If you need to do this, please
+	 * submit a bug report from http://www.apache.org/bug_report.html
+	 * letting us know that you needed to do this.  Please be
+	 * sure to include the operating system you are using.  
+	 */
+
 	else {
+#endif
 	    last_cp = cp;
 	
 	    while (--cp > path && *cp != '/')
@@ -184,8 +205,22 @@ void get_path_info(request_rec *r)
 
 	    while (cp > path && cp[-1] == '/')
 		--cp;
+	} 
+#if defined(ENOENT) && defined(ENOTDIR)
+	else {
+#if defined(EACCES)
+	    if (errno != EACCES) 
+#endif 
+	    log_printf(r->server, 
+	       "access to %s failed for %s, reason: stat: %s (errno = %d)",
+	       r->uri, get_remote_host(r->connection, r->per_dir_config,
+	       REMOTE_NAME), strerror(errno), errno);
+
+	    return HTTP_FORBIDDEN;
 	}
+#endif /* ENOENT && ENOTDIR */
     }
+    return OK;
 }
 
 int directory_walk (request_rec *r)
@@ -224,7 +259,12 @@ int directory_walk (request_rec *r)
      * for the moment, that's not worth the trouble.
      */
 
+#ifdef __EMX__
+    /* Add OS/2 drive name support */
+    if ((test_filename[0] != '/') && (test_filename[1] != ':'))
+#else
     if (test_filename[0] != '/')
+#endif
     {
 /* fake filenames only match Directory sections */
         void *this_conf, *entry_config;
@@ -242,7 +282,11 @@ int directory_walk (request_rec *r)
 	    entry_dir = entry_core->d;
 
 	    this_conf = NULL;
-	    if (is_matchexp(entry_dir)) {
+	    if (entry_core->r) {
+		if (!regexec(entry_core->r, test_filename, 0, NULL, 0))
+		    this_conf = entry_config;
+	    }
+	    else if (entry_core->d_is_matchexp) {
 		if (!strcmp_match(test_filename, entry_dir))
 		    this_conf = entry_config;
 	    }
@@ -261,26 +305,30 @@ int directory_walk (request_rec *r)
 
     no2slash (test_filename);
     num_dirs = count_dirs(test_filename);
-    get_path_info (r);
+
+    res = get_path_info (r);
+    if (res != OK) {
+	return res;
+    }
     
+    if (test_filename[strlen(test_filename)-1] == '/')
+	--num_dirs;
+
     if (S_ISDIR (r->finfo.st_mode)) ++num_dirs;
 
     for (i = 1; i <= num_dirs; ++i) {
         core_dir_config *core_dir =
 	  (core_dir_config *)get_module_config(per_dir_defaults, &core_module);
-	int allowed_here = core_dir->opts;
-	int overrides_here = core_dir->override;
+	int overrides_here;
         void *this_conf = NULL, *htaccess_conf = NULL;
 	char *this_dir = make_dirstr (r->pool, test_filename, i);
-	char *config_name = make_full_path(r->pool, this_dir,
-					   sconf->access_name);
 	int j;
       
 	/* Do symlink checks first, because they are done with the
 	 * permissions appropriate to the *parent* directory...
 	 */
 	
-	if ((res = check_symlinks (this_dir, allowed_here)))
+	if ((res = check_symlinks (this_dir, core_dir->opts)))
 	{
 	    log_reason("Symbolic link not allowed", this_dir, r);
 	    return res;
@@ -301,31 +349,42 @@ int directory_walk (request_rec *r)
 	      (core_dir_config *)get_module_config(entry_config, &core_module);
 	    entry_dir = entry_core->d;
 	
-	    if (is_matchexp(entry_dir) && !strcmp_match(this_dir, entry_dir)) {
-		/* Don't try this wildcard again --- if it ends in '*'
-		 * it'll match again, and subdirectories won't be able to
-		 * override it...
-		 */
+	    if (entry_core->r) {
+		if (!regexec(entry_core->r, this_dir, 0, NULL,
+			     (j == num_sec) ? 0 : REG_NOTEOL)) {
+		    /* Don't try this wildcard again --- if it ends in '*'
+		     * it'll match again, and subdirectories won't be able to
+		     * override it...
+		     */
+		    sec[j] = NULL;
+		    this_conf = entry_config;
+		}
+	    }
+	    else if (entry_core->d_is_matchexp &&
+		     !strcmp_match(this_dir, entry_dir)) {
 		sec[j] = NULL;	
 	        this_conf = entry_config;
 	    }
 	    else if (!strcmp (this_dir, entry_dir))
 	        this_conf = entry_config;
+
+	    if (this_conf) {
+	        per_dir_defaults =
+	           merge_per_dir_configs (r->pool, per_dir_defaults, this_conf);
+	        core_dir =(core_dir_config *)get_module_config(per_dir_defaults,
+	           &core_module);
+	    }
+
 	}
 
-	if (this_conf)
-	{
-	    per_dir_defaults =
-	        merge_per_dir_configs (r->pool, per_dir_defaults, this_conf);
-	    core_dir =(core_dir_config *)get_module_config(per_dir_defaults,
-							   &core_module);
-	}
 	overrides_here = core_dir->override;
 
 	/* If .htaccess files are enabled, check for one.
 	 */
 	
 	if (overrides_here) {
+	    char *config_name = make_full_path(r->pool, this_dir,
+					   sconf->access_name);
 	    res = parse_htaccess (&htaccess_conf, r, overrides_here,
 				  this_dir, config_name);
 	    if (res) return res;
@@ -359,9 +418,15 @@ int location_walk (request_rec *r)
     void *per_dir_defaults = r->per_dir_config;
     
     core_dir_config **url = (core_dir_config **)url_array->elts;
-    int num_url = url_array->nelts;
+    int len, num_url = url_array->nelts;
     char *test_location = pstrdup (r->pool, r->uri);
 
+    /* Collapse multiple slashes, if it's a path URL (we don't want to
+     * do anything to <Location http://...> or such).
+     */
+    if (test_location[0] == '/')
+	no2slash (test_location);
+
     /* Go through the location entries, and check for matches. */
 
     if (num_url) {
@@ -383,12 +448,83 @@ int location_walk (request_rec *r)
 		get_module_config(entry_config, &core_module);
 	    entry_url = entry_core->d;
 
+	    len = strlen(entry_url);
+
 	    this_conf = NULL;
-	    if (is_matchexp(entry_url)) {
+
+	    if (entry_core->r) {
+		if (!regexec(entry_core->r, test_location, 0, NULL, 0))
+		    this_conf = entry_config;
+	    }
+	    else if( entry_core->d_is_matchexp ) {
 		if (!strcmp_match(test_location, entry_url))
 		    this_conf = entry_config;
 	    }
-	    else if (!strncmp (test_location, entry_url, strlen(entry_url)))
+	    else if (!strncmp (test_location, entry_url, len) &&
+		     (entry_url[len - 1] == '/' ||
+		      test_location[len] == '/' || test_location[len] == '\0'))
+	        this_conf = entry_config;
+
+	    if (this_conf)
+	        per_dir_defaults = merge_per_dir_configs (r->pool,
+					    per_dir_defaults, this_conf);
+	}
+
+	r->per_dir_config = per_dir_defaults;
+    }
+
+    return OK;
+}
+
+int file_walk (request_rec *r)
+{
+    core_dir_config *conf = get_module_config(r->per_dir_config, &core_module);
+    array_header *file_array = copy_array (r->pool, conf->sec);
+    void *per_dir_defaults = r->per_dir_config;
+    
+    core_dir_config **file = (core_dir_config **)file_array->elts;
+    int len, num_files = file_array->nelts;
+    char *test_file = pstrdup (r->pool, r->filename);
+
+    /* Collapse multiple slashes */
+    no2slash (test_file);
+
+    /* Go through the file entries, and check for matches. */
+
+    if (num_files) {
+        void *this_conf, *entry_config;
+	core_dir_config *entry_core;
+	char *entry_file;
+	int j;
+
+/* 
+ * we apply the directive sections in some order; should really try them
+ * with the most general first.
+ */
+	for (j = 0; j < num_files; ++j) {
+
+	    entry_config = file[j];
+	    if (!entry_config) continue;
+	    
+	    entry_core =(core_dir_config *)
+		get_module_config(entry_config, &core_module);
+	    entry_file = entry_core->d;
+
+	    len = strlen(entry_file);
+
+	    this_conf = NULL;
+
+	    if (entry_core->r) {
+		if (!regexec(entry_core->r, test_file, 0, NULL, 0))
+		    this_conf = entry_config;
+	    }
+	    else if ( entry_core->d_is_matchexp ) {
+		if (!strcmp_match(test_file, entry_file))
+		    this_conf = entry_config;
+	    }
+	    else if (!strncmp (test_file, entry_file, len) &&
+		     (entry_file[len - 1] == '/' ||
+		      test_file[len] == '/' || test_file[len] == '\0'))
 	        this_conf = entry_config;
 
 	    if (this_conf)
@@ -410,7 +546,7 @@ int location_walk (request_rec *r)
  * These do all access checks, etc., but don't actually run the transaction
  * ... use run_sub_req below for that.  Also, be sure to use destroy_sub_req
  * as appropriate if you're likely to be creating more than a few of these.
- * (An early Shambhala version didn't destroy the sub_reqs used in directory
+ * (An early Apache version didn't destroy the sub_reqs used in directory
  * indexing.  The result, when indexing a directory with 800-odd files in
  * it, was massively excessive storage allocation).
  *
@@ -418,7 +554,7 @@ int location_walk (request_rec *r)
  * structure...
  */
 
-request_rec *make_sub_request (request_rec *r)
+request_rec *make_sub_request (const request_rec *r)
 {
     pool *rrp = make_sub_pool (r->pool);
     request_rec *rr = pcalloc (rrp, sizeof (request_rec));
@@ -427,57 +563,20 @@ request_rec *make_sub_request (request_rec *r)
     return rr;
 }
 
-request_rec *sub_req_lookup_simple (char *new_file, request_rec *r)
-{
-    /* This handles the simple case, common to ..._lookup_uri and _file,
-     * of looking up another file in the same directory.
-     */
-    request_rec *rnew = make_sub_request (r);
-    pool *rnewp = rnew->pool;
-    int res;
-    
-    char *udir = make_dirstr(rnewp, r->uri, count_dirs(r->uri));
-    char *fdir = make_dirstr(rnewp, r->filename, count_dirs(r->filename));
 
-    *rnew = *r;			/* Copy per_dir config, etc. */
-    rnew->pool = rnewp;
-    rnew->uri = make_full_path (rnewp, udir, new_file);
-    rnew->filename = make_full_path (rnewp, fdir, new_file);
-    set_sub_req_protocol (rnew, r);
-	
-    rnew->finfo.st_mode = 0;
-    
-    if ((res = check_symlinks (rnew->filename, allow_options (rnew))))
-    {
-        rnew->status = res;
-    }
-
-    if (rnew->finfo.st_mode == 0 && stat (rnew->filename, &rnew->finfo) < 0)
-        rnew->finfo.st_mode = 0;
-
-    if ((rnew->status == 200) && (res = find_types (rnew)))
-        rnew->status = res;
-    
-    if ((rnew->status == 200) && (res = run_fixups (rnew)))
-        rnew->status = res;
-    
-    return rnew;
-}
-
-
-static int some_auth_required (request_rec *r);
-
-request_rec *sub_req_lookup_uri (char *new_file, request_rec *r)
+request_rec *sub_req_lookup_uri (const char *new_file, const request_rec *r)
 {
     request_rec *rnew;
     int res;
     char *udir;
     
     rnew = make_sub_request (r);
+    rnew->request_time = r->request_time;
     rnew->connection = r->connection; 
     rnew->server = r->server;
     rnew->request_config = create_request_config (rnew->pool);
     rnew->htaccess = r->htaccess; /* copy htaccess cache */
+    rnew->per_dir_config=r->server->lookup_defaults;
     set_sub_req_protocol (rnew, r);
 	
     if (new_file[0] == '/')
@@ -498,6 +597,11 @@ request_rec *sub_req_lookup_uri (char *new_file, request_rec *r)
 
     getparents (rnew->uri);
 	
+    if ((res = location_walk (rnew))) {
+	rnew->status=res;
+	return rnew;
+    }
+
     res = translate_name(rnew);
     if (res)
     {
@@ -511,11 +615,20 @@ request_rec *sub_req_lookup_uri (char *new_file, request_rec *r)
      * of translate_name.
      * Instead we rely on the cache of .htaccess results.
      */
+    /* NB: directory_walk() clears the per_dir_config, so we don't inherit from
+       location_walk() above */
     
     if ((res = directory_walk (rnew))
-	|| (!some_auth_required (rnew) ? 0 :
-	     ((res = check_user_id (rnew)) || (res = check_auth (rnew))))
-	|| (res = check_access (rnew))
+	|| (res = file_walk (rnew))
+	|| (res = location_walk (rnew))
+        || ((satisfies(rnew)==SATISFY_ALL || satisfies(rnew)==SATISFY_NOSPEC)?
+	    ((res = check_access (rnew))
+	     || (some_auth_required (rnew) &&
+		 ((res = check_user_id (rnew)) || (res = check_auth (rnew))))):
+	    ((res = check_access (rnew))
+	     && (!some_auth_required (rnew) ||
+		 ((res = check_user_id (rnew)) || (res = check_auth (rnew)))))
+	    )
 	|| (res = find_types (rnew))
 	|| (res = run_fixups (rnew))
 	)
@@ -526,38 +639,79 @@ request_rec *sub_req_lookup_uri (char *new_file, request_rec *r)
     return rnew;
 }
 
-request_rec *sub_req_lookup_file (char *new_file, request_rec *r)
+request_rec *sub_req_lookup_file (const char *new_file, const request_rec *r)
 {
     request_rec *rnew;
     int res;
     char *fdir;
-    
-    /* Check for a special case... if there are no '/' characters in new_file
-     * at all, then we are looking at a relative lookup in the same directory.
-     * That means we don't have to redo any access checks.
-     */
-
-    if (strchr (new_file, '/') == NULL) 
-        return sub_req_lookup_simple (new_file, r);
 
     rnew = make_sub_request (r);
-    fdir = make_dirstr (rnew->pool, r->filename, count_dirs (r->filename));
-    
+    rnew->request_time = r->request_time;
     rnew->connection = r->connection; /* For now... */
     rnew->server = r->server;
     rnew->request_config = create_request_config (rnew->pool);
     rnew->htaccess = r->htaccess; /* copy htaccess cache */
     set_sub_req_protocol (rnew, r);
-	
-    rnew->uri = "INTERNALLY GENERATED file-relative req";
-    rnew->filename = ((new_file[0] == '/') ?
-		      new_file :
-		      make_full_path (rnew->pool, fdir, new_file));
-	
-    if ((res = directory_walk (rnew))
-	|| (res = check_access (rnew))
-	|| (!some_auth_required (rnew) ? 0 :
-	     ((res = check_user_id (rnew)) && (res = check_auth (rnew))))
+    fdir = make_dirstr (rnew->pool, r->filename, count_dirs (r->filename));
+
+    /* Check for a special case... if there are no '/' characters in new_file
+     * at all, then we are looking at a relative lookup in the same directory.
+     * That means we won't have to redo directory_walk, and we may not
+     * even have to redo access checks.
+     */
+
+    if (strchr (new_file, '/') == NULL) {
+	char *udir = make_dirstr(rnew->pool, r->uri, count_dirs(r->uri));
+
+	rnew->uri = make_full_path (rnew->pool, udir, new_file);
+	rnew->filename = make_full_path (rnew->pool, fdir, new_file);
+	if (stat (rnew->filename, &rnew->finfo) < 0) {
+	    rnew->finfo.st_mode = 0;
+	}
+
+	rnew->per_dir_config = r->per_dir_config;
+
+	if ((res = check_symlinks (rnew->filename, allow_options (rnew)))) {
+	    log_reason ("Symbolic link not allowed", rnew->filename, rnew);
+	    rnew->status = res;
+	    return rnew;
+	}
+	/* do a file_walk, if it doesn't change the per_dir_config then
+	 * we know that we don't have to redo all the access checks */
+	if ((res = file_walk (rnew))) {
+	    rnew->status = res;
+	    return rnew;
+	}
+	if (rnew->per_dir_config == r->per_dir_config) {
+	    if ((res = find_types (rnew)) || (res = run_fixups (rnew))) {
+		rnew->status = res;
+	    }
+	    return rnew;
+	}
+    } else {
+	/* XXX: this should be set properly like it is in the same-dir case
+	 * but it's actually sometimes to impossible to do it... because the
+	 * file may not have a uri associated with it -djg */
+	rnew->uri = "INTERNALLY GENERATED file-relative req";
+	rnew->filename = ((new_file[0] == '/') ?
+			pstrdup(rnew->pool,new_file) :
+			make_full_path (rnew->pool, fdir, new_file));
+	rnew->per_dir_config = r->server->lookup_defaults;
+	res = directory_walk (rnew);
+	if (!res) {
+	    res = file_walk (rnew);
+	}
+    }
+
+    if (res
+        || ((satisfies(rnew)==SATISFY_ALL || satisfies(rnew)==SATISFY_NOSPEC)?
+	    ((res = check_access (rnew))
+	     || (some_auth_required (rnew) &&
+		 ((res = check_user_id (rnew)) || (res = check_auth (rnew))))):
+	    ((res = check_access (rnew))
+	     && (!some_auth_required (rnew) ||
+		 ((res = check_user_id (rnew)) || (res = check_auth (rnew)))))
+	    )
 	|| (res = find_types (rnew))
 	|| (res = run_fixups (rnew))
 	)
@@ -598,10 +752,10 @@ void die(int type, request_rec *r)
      * any attempt to handle the other thing "intelligently"...
      */
 
-    if (r->status != 200) {
+    if (r->status != HTTP_OK) {
         recursive_error = type;
 
-	while (r->prev && r->prev->status != 200)
+	while (r->prev && (r->prev->status != HTTP_OK))
 	  r = r->prev; /* Get back to original error */
 	
 	type = r->status;
@@ -609,7 +763,7 @@ void die(int type, request_rec *r)
     }
        
     r->status = type;
-    
+
     /* Two types of custom redirects --- plain text, and URLs.
      * Plain text has a leading '"', so the URL code, here, is triggered
      * on its absence
@@ -626,12 +780,13 @@ void die(int type, request_rec *r)
 	    r->status = REDIRECT;
 	    table_set (r->headers_out, "Location", custom_response);
 	} else if ( custom_response[0] == '/') {
-	    r->no_cache = 1;	/* Do NOT send USE_LOCAL_COPY for
-				 * error documents!
-				 */
+	    r->no_local_copy = 1; /* Do NOT send USE_LOCAL_COPY for
+				   * error documents!
+				   */
 	    /* This redirect needs to be a GET no matter what the original
 	     * method was.
 	     */
+	    table_set(r->subprocess_env, "REQUEST_METHOD", r->method);
 	    r->method = pstrdup(r->pool, "GET");
 	    r->method_number = M_GET;
 	    internal_redirect (custom_response, r);
@@ -662,7 +817,7 @@ static void decl_die (int status, char *phase, request_rec *r)
     else die (status, r);
 }
 
-static int some_auth_required (request_rec *r)
+int some_auth_required (request_rec *r)
 {
     /* Is there a require line configured for the type of *this* req? */
     
@@ -700,7 +855,8 @@ void process_request_internal (request_rec *r)
 	return;
     }
 
-    if (!r->hostname && (r->proto_num >= 1001)) {
+    if ((!r->hostname && (r->proto_num >= 1001)) ||
+	((r->proto_num == 1001) && !table_get(r->headers_in, "Host"))) {
         /* Client sent us a HTTP/1.1 or later request without telling
 	 * us the hostname, either with a full URL or a Host: header.
 	 * We therefore need to (as per the 1.1 spec) send an error
@@ -713,6 +869,15 @@ void process_request_internal (request_rec *r)
 
     if (!r->proxyreq)
     {
+        /* We don't want TRACE to run through the normal handler set,
+         * we handle it specially.
+         */
+        if (r->method_number == M_TRACE) {
+            send_http_trace(r);
+            finalize_request_protocol(r);
+            return;
+        }
+
 	access_status = unescape_url(r->uri);
 	if (access_status)
 	{
@@ -723,36 +888,72 @@ void process_request_internal (request_rec *r)
 	getparents(r->uri);	/* OK --- shrinking transformations... */
     }
 
+    if ((access_status = location_walk (r))) {
+        die (access_status, r);
+	return;
+    }
+
     if ((access_status = translate_name (r))) {
         decl_die (access_status, "translate", r);
 	return;
     }
-    
+
+    /* NB: directory_walk() clears the per_dir_config, so we don't inherit from
+       location_walk() above */
+
     if ((access_status = directory_walk (r))) {
         die (access_status, r);
 	return;
     }	
+
+    if ((access_status = file_walk (r))) {
+	die (access_status, r);
+	return;
+    }
     
     if ((access_status = location_walk (r))) {
         die (access_status, r);
 	return;
     }	
-    
-    if ((access_status = check_access (r)) != 0) {
-        decl_die (access_status, "check access", r);
+
+    if ((access_status = header_parse (r))) {
+        die (access_status, r);
 	return;
     }
     
-    if (some_auth_required (r)) {
-        if ((access_status = check_user_id (r)) != 0) {
-	    decl_die (access_status, "check user.  No user file?", r);
+    switch (satisfies(r)) {
+    case SATISFY_ALL: case SATISFY_NOSPEC:
+	if ((access_status = check_access (r)) != 0) {
+	    decl_die (access_status, "check access", r);
 	    return;
 	}
-
-	if ((access_status = check_auth (r)) != 0) {
-	    decl_die (access_status, "check access.  No groups file?", r);
-	    return;
+	if (some_auth_required (r)) {
+	    if ((access_status = check_user_id (r)) != 0) {
+		decl_die (access_status, "check user.  No user file?", r);
+		return;
+	    }
+	    if ((access_status = check_auth (r)) != 0) {
+		decl_die (access_status, "check access.  No groups file?", r);
+		return;
+	    }
 	}
+	break;
+    case SATISFY_ANY:
+	if ((access_status = check_access (r)) != 0) {
+	    if (!some_auth_required (r)) {
+		decl_die (access_status, "check access", r);
+		return;
+	    }
+	    if ((access_status = check_user_id (r)) != 0) {
+		decl_die (access_status, "check user.  No user file?", r);
+		return;
+	    }
+	    if ((access_status = check_auth (r)) != 0) {
+		decl_die (access_status, "check access.  No groups file?", r);
+		return;
+	    }
+	}
+	break;
     }
 
     if ((access_status = find_types (r)) != 0) {
@@ -765,8 +966,13 @@ void process_request_internal (request_rec *r)
 	return;
     }
 
-    if ((access_status = invoke_handler (r)) != 0)
+    if ((access_status = invoke_handler (r)) != 0) {
         die (access_status, r);
+	return;
+    }
+
+   /* Take care of little things that need to happen when we're done */
+   finalize_request_protocol (r);
 }
 
 void process_request (request_rec *r)
@@ -801,10 +1007,10 @@ table *rename_original_env (pool *p, table *t)
     return new;
 }
 
-request_rec *internal_internal_redirect (char *new_uri, request_rec *r)
+request_rec *internal_internal_redirect (const char *new_uri, request_rec *r)
 {
     request_rec *new = (request_rec *)pcalloc(r->pool, sizeof(request_rec));
-    char t[10];			/* Long enough... */
+    char t[256];		/* Long enough... */
   
     new->connection = r->connection;
     new->server = r->server;
@@ -824,13 +1030,20 @@ request_rec *internal_internal_redirect (char *new_uri, request_rec *r)
     
     /* Inherit the rest of the protocol info... */
 
+    new->the_request = r->the_request;
+
     new->method = r->method;
     new->method_number = r->method_number;
+    new->allowed = r->allowed;
     
     new->status = r->status;
     new->assbackwards = r->assbackwards;
     new->header_only = r->header_only;
     new->protocol = r->protocol;
+    new->proto_num = r->proto_num;
+    new->hostname = r->hostname;
+    new->hostlen = r->hostlen;
+    new->request_time = r->request_time;
     new->main = r->main;
 
     new->headers_in = r->headers_in;
@@ -843,14 +1056,15 @@ request_rec *internal_internal_redirect (char *new_uri, request_rec *r)
     new->no_cache = r->no_cache; /* If we've already made up our minds
 				  * about this, don't change 'em back!
 				  */
+    new->no_local_copy = r->no_local_copy;
 
-    sprintf (t, "%d", r->status);
+    ap_snprintf (t, sizeof(t), "%d", r->status);
     table_set (new->subprocess_env, "REDIRECT_STATUS", pstrdup (r->pool, t));
 
     return new;
 }
 
-void internal_redirect (char *new_uri, request_rec *r)
+void internal_redirect (const char *new_uri, request_rec *r)
 {
     request_rec *new = internal_internal_redirect(new_uri, r);
     process_request_internal (new);
@@ -861,7 +1075,7 @@ void internal_redirect (char *new_uri, request_rec *r)
  * an internal redirect.
  */
 
-void internal_redirect_handler (char *new_uri, request_rec *r)
+void internal_redirect_handler (const char *new_uri, request_rec *r)
 {
     request_rec *new = internal_internal_redirect(new_uri, r);
     if (r->handler)
