@@ -1,6 +1,5 @@
- 
 /* ====================================================================
- * Copyright (c) 1995 The Apache Group.  All rights reserved.
+ * Copyright (c) 1995-1997 The Apache Group.  All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
@@ -51,7 +50,6 @@
  *
  */
 
-
 /*
  * httpd.h: header for simple (ha! not anymore) http daemon
  */
@@ -67,12 +65,14 @@
 /* Define this to be the default server home dir. Anything later in this
  * file with a relative pathname will have this added.
  */
+#ifndef HTTPD_ROOT
 #ifdef __EMX__
 /* Set default for OS/2 file system */ 
 #define HTTPD_ROOT "/os2httpd"
 #else
 #define HTTPD_ROOT "/usr/local/etc/httpd"
 #endif
+#endif
 
 /* Root of server */
 #ifdef __EMX__
@@ -104,8 +104,12 @@
 /* --------- Default user name and group name running standalone ---------- */
 /* --- These may be specified as numbers by placing a # before a number --- */
 
+#ifndef DEFAULT_USER
 #define DEFAULT_USER "#-1"
+#endif
+#ifndef DEFAULT_GROUP
 #define DEFAULT_GROUP "#-1"
+#endif
 
 /* The name of the log files */
 #ifdef __EMX__
@@ -131,7 +135,7 @@
 
 /* Define this to be what type you'd like returned for files with unknown */
 /* suffixes */
-#define DEFAULT_TYPE "text/html"
+#define DEFAULT_TYPE "text/plain"
 
 /* Define this to be what your per-directory security files are called */
 #ifdef __EMX__
@@ -142,7 +146,9 @@
 #endif
 
 /* The name of the server config file */
+#ifndef SERVER_CONFIG_FILE
 #define SERVER_CONFIG_FILE "conf/httpd.conf"
+#endif
 
 /* The name of the document config file */
 #define RESOURCE_CONFIG_FILE "conf/srm.conf"
@@ -159,35 +165,41 @@
 #define DEFAULT_USER_DIR "public_html"
 
 /* The default path for CGI scripts if none is currently set */
+#ifndef DEFAULT_PATH
 #define DEFAULT_PATH "/bin:/usr/bin:/usr/ucb:/usr/bsd:/usr/local/bin"
+#endif
 
 /* The path to the Bourne shell, for parsed docs */
+#ifndef SHELL_PATH
 #ifdef __EMX__
 /* Set default for OS/2 file system */ 
 #define SHELL_PATH "CMD.EXE"
 #else
 #define SHELL_PATH "/bin/sh"
 #endif
+#endif
+
+/* The path to the suExec wrapper, can be overridden in Configuration */
+#ifndef SUEXEC_BIN
+#define SUEXEC_BIN "/usr/local/etc/httpd/sbin/suexec"
+#endif
 
 /* The default string lengths */
 #define MAX_STRING_LEN HUGE_STRING_LEN
 #define HUGE_STRING_LEN 8192
 
 /* The timeout for waiting for messages */
-#define DEFAULT_TIMEOUT 1200
+#define DEFAULT_TIMEOUT 300
 
 /* The timeout for waiting for keepalive timeout until next request */
 #define DEFAULT_KEEPALIVE_TIMEOUT 15
 
 /* The number of requests to entertain per connection */
-#define DEFAULT_KEEPALIVE 5
+#define DEFAULT_KEEPALIVE 100
 
 /* The size of the server's internal read-write buffers */
 #define IOBUFSIZE 8192
 
-/* The number of header lines we will accept from a client */
-#define MAX_HEADERS 200
-
 /* Number of servers to spawn off by default --- also, if fewer than
  * this free when the caretaker checks, it will spawn more.
  */
@@ -210,13 +222,13 @@
  * We keep a hard maximum number of servers, for two reasons --- first off,
  * in case something goes seriously wrong, we want to stop the fork bomb
  * short of actually crashing the machine we're running on by filling some
- * kernel table (I was originally going to make this 256, but it turns out
- * that that would actually fill the process table on reasonably configured
- * machines).  Secondly, it keeps the size of the scoreboard file small
+ * kernel table.  Secondly, it keeps the size of the scoreboard file small
  * enough that we can read the whole thing without worrying too much about
  * the overhead.
  */
-#define HARD_SERVER_LIMIT 150
+#ifndef HARD_SERVER_LIMIT
+#define HARD_SERVER_LIMIT 256
+#endif
 
 /* Number of requests to try to handle in a single process.  If <= 0,
  * the children don't die off.  That's the default here, since I'm still
@@ -225,49 +237,124 @@
 
 #define DEFAULT_MAX_REQUESTS_PER_CHILD 0
 
-/* If you have altered Apache and wish to change the SERVER_VERSION define
- * below, please keep to the HTTP/1.0 specification.  This states that
+/* If you have altered Apache and wish to change the SERVER_VERSION
+ * identifier below, please keep to the HTTP specification.  This states that
  * the identification string should consist of product tokens with an optional
  * slash and version designator.  Sub-products which form a significant part 
- * of the application can be listed, separated by whitespace.  The tokens
- * are listed in order of their significance for identifying the application.
+ * of the application can be listed, separated by whitespace, by adding
+ * their product tokens to EXTRA_CFLAGS in the Configuration file like so.
+ *
+ * EXTRA_CFLAGS="-DSERVER_SUBVERSION="MrWidget/0.1-alpha"
+ *
+ * The tokens are listed in order of their significance for identifying the
+ * application.
  *
  * "Product tokens should be short and to the point -- use of them for 
  * advertizing or other non-essential information is explicitly forbidden."
  *
- * Example: "Apache/1.1b3 MrWidget/0.1-alpha" 
+ * Example: "Apache/1.1.0 MrWidget/0.1-alpha" 
  */
 
-#define SERVER_VERSION "Apache/1.1.3" /* SEE COMMENTS ABOVE */
+#define SERVER_BASEVERSION "Apache/1.2.0" /* SEE COMMENTS ABOVE */
+#ifdef SERVER_SUBVERSION
+#define SERVER_VERSION	SERVER_BASEVERSION " " SERVER_SUBVERSION
+#else
+#define SERVER_VERSION	SERVER_BASEVERSION
+#endif
 
-#define SERVER_PROTOCOL "HTTP/1.0"
+#define SERVER_PROTOCOL "HTTP/1.1"
 #define SERVER_SUPPORT "http://www.apache.org/"
 
 #define DECLINED -1		/* Module declines to handle */
 #define OK 0			/* Module has handled this stage. */
 
-/* ------------------------------ error types ------------------------------ */
+/* ----------------------- HTTP Status Codes  ------------------------- */
 
-#define DOCUMENT_FOLLOWS 200
-#define REDIRECT 302
-#define USE_LOCAL_COPY 304
-#define BAD_REQUEST 400
-#define AUTH_REQUIRED 401
-#define FORBIDDEN 403
-#define NOT_FOUND 404
-#define SERVER_ERROR 500
-#define NOT_IMPLEMENTED 501
-#define BAD_GATEWAY 502
-#define SERVICE_UNAVAILABLE 503
-#define RESPONSE_CODES 10
+#define RESPONSE_CODES 38
 
-#define METHODS 6
+#define HTTP_CONTINUE                      100
+#define HTTP_SWITCHING_PROTOCOLS           101
+#define HTTP_OK                            200
+#define HTTP_CREATED                       201
+#define HTTP_ACCEPTED                      202
+#define HTTP_NON_AUTHORITATIVE             203
+#define HTTP_NO_CONTENT                    204
+#define HTTP_RESET_CONTENT                 205
+#define HTTP_PARTIAL_CONTENT               206
+#define HTTP_MULTIPLE_CHOICES              300
+#define HTTP_MOVED_PERMANENTLY             301
+#define HTTP_MOVED_TEMPORARILY             302
+#define HTTP_SEE_OTHER                     303
+#define HTTP_NOT_MODIFIED                  304
+#define HTTP_USE_PROXY                     305
+#define HTTP_BAD_REQUEST                   400
+#define HTTP_UNAUTHORIZED                  401
+#define HTTP_PAYMENT_REQUIRED              402
+#define HTTP_FORBIDDEN                     403
+#define HTTP_NOT_FOUND                     404
+#define HTTP_METHOD_NOT_ALLOWED            405
+#define HTTP_NOT_ACCEPTABLE                406
+#define HTTP_PROXY_AUTHENTICATION_REQUIRED 407
+#define HTTP_REQUEST_TIME_OUT              408
+#define HTTP_CONFLICT                      409
+#define HTTP_GONE                          410
+#define HTTP_LENGTH_REQUIRED               411
+#define HTTP_PRECONDITION_FAILED           412
+#define HTTP_REQUEST_ENTITY_TOO_LARGE      413
+#define HTTP_REQUEST_URI_TOO_LARGE         414
+#define HTTP_UNSUPPORTED_MEDIA_TYPE        415
+#define HTTP_INTERNAL_SERVER_ERROR         500
+#define HTTP_NOT_IMPLEMENTED               501
+#define HTTP_BAD_GATEWAY                   502
+#define HTTP_SERVICE_UNAVAILABLE           503
+#define HTTP_GATEWAY_TIME_OUT              504
+#define HTTP_VERSION_NOT_SUPPORTED         505
+#define HTTP_VARIANT_ALSO_VARIES           506
+
+#define DOCUMENT_FOLLOWS    HTTP_OK
+#define PARTIAL_CONTENT     HTTP_PARTIAL_CONTENT
+#define MULTIPLE_CHOICES    HTTP_MULTIPLE_CHOICES
+#define MOVED               HTTP_MOVED_PERMANENTLY
+#define REDIRECT            HTTP_MOVED_TEMPORARILY
+#define USE_LOCAL_COPY      HTTP_NOT_MODIFIED
+#define BAD_REQUEST         HTTP_BAD_REQUEST
+#define AUTH_REQUIRED       HTTP_UNAUTHORIZED
+#define FORBIDDEN           HTTP_FORBIDDEN
+#define NOT_FOUND           HTTP_NOT_FOUND
+#define METHOD_NOT_ALLOWED  HTTP_METHOD_NOT_ALLOWED
+#define NOT_ACCEPTABLE      HTTP_NOT_ACCEPTABLE
+#define LENGTH_REQUIRED     HTTP_LENGTH_REQUIRED
+#define PRECONDITION_FAILED HTTP_PRECONDITION_FAILED
+#define SERVER_ERROR        HTTP_INTERNAL_SERVER_ERROR
+#define NOT_IMPLEMENTED     HTTP_NOT_IMPLEMENTED
+#define BAD_GATEWAY         HTTP_BAD_GATEWAY
+#define VARIANT_ALSO_VARIES HTTP_VARIANT_ALSO_VARIES
+
+#define is_HTTP_INFO(x)         (((x) >= 100)&&((x) < 200))
+#define is_HTTP_SUCCESS(x)      (((x) >= 200)&&((x) < 300))
+#define is_HTTP_REDIRECT(x)     (((x) >= 300)&&((x) < 400))
+#define is_HTTP_ERROR(x)        (((x) >= 400)&&((x) < 600))
+#define is_HTTP_CLIENT_ERROR(x) (((x) >= 400)&&((x) < 500))
+#define is_HTTP_SERVER_ERROR(x) (((x) >= 500)&&((x) < 600))
+
+#define status_drops_connection(x) (((x) == HTTP_BAD_REQUEST)           || \
+                                    ((x) == HTTP_REQUEST_TIME_OUT)      || \
+                                    ((x) == HTTP_LENGTH_REQUIRED)       || \
+                                    ((x) == HTTP_REQUEST_ENTITY_TOO_LARGE) || \
+                                    ((x) == HTTP_REQUEST_URI_TOO_LARGE) || \
+                                    ((x) == HTTP_INTERNAL_SERVER_ERROR) || \
+                                    ((x) == HTTP_SERVICE_UNAVAILABLE))
+
+
+#define METHODS 8
 #define M_GET 0
 #define M_PUT 1
 #define M_POST 2
 #define M_DELETE 3
 #define M_CONNECT 4
-#define M_INVALID 5
+#define M_OPTIONS 5
+#define M_TRACE 6
+#define M_INVALID 7
 
 #define CGI_MAGIC_TYPE "application/x-httpd-cgi"
 #define INCLUDES_MAGIC_TYPE "text/x-server-parsed-html"
@@ -281,6 +368,17 @@
 #define LF 10
 #define CR 13
 
+/* Possible values for request_rec.read_body (set by handling module):
+ *    REQUEST_NO_BODY          Send 413 error if message has any body
+ *    REQUEST_CHUNKED_ERROR    Send 411 error if body without Content-Length
+ *    REQUEST_CHUNKED_DECHUNK  If chunked, remove the chunks for me.
+ *    REQUEST_CHUNKED_PASS     Pass the chunks to me without removal.
+ */
+#define REQUEST_NO_BODY          0
+#define REQUEST_CHUNKED_ERROR    1
+#define REQUEST_CHUNKED_DECHUNK  2
+#define REQUEST_CHUNKED_PASS     3
+
 /* Things which may vary per file-lookup WITHIN a request ---
  * e.g., state of MIME config.  Basically, the name of an object, info
  * about the object, and any other info we may ahve which may need to
@@ -342,6 +440,8 @@ struct request_rec {
   char *hostname;		/* Host, as set by full URI or Host: */
   int hostlen;			/* Length of http://host:port in full URI */
 
+  time_t request_time;		/* When the request started */
+
   char *status_line;		/* Status line, if set by script */
   int status;			/* In any case */
   
@@ -351,10 +451,24 @@ struct request_rec {
   
   char *method;			/* GET, HEAD, POST, etc. */
   int method_number;		/* M_GET, M_POST, etc. */
+  int allowed;			/* Allowed methods - for 405, OPTIONS, etc */
 
   int sent_bodyct;		/* byte count in stream is for body */
   long bytes_sent;		/* body byte count, for easy access */
-  
+
+  /* HTTP/1.1 connection-level features */
+
+  int chunked;			/* sending chunked transfer-coding */
+  int byterange;		/* number of byte ranges */
+  char *boundary;		/* multipart/byteranges boundary */
+  char *range;			/* The Range: header */
+  long clength;			/* The "real" content length */
+
+  long remaining;		/* bytes left to read */
+  long read_length;		/* bytes that have been read */
+  int read_body;   		/* how the request body should be read */
+  int read_chunked;		/* reading chunked transfer-coding */
+
   /* MIME header environments, in and out.  Also, an array containing
    * environment variables to be passed to subprocesses, so people can
    * write modules to add to that environment.
@@ -377,9 +491,11 @@ struct request_rec {
   char *handler;		/* What we *really* dispatch on           */
 
   char *content_encoding;
-  char *content_language;
+  char *content_language;	/* for back-compat. only -- do not use */
+  array_header *content_languages; /* array of (char*) */
   
   int no_cache;
+  int no_local_copy;
   
   /* What object is being requested (either directly, or via include
    * or content-negotiation mapping).
@@ -417,6 +533,7 @@ struct conn_rec {
   
   pool *pool;
   server_rec *server;
+  server_rec *base_server;      /* Physical vhost this conn come in on */
   
   /* Information about the connection itself */
 
@@ -434,7 +551,7 @@ struct conn_rec {
                                  * "" if it has and no address was found.
                                  * N.B. Only access this though
 				 * get_remote_host() */
-  char *remote_logname;		/* Only ever set if doing_rfc931
+  char *remote_logname;		/* Only ever set if doing rfc1413 lookups.
                                  * N.B. Only access this through
 				 * get_remote_logname() */
     char *user;			/* If an authentication check was made,
@@ -450,54 +567,73 @@ struct conn_rec {
 
 /* Per-vhost config... */
 
+/* The address 255.255.255.255, when used as a virtualhost address,
+ * will become the "default" server when the ip doesn't match other vhosts.
+ */
+#define DEFAULT_VHOST_ADDR 0xfffffffful
+
+typedef struct server_addr_rec server_addr_rec;
+struct server_addr_rec {
+    server_addr_rec *next;
+    struct in_addr host_addr;	/* The bound address, for this server */
+    unsigned short host_port;	/* The bound port, for this server */   
+    char *virthost;		/* The name given in <VirtualHost> */
+};
+
+
 struct server_rec {
 
-  server_rec *next;
+    server_rec *next;
   
-  /* Full locations of server config info */
+    /* Full locations of server config info */
   
-  char *srm_confname;
-  char *access_confname;
+    char *srm_confname;
+    char *access_confname;
   
-  /* Contact information */
+    /* Contact information */
   
-  char *server_admin;
-  char *server_hostname;
-  short port;                    /* for redirects, etc. */
+    char *server_admin;
+    char *server_hostname;
+    unsigned short port;           /* for redirects, etc. */
   
-  /* Log files --- note that transfer log is now in the modules... */
+    /* Log files --- note that transfer log is now in the modules... */
   
-  char *error_fname;
-  FILE *error_log;
+    char *error_fname;
+    FILE *error_log;
   
-  /* Module-specific configuration for server, and defaults... */
+    /* Module-specific configuration for server, and defaults... */
 
-  int is_virtual;               /* true if this is the virtual server */
-  void *module_config;		/* Config vector containing pointers to
+    int is_virtual;             /* true if this is the virtual server */
+    void *module_config;	/* Config vector containing pointers to
 				 * modules' per-server config structures.
 				 */
-  void *lookup_defaults;	/* MIME type info, etc., before we start
+    void *lookup_defaults;	/* MIME type info, etc., before we start
 				 * checking per-directory info.
 				 */
-  /* Transaction handling */
+    /* Transaction handling */
 
-  struct in_addr host_addr;	/* The bound address, for this server */
-  short host_port;              /* The bound port, for this server */
-  int timeout;			/* Timeout, in seconds, before we give up */
-  int keep_alive_timeout;	/* Seconds we'll wait for another request */
-  int keep_alive;		/* Maximum requests per connection */
+    server_addr_rec *addrs;
+    int timeout;		/* Timeout, in seconds, before we give up */
+    int keep_alive_timeout;	/* Seconds we'll wait for another request */
+    int keep_alive_max;		/* Maximum requests per connection */
+    int keep_alive;		/* Use persistent connections? */
+    int send_buffer_size;       /* size of TCP send buffer (in bytes) */
 
-  char *path;			/* Pathname for ServerPath */
-  int pathlen;			/* Length of path */
+    char *path;			/* Pathname for ServerPath */
+    int pathlen;		/* Length of path */
 
-  char *names;			/* Wildcarded names for HostAlias servers */
-  char *virthost;		/* The name given in <VirtualHost> */
+    char *names;		/* Wildcarded names for ServerAlias servers */
+
+    uid_t server_uid;		/* effective user id when calling exec wrapper */
+    gid_t server_gid;		/* effective group id when calling exec wrapper */
 };
 
 /* These are more like real hosts than virtual hosts */
 struct listen_rec {
     listen_rec *next;
     struct sockaddr_in local_addr; /* local IP address and port */
+    int fd;
+    int used;	/* Only used during restart */
 /* more stuff here, like which protocol is bound to the port */
 };
 
@@ -507,54 +643,71 @@ struct listen_rec {
 /* Time */
 extern const char month_snames[12][4];
 
-struct tm *get_gmtoff(long *tz);
+struct tm *get_gmtoff(int *tz);
 char *get_time();
-char *ht_time (pool *p, time_t t, char *fmt, int gmt);     
+char *ht_time (pool *p, time_t t, const char *fmt, int gmt);     
 char *gm_timestr_822(pool *p, time_t t);
      
-/* String handling */     
+/* String handling. The *_nc variants allow you to use non-const char **s as
+arguments (unfortunately C won't automatically convert a char ** to a const
+char **) */     
      
-char *getword(pool *p, char **line, char stop);
-char *getword_nulls (pool *p, char **line, char stop);
-char *getword_conf (pool *p, char **line);      
+char *getword(pool *p, const char **line, char stop);
+char *getword_nc(pool *p, char **line, char stop);
+char *getword_white(pool *p, const char **line);
+char *getword_white_nc(pool *p, char **line);
+char *getword_nulls (pool *p, const char **line, char stop);
+char *getword_nulls_nc (pool *p, char **line, char stop);
+char *getword_conf (pool *p, const char **line);      
+char *getword_conf_nc (pool *p, char **line);      
 
 char *get_token (pool *p, char **accept_line, int accept_white);
+int find_token (pool *p, const char *line, const char *tok);
+int find_last_token (pool *p, const char *line, const char *tok);
      
-int is_url(char *u);
+int is_url(const char *u);
 extern int unescape_url(char *url);
 void no2slash(char *name);
 void getparents(char *name);
 char *escape_path_segment(pool *p, const char *s);
 char *os_escape_path(pool *p,const char *path,int partial);
-char *escape_uri (pool *p, char *s);
+#define escape_uri(ppool,path) os_escape_path(ppool,path,1)
 extern char *escape_html(pool *p, const char *s);
-char *construct_url (pool *p, char *path, server_rec *s);     
-char *escape_shell_cmd (pool *p, char *s);
+char *construct_server(pool *p, const char *hostname, unsigned port);
+char *construct_url (pool *p, const char *path, const server_rec *s);     
+char *escape_shell_cmd (pool *p, const char *s);
      
-int count_dirs(char *path);
-char *make_dirstr(pool *a, char *s, int n);
-char *make_full_path(pool *a, char *dir, char *f);
+int count_dirs(const char *path);
+char *make_dirstr(pool *a, const char *s, int n);
+char *make_full_path(pool *a, const char *dir, const char *f);
      
-int is_matchexp(char *str);
-int strcmp_match(char *str, char *exp);
-int strcasecmp_match(char *str, char *exp);
-char *uudecode (pool *, char *);
+int is_matchexp(const char *str);
+int strcmp_match(const char *str, const char *exp);
+int strcasecmp_match(const char *str, const char *exp);
+char *uudecode (pool *, const char *);
+
+char *pregsub(pool *p, const char *input, const char *source,
+	      size_t nmatch, regmatch_t pmatch[]);
 
 void str_tolower (char *);
 int ind (const char *, char);	/* Sigh... */
 int rind (const char *, char);     
 
 int cfg_getline(char *s, int n, FILE *f);
-     
+
+#ifdef NEED_STRERROR
+char *strerror (int err);
+#endif
+
 /* Misc system hackery */
      
-uid_t uname2id(char *name);
-gid_t gname2id(char *name);
-int is_directory(char *name);
-int can_exec(struct stat *);     
-void chdir_file(char *file);
+uid_t uname2id(const char *name);
+gid_t gname2id(const char *name);
+int is_directory(const char *name);
+int can_exec(const struct stat *);     
+void chdir_file(const char *file);
      
 char *get_local_host(pool *);
-unsigned long get_virthost_addr (char *hostname, short int *port);
+unsigned long get_virthost_addr (const char *hostname, unsigned short *port);
 
 extern time_t restart_time;
