@@ -1,6 +1,5 @@
-
 /* ====================================================================
- * Copyright (c) 1995 The Apache Group.  All rights reserved.
+ * Copyright (c) 1995-1997 The Apache Group.  All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
@@ -51,7 +50,6 @@
  *
  */
 
-
 /*
  * Security options etc.
  * 
@@ -63,6 +61,7 @@
 #include "http_core.h"
 #include "http_config.h"
 #include "http_log.h"
+#include "http_request.h"
 
 typedef struct {
     char *from;
@@ -95,7 +94,7 @@ void *create_access_dir_config (pool *p, char *dummy)
     return (void *)conf;
 }
 
-char *order (cmd_parms *cmd, void *dv, char *arg)
+const char *order (cmd_parms *cmd, void *dv, char *arg)
 {
     access_dir_conf *d = (access_dir_conf *)dv;
     int i, order;
@@ -112,7 +111,7 @@ char *order (cmd_parms *cmd, void *dv, char *arg)
     return NULL;
 }
 
-char *allow_cmd (cmd_parms *cmd, void *dv, char *from, char *where)
+const char *allow_cmd (cmd_parms *cmd, void *dv, char *from, char *where)
 {
     access_dir_conf *d = (access_dir_conf *)dv;
     allowdeny *a;
@@ -143,7 +142,7 @@ int in_domain(const char *domain, const char *what) {
     int wl=strlen(what);
 
     if((wl-dl) >= 0) {
-        if (strcmp(domain,&what[wl-dl]) != 0) return 0;
+        if (strcasecmp(domain,&what[wl-dl]) != 0) return 0;
 
 	/* Make sure we matched an *entire* subdomain --- if the user
 	 * said 'allow from good.com', we don't want people from nogood.com
@@ -168,27 +167,56 @@ int in_ip(char *domain, char *what) {
     return (what[l] == '\0' || what[l] == '.');
 }
 
+static int is_ip(const char *host)
+{
+    while ((*host == '.') || isdigit(*host))
+        host++;
+    return (*host == '\0');
+}
+
 int find_allowdeny (request_rec *r, array_header *a, int method)
 {
     allowdeny *ap = (allowdeny *)a->elts;
     int mmask = (1 << method);
-    int i, gothost=0;
-    const char *remotehost=NULL;
+    int i;
+    int gothost = 0;
+    const char *remotehost = NULL;
 
     for (i = 0; i < a->nelts; ++i) {
         if (!(mmask & ap[i].limited))
 	    continue;
+
+	if (!strncmp(ap[i].from,"env=",4) && table_get(r->subprocess_env,ap[i].from+4))
+	    return 1;
+	    
+        if (ap[i].from && !strcmp(ap[i].from, "user-agents")) {
+	    char * this_agent = table_get(r->headers_in, "User-Agent");
+	    int j;
+  
+	    if (!this_agent) return 0;
+  
+	    for (j = i+1; j < a->nelts; ++j) {
+	        if (strstr(this_agent, ap[j].from)) return 1;
+	    }
+	    return 0;
+	}
+	
 	if (!strcmp (ap[i].from, "all"))
 	    return 1;
-	if (!gothost)
-	{
+
+	if (!gothost) {
 	    remotehost = get_remote_host(r->connection, r->per_dir_config,
-					 REMOTE_HOST);
-	    gothost = 1;
+	                                 REMOTE_HOST);
+
+	    if ((remotehost == NULL) || is_ip(remotehost))
+	        gothost = 1;
+	    else
+	        gothost = 2;
 	}
-        if (remotehost != NULL && isalpha(remotehost[0]))
-            if (in_domain(ap[i].from, remotehost))
-                return 1;
+
+        if ((gothost == 2) && in_domain(ap[i].from, remotehost))
+            return 1;
+
         if (in_ip (ap[i].from, r->connection->remote_ip))
             return 1;
     }
@@ -224,8 +252,11 @@ int check_dir_access (request_rec *r)
 	    ret = FORBIDDEN;
     }
 
-    if (ret == FORBIDDEN)
+    if (ret == FORBIDDEN && (
+        satisfies(r) != SATISFY_ANY || !some_auth_required(r)
+    )) {
 	log_reason ("Client denied by server configuration", r->filename, r);
+    }
 
     return ret;
 }
@@ -247,5 +278,6 @@ module access_module = {
    check_dir_access,		/* check access */
    NULL,			/* type_checker */
    NULL,			/* fixups */
-   NULL				/* logger */
+   NULL,			/* logger */
+   NULL				/* header parser */
 };
