@@ -1,6 +1,5 @@
-
 /* ====================================================================
- * Copyright (c) 1995 The Apache Group.  All rights reserved.
+ * Copyright (c) 1995-1997 The Apache Group.  All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
@@ -51,17 +50,16 @@
  *
  */
 
-
 /*
  * http_auth: authentication
  * 
  * Rob McCool & Brian Behlendorf.
  * 
- * Adapted to Shambhala by rst.
+ * Adapted to Apache by rst.
  *
  * Version 0.5 May 1996
  *
- * Brutally raped by Dirk.vanGulik@jrc.it to
+ * Modified by Dirk.vanGulik@jrc.it to
  * 
  * Adapted to allow anonymous logins, just like with Anon-FTP, when
  * one gives the magic user name 'anonymous' and ones email address
@@ -108,7 +106,7 @@ typedef struct  {
     int   auth_anon_logemail;
     int   auth_anon_verifyemail;
     int   auth_anon_mustemail;
-    int   auth_anon_authorative;
+    int   auth_anon_authoritative;
 
 } anon_auth_config_rec;
 
@@ -126,38 +124,38 @@ void *create_anon_auth_dir_config (pool *p, char *d)
     sec -> auth_anon_logemail		=1;
     sec -> auth_anon_verifyemail	=0;
     sec -> auth_anon_mustemail		=1;
-    sec -> auth_anon_authorative        =0;
+    sec -> auth_anon_authoritative        =0;
     return sec;
 }
 
-char *anon_set_passwd_flag (cmd_parms *cmd, 
+const char *anon_set_passwd_flag (cmd_parms *cmd, 
 	anon_auth_config_rec *sec, int arg) {
     sec->auth_anon_mustemail=arg;
     return NULL;
 }
 
-char *anon_set_userid_flag (cmd_parms *cmd, 
+const char *anon_set_userid_flag (cmd_parms *cmd, 
 	anon_auth_config_rec *sec, int arg) {
     sec->auth_anon_nouserid=arg;
     return NULL;
 }
-char *anon_set_logemail_flag (cmd_parms *cmd, 
+const char *anon_set_logemail_flag (cmd_parms *cmd, 
 	anon_auth_config_rec *sec, int arg) {
     sec->auth_anon_logemail=arg;
     return NULL;
 }
-char *anon_set_verifyemail_flag (cmd_parms *cmd, 
+const char *anon_set_verifyemail_flag (cmd_parms *cmd, 
 	anon_auth_config_rec *sec, int arg) {
     sec->auth_anon_verifyemail=arg;
     return NULL;
 }
-char *anon_set_authorative_flag (cmd_parms *cmd, 
+const char *anon_set_authoritative_flag (cmd_parms *cmd, 
 	anon_auth_config_rec *sec, int arg) {
-    sec->auth_anon_authorative=arg;
+    sec->auth_anon_authoritative=arg;
     return NULL;
 }
 
-char *anon_set_string_slots (cmd_parms *cmd, 
+const char *anon_set_string_slots (cmd_parms *cmd, 
 	anon_auth_config_rec *sec, char *arg) {
   
     auth_anon 	* first;
@@ -190,7 +188,7 @@ command_rec anon_auth_cmds[] = {
 	"Limited to 'on' or 'off'" },
 { "Anonymous_LogEmail", anon_set_logemail_flag, NULL, OR_AUTHCFG, FLAG, 
 	"Limited to 'on' or 'off'" },
-{ "Anonymous_Authorative", anon_set_authorative_flag, NULL, OR_AUTHCFG, FLAG, 
+{ "Anonymous_Authoritative", anon_set_authoritative_flag, NULL, OR_AUTHCFG, FLAG, 
 	"Limited to 'on' or 'off'" },
 
 { NULL }
@@ -236,19 +234,21 @@ int anon_authenticate_basic_user (request_rec *r)
 	( (!sec->auth_anon_mustemail) || strlen(send_pw)  ) &&
 	/* does the password look like an email address ? */
 	( (!sec->auth_anon_verifyemail) || 
-	     (strpbrk("@",send_pw) != NULL) || 
-	     (strpbrk(".",send_pw) != NULL) 
+	     ((strpbrk("@",send_pw) != NULL) 
+	      &&
+	      (strpbrk(".",send_pw) != NULL))
 	  ) 
 	) {
       if (sec->auth_anon_logemail) {
-	sprintf(errstr,"Anonymous: Passwd <%s> Accepted", 
+	ap_snprintf(errstr, sizeof(errstr), "Anonymous: Passwd <%s> Accepted", 
 			send_pw ? send_pw : "\'none\'");
 	log_error (errstr, r->server );
       }
       return OK;
     } else {
-        if (sec->auth_anon_authorative) {
-	sprintf(errstr,"Anonymous: Authorative, Passwd <%s> not accepted",
+        if (sec->auth_anon_authoritative) {
+	ap_snprintf(errstr, sizeof(errstr),
+		"Anonymous: Authoritative, Passwd <%s> not accepted",
 		send_pw ? send_pw : "\'none\'");
 	log_error(errstr,r->server);
 	return AUTH_REQUIRED;
@@ -294,5 +294,6 @@ module anon_auth_module = {
    NULL,			/* check access */
    NULL,			/* type_checker */
    NULL,			/* fixups */
-   NULL				/* logger */
+   NULL,			/* logger */
+   NULL				/* header parser */
 };
