@@ -1,5 +1,5 @@
 /* ====================================================================
- * Copyright (c) 1995 The Apache Group.  All rights reserved.
+ * Copyright (c) 1995-1997 The Apache Group.  All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
@@ -50,7 +50,6 @@
  *
  */
 
-
 /*
  * mod_auth_db: authentication
  * 
@@ -73,6 +72,12 @@
  * On some BSD systems (e.g. FreeBSD and NetBSD) dbm is automatically
  * mapped to Berkeley DB. You can use either mod_auth_dbm or
  * mod_auth_db. The latter makes it more obvious that it's Berkeley.
+ *
+ * dirkx - Added Authoritative control to allow passing on to lower  
+ *         modules if and only if the user-id is not known to this
+ *         module. A known user with a faulty or absent password still
+ *         causes an AuthRequired. The default is 'Authoritative', i.e.
+ *         no control is passed along.
  */
 
 #include "httpd.h"
@@ -86,12 +91,25 @@ typedef struct  {
 
     char *auth_dbpwfile;
     char *auth_dbgrpfile;
-
+    int   auth_dbauthoritative;
 } db_auth_config_rec;
 
 void *create_db_auth_dir_config (pool *p, char *d)
 {
-    return pcalloc (p, sizeof(db_auth_config_rec));
+    db_auth_config_rec *sec
+	= (db_auth_config_rec *)pcalloc (p, sizeof(db_auth_config_rec));
+    sec->auth_dbpwfile = NULL;
+    sec->auth_dbgrpfile = NULL;
+    sec->auth_dbauthoritative=1; /* fortress is secure by default */
+    return sec;
+}
+
+const char *set_db_slot (cmd_parms *cmd, void *offset, char *f, char *t)
+{
+    if (!t || strcmp(t, "db"))
+        return DECLINE_CMD;
+
+    return set_string_slot(cmd, offset, f);
 }
 
 command_rec db_auth_cmds[] = {
@@ -101,6 +119,16 @@ command_rec db_auth_cmds[] = {
 { "AuthDBGroupFile", set_string_slot,
     (void*)XtOffsetOf(db_auth_config_rec, auth_dbgrpfile),
     OR_AUTHCFG, TAKE1, NULL },
+{ "AuthUserFile", set_db_slot,
+    (void*)XtOffsetOf(db_auth_config_rec, auth_dbpwfile),
+    OR_AUTHCFG, TAKE12, NULL },
+{ "AuthGroupFile", set_db_slot,
+    (void*)XtOffsetOf(db_auth_config_rec, auth_dbgrpfile),
+    OR_AUTHCFG, TAKE12, NULL },
+{ "AuthDBAuthoritative", set_flag_slot,
+    (void*)XtOffsetOf(db_auth_config_rec, auth_dbauthoritative),
+    OR_AUTHCFG, FLAG, 
+    "Set to 'no' to allow access control to be passed along to lower modules if the userID is not known to this module" },
 { NULL }
 };
 
@@ -171,7 +199,9 @@ int db_authenticate_basic_user (request_rec *r)
         return DECLINED;
 	
     if(!(real_pw = get_db_pw(r, c->user, sec->auth_dbpwfile))) {
-        sprintf(errstr,"DB user %s not found", c->user);
+	if (!(sec -> auth_dbauthoritative))
+	    return DECLINED; 
+        ap_snprintf(errstr, sizeof(errstr), "DB user %s not found", c->user);
 	log_reason (errstr, r->filename, r);
 	note_basic_auth_failure (r);
 	return AUTH_REQUIRED;
@@ -181,7 +211,8 @@ int db_authenticate_basic_user (request_rec *r)
     if (colon_pw) *colon_pw='\0';   
     /* anyone know where the prototype for crypt is? */
     if(strcmp(real_pw,(char *)crypt(sent_pw,real_pw))) {
-        sprintf(errstr,"user %s: password mismatch",c->user);
+        ap_snprintf(errstr, sizeof(errstr), 
+		"user %s: password mismatch",c->user);
 	log_reason (errstr, r->uri, r);
 	note_basic_auth_failure (r);
 	return AUTH_REQUIRED;
@@ -203,7 +234,8 @@ int db_check_auth(request_rec *r) {
     require_line *reqs = reqs_arr ? (require_line *)reqs_arr->elts : NULL;
 
     register int x;
-    char *t, *w;
+    const char *t;
+    char *w;
 
     if (!sec->auth_dbgrpfile) return DECLINED;
     if (!reqs_arr) return DECLINED;
@@ -216,11 +248,15 @@ int db_check_auth(request_rec *r) {
         w = getword(r->pool, &t, ' ');
 	
         if(!strcmp(w,"group") && sec->auth_dbgrpfile) {
-           char *orig_groups,*groups,*v;
+	   const char *orig_groups,*groups;
+           char *v;
 
            if (!(groups = get_db_grp(r, user, sec->auth_dbgrpfile))) {
-               sprintf(errstr,"user %s not in DB group file %s",
-		       user, sec->auth_dbgrpfile);
+	       if (!(sec->auth_dbauthoritative))
+		 return DECLINED;
+               ap_snprintf(errstr, sizeof(errstr), 
+			"user %s not in DB group file %s",
+			user, sec->auth_dbgrpfile);
 	       log_reason (errstr, r->filename, r);
 	       note_basic_auth_failure (r);
 	       return AUTH_REQUIRED;
@@ -235,7 +271,8 @@ int db_check_auth(request_rec *r) {
                        return OK;
                }
            }
-           sprintf(errstr,"user %s not in right group",user);
+           ap_snprintf(errstr, sizeof(errstr), 
+		"user %s not in right group",user);
 	   log_reason (errstr, r->filename, r);
            note_basic_auth_failure(r);
 	   return AUTH_REQUIRED;
@@ -261,7 +298,6 @@ module db_auth_module = {
    NULL,			/* check access */
    NULL,			/* type_checker */
    NULL,			/* fixups */
-   NULL				/* logger */
+   NULL,			/* logger */
+   NULL				/* header parser */
 };
-
-
