@@ -1,5 +1,5 @@
 /* ====================================================================
- * Copyright (c) 1995 The Apache Group.  All rights reserved.
+ * Copyright (c) 1995-1997 The Apache Group.  All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
@@ -153,7 +153,7 @@ void *create_cern_meta_config (pool *p, server_rec *dummy)
     return new;
 }   
 
-char *set_metadir (cmd_parms *parms, void *dummy, char *arg)
+const char *set_metadir (cmd_parms *parms, void *dummy, char *arg)
 {       
     cern_meta_config *cmc ;
 
@@ -163,7 +163,7 @@ char *set_metadir (cmd_parms *parms, void *dummy, char *arg)
     return NULL;
 }
 
-char *set_metasuffix (cmd_parms *parms, void *dummy, char *arg)
+const char *set_metasuffix (cmd_parms *parms, void *dummy, char *arg)
 {       
     cern_meta_config *cmc ;
 
@@ -327,4 +327,5 @@ module cern_meta_module = {
    NULL,			/* type_checker */
    add_cern_meta_data,		/* fixups */
    NULL,			/* logger */
+   NULL				/* header parser */
 };
