@@ -1,6 +1,5 @@
-
 /* ====================================================================
- * Copyright (c) 1995 The Apache Group.  All rights reserved.
+ * Copyright (c) 1995-1997 The Apache Group.  All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
@@ -51,7 +50,6 @@
  *
  */
 
-
 /*
  * mod_digest: MD5 digest authentication
  * 
@@ -84,9 +82,17 @@ void *create_digest_dir_config (pool *p, char *d)
     return pcalloc (p, sizeof(digest_config_rec));
 }
 
+const char *set_digest_slot (cmd_parms *cmd, void *offset, char *f, char *t)
+{
+    if (t && strcmp(t, "standard"))
+	return pstrcat(cmd->pool, "Invalid auth file type: ",  t, NULL);
+
+    return set_string_slot(cmd, offset, f);
+}
+
 command_rec digest_cmds[] = {
-{ "AuthDigestFile", set_string_slot,
-    (void*)XtOffsetOf(digest_config_rec,pwfile), OR_AUTHCFG, TAKE1, NULL },
+{ "AuthDigestFile", set_digest_slot,
+  (void*)XtOffsetOf(digest_config_rec,pwfile), OR_AUTHCFG, TAKE12, NULL },
 { NULL }
 };
 
@@ -96,7 +102,8 @@ char *get_hash(request_rec *r, char *user, char *auth_pwfile)
 {
     FILE *f;
     char l[MAX_STRING_LEN];
-    char *rpw, *w, *x;
+    const char *rpw;
+    char *w, *x;
 
     if(!(f=pfopen(r->pool, auth_pwfile, "r"))) {
         log_reason ("Could not open password file", auth_pwfile, r);
@@ -120,7 +127,7 @@ char *get_hash(request_rec *r, char *user, char *auth_pwfile)
 /* Parse the Authorization header, if it exists */
 
 int get_digest_rec(request_rec *r, digest_header_rec *response) {
-  char *auth_line = table_get(r->headers_in, "Authorization");
+  const char *auth_line = table_get(r->headers_in, "Authorization");
   int l;
   int s = 0, vk = 0, vv = 0;
   char *t, *key, *value;
@@ -270,14 +277,14 @@ int authenticate_digest_user (request_rec *r)
         return DECLINED;
 	
     if (!(a1 = get_hash(r, c->user, sec->pwfile))) {
-        sprintf(errstr,"user %s not found",c->user);
+        ap_snprintf(errstr, sizeof(errstr), "user %s not found",c->user);
 	log_reason (errstr, r->uri, r);
 	note_digest_auth_failure (r);
 	return AUTH_REQUIRED;
     }
     /* anyone know where the prototype for crypt is? */
     if(strcmp(response->digest, find_digest(r, response, a1))) {
-        sprintf(errstr,"user %s: password mismatch",c->user);
+        ap_snprintf(errstr, sizeof(errstr), "user %s: password mismatch",c->user);
 	log_reason (errstr, r->uri, r);
 	note_digest_auth_failure (r);
 	return AUTH_REQUIRED;
@@ -292,7 +299,8 @@ int digest_check_auth (request_rec *r) {
     int m = r->method_number;
     int method_restricted = 0;    
     register int x;
-    char *t, *w;
+    const char *t;
+    char *w;
     array_header *reqs_arr;
     require_line *reqs;
 
@@ -350,7 +358,6 @@ module digest_module = {
    NULL,			/* check access */
    NULL,			/* type_checker */
    NULL,			/* fixups */
-   NULL				/* logger */
+   NULL,			/* logger */
+   NULL				/* header parser */
 };
-
-
