@@ -1,6 +1,5 @@
-
 /* ====================================================================
- * Copyright (c) 1995 The Apache Group.  All rights reserved.
+ * Copyright (c) 1995-1997 The Apache Group.  All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
@@ -51,14 +50,13 @@
  *
  */
 
-
 /*
  * http_dir.c: Handles the on-the-fly html index generation
  * 
  * Rob McCool
  * 3/23/93
  * 
- * Adapted to Shambhala by rst.
+ * Adapted to Apache by rst.
  */
 
 #include "httpd.h"
@@ -126,7 +124,7 @@ void push_item(array_header *arr, char *type, char *to, char *path, char *data)
         p->apply_to = NULL;
 }
 
-char *add_alt(cmd_parms *cmd, void *d, char *alt, char *to)
+const char *add_alt(cmd_parms *cmd, void *d, char *alt, char *to)
 {
     if (cmd->info == BY_PATH)
         if(!strcmp(to,"**DIRECTORY**"))
@@ -136,12 +134,12 @@ char *add_alt(cmd_parms *cmd, void *d, char *alt, char *to)
     return NULL;
 }
 
-char *add_icon(cmd_parms *cmd, void *d, char *icon, char *to)
+const char *add_icon(cmd_parms *cmd, void *d, char *icon, char *to)
 {
     char *iconbak = pstrdup (cmd->pool, icon);
 
     if(icon[0] == '(') {
-        char *alt = getword (cmd->pool, &iconbak, ',');
+        char *alt = getword_nc (cmd->pool, &iconbak, ',');
         iconbak[strlen(iconbak) - 1] = '\0'; /* Lose closing paren */
         add_alt(cmd, d, &alt[1], to);
     }
@@ -154,40 +152,40 @@ char *add_icon(cmd_parms *cmd, void *d, char *icon, char *to)
     return NULL;
 }
 
-char *add_desc(cmd_parms *cmd, void *d, char *desc, char *to)
+const char *add_desc(cmd_parms *cmd, void *d, char *desc, char *to)
 {
     push_item(((dir_config_rec *)d)->desc_list, cmd->info, to, cmd->path,desc);
     return NULL;
 }
 
-char *add_ignore(cmd_parms *cmd, void *d, char *ext) {
+const char *add_ignore(cmd_parms *cmd, void *d, char *ext) {
     push_item(((dir_config_rec *)d)->ign_list, 0, ext, cmd->path, NULL);
     return NULL;
 }
 
-char *add_header(cmd_parms *cmd, void *d, char *name) {
+const char *add_header(cmd_parms *cmd, void *d, char *name) {
     push_item(((dir_config_rec *)d)->hdr_list, 0, NULL, cmd->path, name);
     return NULL;
 }
 
-char *add_readme(cmd_parms *cmd, void *d, char *name) {
+const char *add_readme(cmd_parms *cmd, void *d, char *name) {
     push_item(((dir_config_rec *)d)->rdme_list, 0, NULL, cmd->path, name);
     return NULL;
 }
 
 
-char *add_opts_int(cmd_parms *cmd, void *d, int opts) {
+const char *add_opts_int(cmd_parms *cmd, void *d, int opts) {
     push_item(((dir_config_rec *)d)->opts_list, (char*)opts, NULL,
 	      cmd->path, NULL);
     return NULL;
 }
 
-char *fancy_indexing (cmd_parms *cmd, void *d, int arg)
+const char *fancy_indexing (cmd_parms *cmd, void *d, int arg)
 {
     return add_opts_int (cmd, d, arg? FANCY_INDEXING : 0);
 }
 
-char *add_opts(cmd_parms *cmd, void *d, char *optstr) {
+const char *add_opts(cmd_parms *cmd, void *d, const char *optstr) {
     char *w;
     int opts = 0;
 
@@ -691,21 +689,24 @@ int index_directory(request_rec *r, dir_config_rec *dir_conf)
     struct DIR_TYPE *dstruct;
     int num_ent=0,x;
     struct ent *head,*p;
-    struct ent **ar;
+    struct ent **ar = NULL;
     char *tmp;
     int dir_opts = find_opts(dir_conf, r);
 
-    if(!(d=opendir(name))) return FORBIDDEN;
+    if(!(d=opendir(name))) {
+        log_reason ("Can't open directory for index", r->filename, r);
+        return HTTP_FORBIDDEN;
+    }
 
     r->content_type = "text/html";
     
-    soft_timeout ("send directory", r);
     send_http_header(r);
 
     if (r->header_only) {
 	closedir (d);
 	return 0;
     }
+    hard_timeout("send directory", r);
 
     /* Spew HTML preamble */
     
@@ -714,8 +715,14 @@ int index_directory(request_rec *r, dir_config_rec *dir_conf)
     while (title_endp > title_name && *title_endp == '/')
 	*title_endp-- = '\0';
     
-    rvputs(r, "<HEAD><TITLE>Index of ", title_name, "</TITLE></HEAD><BODY>\n",
-	   NULL);
+    rvputs
+	(
+	    r,
+	    "<HTML><HEAD>\n<TITLE>Index of ",
+	    title_name,
+	    "</TITLE>\n</HEAD><BODY>\n",
+	    NULL
+	);
 
     if((!(tmp = find_header(dir_conf,r))) || (!(insert_readme(name,tmp,0,r))))
         rvputs(r, "<H1>Index of ", title_name, "</H1>\n", NULL);
@@ -732,32 +739,36 @@ int index_directory(request_rec *r, dir_config_rec *dir_conf)
             num_ent++;
         }
     }
-    ar=(struct ent **) palloc(r->pool, num_ent*sizeof(struct ent *));
-    p=head;
-    x=0;
-    while(p) {
-        ar[x++]=p;
-        p = p->next;
-    }
+    if (num_ent > 0) {
+        ar=(struct ent **) palloc(r->pool, num_ent*sizeof(struct ent *));
+        p=head;
+        x=0;
+        while(p) {
+            ar[x++]=p;
+            p = p->next;
+        }
     
-    qsort((void *)ar,num_ent,sizeof(struct ent *),
+        qsort((void *)ar,num_ent,sizeof(struct ent *),
 #ifdef ULTRIX_BRAIN_DEATH
-          (int (*))dsortf);
+	      (int (*))dsortf);
 #else
-          (int (*)(const void *,const void *))dsortf);
+	      (int (*)(const void *,const void *))dsortf);
 #endif
-     output_directories(ar, num_ent, dir_conf, r, dir_opts);
-     closedir(d);
+    }
+    output_directories(ar, num_ent, dir_conf, r, dir_opts);
+    closedir(d);
 
-     if (dir_opts & FANCY_INDEXING)
-         if((tmp = find_readme(dir_conf, r)))
-             insert_readme(name,tmp,1,r);
-     else {
-         rputs("</UL>", r);
-     }
+    if (dir_opts & FANCY_INDEXING)
+        if((tmp = find_readme(dir_conf, r)))
+            insert_readme(name,tmp,1,r);
+    else {
+        rputs("</UL>", r);
+    }
 
-     rputs("</BODY>", r);
-     return 0;
+    rputs ("</BODY></HTML>\n", r);
+
+    kill_timeout(r);
+    return 0;
 }
 
 /* The formal handler... */
@@ -766,7 +777,7 @@ int handle_dir (request_rec *r)
 {
     dir_config_rec *d =
       (dir_config_rec *)get_module_config (r->per_dir_config, &dir_module);
-    char *names_ptr = d->index_names ? d->index_names : DEFAULT_INDEX;
+    const char *names_ptr = d->index_names ? d->index_names : DEFAULT_INDEX;
     int allow_opts = allow_options (r);
     int error_notfound = 0;
 
@@ -780,8 +791,8 @@ int handle_dir (request_rec *r)
 			 "/", NULL);
 
 	table_set (r->headers_out, "Location",
-		   construct_url(r->pool, ifile, r->server));
-	return REDIRECT;
+		   construct_url(r->pool, ifile, r->server)); 
+	return HTTP_MOVED_PERMANENTLY;
     }
 
     /* KLUDGE --- make the sub_req lookups happen in the right directory.
@@ -796,7 +807,7 @@ int handle_dir (request_rec *r)
 	char *name_ptr = getword_conf (r->pool, &names_ptr);
 	request_rec *rr = sub_req_lookup_uri (name_ptr, r);
            
-	if (rr->status == 200 && rr->finfo.st_mode != 0) {
+	if (rr->status == HTTP_OK && rr->finfo.st_mode != 0) {
 	    char* new_uri = escape_uri(r->pool, rr->uri);
 
 	    if (rr->args != NULL)
@@ -809,32 +820,49 @@ int handle_dir (request_rec *r)
 	    return OK;
 	}
 
-       /* If the request returned something other than 404 (or 200),
-        * it means the module encountered some sort of problem. To be
-        * secure, we should return the error, rather than create
-        * along a (possibly unsafe) directory index.
-        *
-        * So we store the error, and if none of the listed files
-        * exist, we return the last error response we got, instead
-        * of a directory listing.
-        */
-       if (rr->status && rr->status != 404 && rr->status != 200)
-           error_notfound = rr->status;
+	/* If the request returned a redirect, propagate it to the client */
 
-       destroy_sub_req (rr);
+	if (is_HTTP_REDIRECT(rr->status) ||
+	    (rr->status == HTTP_NOT_ACCEPTABLE && *names_ptr == '\0')) {
+
+	    error_notfound = rr->status;
+	    r->notes = overlay_tables(r->pool, r->notes, rr->notes);
+	    r->headers_out = overlay_tables(r->pool, r->headers_out,
+	                                            rr->headers_out);
+	    r->err_headers_out = overlay_tables(r->pool, r->err_headers_out,
+	                                                rr->err_headers_out);
+	    destroy_sub_req(rr);
+	    return error_notfound;
+	}
+            
+	/* If the request returned something other than 404 (or 200),
+	 * it means the module encountered some sort of problem. To be
+	 * secure, we should return the error, rather than create
+	 * along a (possibly unsafe) directory index.
+	 *
+	 * So we store the error, and if none of the listed files
+	 * exist, we return the last error response we got, instead
+	 * of a directory listing.
+	 */
+	if (rr->status && rr->status != HTTP_NOT_FOUND && rr->status != HTTP_OK)
+	    error_notfound = rr->status;
+
+        destroy_sub_req (rr);
     }
 
     if (error_notfound)
-       return error_notfound;
- 
+	return error_notfound;
+
     if (r->method_number != M_GET) return NOT_IMPLEMENTED;
     
     /* OK, nothing easy.  Trot out the heavy artillery... */
 
     if (allow_opts & OPT_INDEXES) 
         return index_directory (r, d);
-    else
-        return FORBIDDEN;
+    else {
+        log_reason ("Directory index forbidden by rule", r->filename, r);
+        return HTTP_FORBIDDEN;
+    }
 }
 
 
@@ -858,5 +886,6 @@ module dir_module = {
    NULL,			/* check access */
    NULL,			/* type_checker */
    NULL,			/* fixups */
-   NULL				/* logger */
+   NULL,			/* logger */
+   NULL				/* header parser */
 };
