@@ -1,6 +1,5 @@
-
 /* ====================================================================
- * Copyright (c) 1995 The Apache Group.  All rights reserved.
+ * Copyright (c) 1995-1997 The Apache Group.  All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
@@ -51,7 +50,6 @@
  *
  */
 
-
 /*
  * A first stab at dynamic loading, using the GNU dld library
  * (or at least, an embarassingly old version of it...).
@@ -187,5 +185,6 @@ module dld_module = {
    NULL,			/* check auth */
    NULL,			/* check access */
    NULL,			/* type_checker */
-   NULL				/* logger */
+   NULL,			/* logger */
+   NULL				/* header parser */
 };
