@@ -1,6 +1,5 @@
-
 /* ====================================================================
- * Copyright (c) 1995 The Apache Group.  All rights reserved.
+ * Copyright (c) 1995-1997 The Apache Group.  All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
@@ -98,7 +97,7 @@
 
 #define IMAP_MAGIC_TYPE "application/x-httpd-imap"
 #define LARGEBUF 500
-#define SMALLBUF 100
+#define SMALLBUF 256
 #define MAXVERTS 100
 #define X 0
 #define Y 1
@@ -355,7 +354,9 @@ int read_quoted(char *string, char *quoted_part)
   return(string - starting_pos); /* return the total characters read */
 }
 
-
+/*
+ * url needs to point to a string with at least SMALLBUF memory allocated
+ */
 void imap_url(request_rec *r, char *base, char *value, char *url) 
 {
 /* translates a value into a URL. */
@@ -366,25 +367,28 @@ void imap_url(request_rec *r, char *base, char *value, char *url)
   char my_base[SMALLBUF] = {'\0'};
 
   if ( ! strcasecmp(value, "map" ) || ! strcasecmp(value, "menu") ) {
-    if (r->server->port == 80 ) { 
-      sprintf(url, "http://%s%s", r->server->server_hostname, r->uri);
+    if (r->server->port == DEFAULT_PORT ) { 
+      ap_snprintf(url, SMALLBUF,
+		"http://%s%s", r->server->server_hostname, r->uri);
     }
     else {
-      sprintf(url, "http://%s:%d%s", r->server->server_hostname,
+      ap_snprintf(url, SMALLBUF, "http://%s:%d%s", r->server->server_hostname,
 	      r->server->port, r->uri);      
     }
     return;  
   }
 
   if ( ! strcasecmp(value, "nocontent") || ! strcasecmp(value, "error") ) {
-    strcpy(url, value);
+    strncpy(url, value, SMALLBUF-1);
+    url[SMALLBUF-1] = '\0';
     return;    /* these are handled elsewhere, so just copy them */
   }
 
   if ( ! strcasecmp(value, "referer" ) ) {
     referer = table_get(r->headers_in, "Referer");
     if ( referer && *referer ) {
-      strcpy(url, referer);
+      strncpy(url, referer, SMALLBUF-1);
+      url[SMALLBUF-1] = '\0';
       return;
     }
     else {
@@ -396,27 +400,35 @@ void imap_url(request_rec *r, char *base, char *value, char *url)
   while ( isalpha(*string_pos) )
     string_pos++;    /* go along the URL from the map until a non-letter */
   if ( *string_pos == ':' ) { 
-    strcpy(url, value);        /* if letters and then a colon (like http:) */
+    strncpy(url, value, SMALLBUF-1);        /* if letters and then a colon (like http:) */
+    url[SMALLBUF-1] = '\0';
     return;                    /* it's an absolute URL, so use it! */
   }
 
   if ( ! base || ! *base ) {
     if ( value && *value ) {  
-      strcpy(url, value);   /* no base: use what is given */
+      strncpy(url, value, SMALLBUF-1);   /* no base: use what is given */
+      url[SMALLBUF-1] = '\0';
     }         
     else {                  
-      if (r->server->port == 80 ) {  
-	sprintf(url, "http://%s/", r->server->server_hostname);
+      if (r->server->port == DEFAULT_PORT ) {  
+	ap_snprintf(url, SMALLBUF, "http://%s/", r->server->server_hostname);
       }            
-      if (r->server->port != 80 ) {
-	sprintf(url, "http://%s:%d/", r->server->server_hostname, 
-		r->server->port);
+      if (r->server->port != DEFAULT_PORT ) {
+	ap_snprintf(url, SMALLBUF, "http://%s:%d/",
+		r->server->server_hostname, r->server->port);
       }                     /* no base, no value: pick a simple default */
     }
     return;  
   }
 
-  strcpy(my_base, base);  /* must be a relative URL to be combined with base */
+  strncpy(my_base, base, sizeof(my_base)-1);  /* must be a relative URL to be combined with base */
+  my_base[sizeof(my_base)-1] = '\0';
+  if (strchr(my_base, '/') == NULL && (!strncmp(value, "../", 3) || !strcmp(value, "..")) ) {
+    url[0] = '\0';
+    log_reason("invalid base directive in map file", r->uri, r);
+    return;
+  }
   string_pos = my_base; 
   while (*string_pos) {  
     if (*string_pos == '/' && *(string_pos+1) == '/') {
@@ -474,10 +486,10 @@ void imap_url(request_rec *r, char *base, char *value, char *url)
   }                   /* by this point, value does not start with '..' */
 
   if ( value && *value ) {
-    sprintf(url, "%s%s", my_base, value);   
+    ap_snprintf(url, SMALLBUF, "%s%s", my_base, value);   
   }
   else {
-    sprintf(url, "%s", my_base);   
+    ap_snprintf(url, SMALLBUF, "%s", my_base);   
   }
   return;
 }
@@ -488,10 +500,7 @@ int imap_reply(request_rec *r, char *redirect)
     return SERVER_ERROR;  /* they actually requested an error! */
   }
   if ( ! strcasecmp(redirect, "nocontent") ) {
-    r->status_line = pstrdup(r->pool, "204 No Content");
-    soft_timeout ("send no content", r);
-    send_http_header(r);
-    return OK;            /* tell the client to keep the page it has */
+    return HTTP_NO_CONTENT; /* tell the client to keep the page it has */
   }
   if (redirect && *redirect ) { 
     table_set(r->headers_out, "Location", redirect);
@@ -502,28 +511,17 @@ int imap_reply(request_rec *r, char *redirect)
 
 void menu_header(request_rec *r, char *menu)
 {
-  if (! strcasecmp(menu, "formatted")) {
-    r->content_type = "text/html";
-    soft_timeout ("send menu", r);
-    send_http_header(r);
-    rvputs(r, "<html>\n<head><title>Menu for ", r->uri,
-	   "</title></head>\n\n<body>\n", NULL);
+  r->content_type = "text/html";
+  send_http_header(r);
+  hard_timeout("send menu", r);   /* killed in menu_footer */
+
+  rvputs(r, "<html><head>\n<title>Menu for ", r->uri,
+	    "</title>\n</head><body>\n", NULL);
+
+  if (!strcasecmp(menu, "formatted")) {
     rvputs(r, "<h1>Menu for ", r->uri, "</h1>\n<hr>\n\n", NULL);
   } 
-  if (! strcasecmp(menu, "semiformatted")) {
-    r->content_type = "text/html";
-    soft_timeout ("send menu", r);
-    send_http_header(r);
-    rvputs(r, "<html>\n<head><title>Menu for ", r->uri,
-	   "</title></head>\n\n<body>\n", NULL);
-  } 
-  if (! strcasecmp(menu, "unformatted")) {
-    r->content_type = "text/html";
-    soft_timeout ("send menu", r);
-    send_http_header(r);
-    rvputs(r, "<html>\n<head><title>Menu for ", r->uri,
-	   "</title></head>\n\n<body>\n", NULL);
-  }
+
   return;
 }
 
@@ -596,11 +594,15 @@ void menu_directive(request_rec *r, char *menu, char *href, char *text)
 void menu_footer(request_rec *r)
 {
   rputs("\n\n</body>\n</html>\n", r);  /* finish the menu */
+  kill_timeout(r);
 }
 
 int imap_handler(request_rec *r)
 {
   char input[LARGEBUF] = {'\0'};
+	/* size of input can not be lowered without changing hard-coded
+	 * checks
+	 */
   char href_text[SMALLBUF] = {'\0'};
   char base[SMALLBUF] = {'\0'};
   char redirect[SMALLBUF] = {'\0'};
@@ -676,7 +678,7 @@ int imap_handler(request_rec *r)
     } /* blank lines and comments are ignored if we aren't printing a menu */
 
 
-    if (sscanf(input, "%s %s", directive, value) != 2) {
+    if (sscanf(input, "%255s %255s", directive, value) != 2) {
       continue;                           /* make sure we read two fields */
     }
     /* Now skip what we just read... we can't use ANSIism %n */
@@ -699,7 +701,8 @@ int imap_handler(request_rec *r)
       imap_url(r, NULL, value, mapdflt);
       if (showmenu) {              /* print the default if there's a menu */
 	if (! *href_text) {           /* if we didn't find a "href text" */
-	  strcpy(href_text, mapdflt); /* use the href itself as text */
+	  strncpy(href_text, mapdflt, sizeof(href_text)-1); /* use the href itself as text */
+	  href_text[sizeof(href_text)-1] = '\0';
 	}
 	imap_url(r, base, mapdflt, redirect); 
 	menu_default(r, imap_menu, redirect, href_text);
@@ -709,7 +712,7 @@ int imap_handler(request_rec *r)
 
     vertex = 0;
     while ( vertex < MAXVERTS &&  
-     sscanf(string_pos, "%lf,%lf",
+     sscanf(string_pos, "%lf, %lf",
      &pointarray[vertex][X], &pointarray[vertex][Y])   == 2)
     {
 	/* Now skip what we just read... we can't use ANSIism %n */
@@ -718,6 +721,8 @@ int imap_handler(request_rec *r)
 	while(isdigit(*string_pos))	/* and the 1st number */
 	    string_pos++;
 	string_pos++;			/* skip the ',' */
+	while(isspace(*string_pos))	/* past any more whitespace */
+	    string_pos++;
 	while(isdigit(*string_pos))	/* 2nd number */
 	    string_pos++;
 	vertex++;
@@ -730,7 +735,8 @@ int imap_handler(request_rec *r)
     if (showmenu) {
       read_quoted(string_pos, href_text); /* href text could be here instead */
       if (! *href_text) {           /* if we didn't find a "href text" */
-	strcpy(href_text, value);  /* use the href itself in the menu */
+	strncpy(href_text, value, sizeof(href_text)-1);  /* use the href itself in the menu */
+	href_text[sizeof(href_text)-1] = '\0';
       }
       imap_url(r, base, value, redirect); 
       menu_directive(r, imap_menu, redirect, href_text);
@@ -775,7 +781,8 @@ int imap_handler(request_rec *r)
     if ( ! strcasecmp(directive, "point" ) ) {         /* point */
       
       if (is_closer(testpoint, pointarray, &closest_yet) ) {
-	strcpy(closest, value);  /* if the closest point yet save it */
+	strncpy(closest, value, sizeof(closest)-1);  /* if the closest point yet save it */
+	closest[sizeof(closest)-1] = '\0';
       }
       
       continue;    
@@ -825,5 +832,6 @@ module imap_module = {
    NULL,			/* check access */
    NULL,			/* type_checker */
    NULL,			/* fixups */
-   NULL				/* logger */
+   NULL,			/* logger */
+   NULL				/* header parser */
 };
