@@ -1,5 +1,5 @@
 /* ====================================================================
- * Copyright (c) 1995, 1996 The Apache Group.  All rights reserved.
+ * Copyright (c) 1995-1997 The Apache Group.  All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
@@ -54,14 +54,14 @@
  * Info Module.  Display configuration information for the server and
  * all included modules.
  *
- * <Location /info>
+ * <Location /server-info>
  * SetHandler server-info
  * </Location>
  *
- * GET /info - Returns full configuration page for server and all modules
- * GET /info?server - Returns server configuration only
- * GET /info?module_name - Returns configuration for a single module
- * GET /info?list - Returns quick list of included modules
+ * GET /server-info - Returns full configuration page for server and all modules
+ * GET /server-info?server - Returns server configuration only
+ * GET /server-info?module_name - Returns configuration for a single module
+ * GET /server-info?list - Returns quick list of included modules
  *
  * Rasmus Lerdorf <rasmus@vex.net>, May 1996
  *
@@ -88,29 +88,50 @@ extern module *top_module;
 
 char *mod_info_html_cmd_string(char *string) {
 	char *s,*t;
-	static char ret[64];  /* What is the max size of a command? */
+	static char ret[256];  /* What is the max size of a command? */
+	char *end_ret;
 
 	ret[0]='\0';
 	s = string;
 	t=ret;	
-	while(*s) {
-		if(*s=='<') { strcat(t,"&lt;"); t+=4*sizeof(char); }
-		else if(*s=='>') { strcat(t,"&gt;"); t+=4*sizeof(char); }
-		else *t++=*s;
+	end_ret = t + sizeof(ret);
+	while((*s) && ((t-ret) < sizeof(ret))) {
+		if(*s=='<') { 
+			strncpy(t,"&lt;", end_ret - t);
+			t+=4;
+		} else if(*s=='>') {
+			strncpy(t,"&gt;", end_ret - t);
+			t+=4;
+		} else if(*s=='&') {
+		    	strncpy(t,"&amp;", end_ret - t);
+			t+=5;
+		} else {
+		    *t++=*s;
+		}
 		s++;
-		*t='\0';
 	}
+	*t='\0';
 	return(ret);
 }
 
-mod_info_config_lines *mod_info_load_config(pool *p, char *filename) {
+mod_info_config_lines *mod_info_load_config(pool *p, char *filename, request_rec *r) {
 	char s[MAX_STRING_LEN];
 	FILE *fp;
 	mod_info_config_lines *new, *ret=NULL, *prev=NULL;
-	char *t,*tt,o;
+	char *t,*tt,o, *msg;
 	
 	fp = pfopen(p,filename,"r");
-	if(!fp) return NULL;
+	if(!fp) {
+	    msg = pstrcat
+		    (
+			r->pool,
+			"mod_info: couldn't open config file ",
+			filename,
+			NULL
+		    );
+	    log_error (msg, r->server);
+	    return NULL;
+	}
 	while(!cfg_getline(s,MAX_STRING_LEN,fp)) {
 		if(*s=='#') continue; /* skip comments */
 		new = palloc(p,sizeof(struct mod_info_config_lines));
@@ -234,9 +255,7 @@ void mod_info_module_cmds(request_rec *r, mod_info_config_lines *cfg, command_re
 
 int display_info(request_rec *r) {
 	module *modp = NULL;
-	char buf[256];
-    extern char *module_names[];
-    char **names = module_names;
+	char buf[512], *cfname;
 	command_rec *cmd=NULL;
 	handler_rec *hand=NULL;
 	server_rec *serv = r->server;
@@ -258,27 +277,26 @@ int display_info(request_rec *r) {
 	extern char server_root[MAX_STRING_LEN];
 	extern char server_confname[MAX_STRING_LEN];
 
-	/* Init timeout */
-	soft_timeout ("send server info", r);
 	r->content_type = "text/html";		
 	send_http_header(r);
 	if(r->header_only) {
 		return 0;
     }
+	hard_timeout("send server info", r);
 	
 	rputs("<html><head><title>Server Information</title></head>\n",r);
 	rputs("<body><h1 align=center>Apache Server Information</h1>\n",r);
 	if(!r->args || strcasecmp(r->args,"list")) {
-		sprintf(buf,"%s/%s",server_root,server_confname);
-		mod_info_cfg_httpd = mod_info_load_config(r->pool,buf);
-		sprintf(buf,"%s/%s",server_root,serv->srm_confname);
-		mod_info_cfg_srm = mod_info_load_config(r->pool,buf);
-		sprintf(buf,"%s/%s",server_root,serv->access_confname);
-		mod_info_cfg_access = mod_info_load_config(r->pool,buf);
+		cfname = server_root_relative (r->pool, server_confname);
+		mod_info_cfg_httpd = mod_info_load_config (r->pool, cfname, r);
+		cfname = server_root_relative (r->pool, serv->srm_confname);
+		mod_info_cfg_srm = mod_info_load_config(r->pool, cfname, r);
+		cfname = server_root_relative (r->pool, serv->access_confname);
+		mod_info_cfg_access = mod_info_load_config (r->pool, cfname, r);
 		if(!r->args) {
 			rputs("<tt><a href=\"#server\">Server Settings</a>, ",r);
-			for(modp = top_module, names=module_names; modp; modp = modp->next, names++) {
-				sprintf(buf,"<a href=\"#%s\">%s</a>",*names,*names);
+			for(modp = top_module; modp; modp = modp->next) {
+				ap_snprintf(buf, sizeof(buf), "<a href=\"#%s\">%s</a>",modp->name,modp->name);
 				rputs(buf, r);
 				if(modp->next) rputs(", ",r);
 			}
@@ -286,42 +304,42 @@ int display_info(request_rec *r) {
 
 		}
 		if(!r->args || !strcasecmp(r->args,"server")) {	
-			sprintf(buf,"<a name=\"server\"><strong>Server Version:</strong> <font size=+1><tt>%s</tt></a></font><br>\n",SERVER_VERSION);
+			ap_snprintf(buf, sizeof(buf), "<a name=\"server\"><strong>Server Version:</strong> <font size=+1><tt>%s</tt></a></font><br>\n",SERVER_VERSION);
 			rputs(buf,r);
-			sprintf(buf,"<strong>API Version:</strong> <tt>%d</tt><br>\n",MODULE_MAGIC_NUMBER);
+			ap_snprintf(buf, sizeof(buf), "<strong>API Version:</strong> <tt>%d</tt><br>\n",MODULE_MAGIC_NUMBER);
 			rputs(buf,r);
-			sprintf(buf,"<strong>Run Mode:</strong> <tt>%s</tt><br>\n",standalone?"standalone":"inetd");
+			ap_snprintf(buf, sizeof(buf), "<strong>Run Mode:</strong> <tt>%s</tt><br>\n",standalone?"standalone":"inetd");
 			rputs(buf,r);
-			sprintf(buf,"<strong>User/Group:</strong> <tt>%s(%d)/%d</tt><br>\n",user_name,(int)user_id,(int)group_id);
+			ap_snprintf(buf, sizeof(buf), "<strong>User/Group:</strong> <tt>%s(%d)/%d</tt><br>\n",user_name,(int)user_id,(int)group_id);
 			rputs(buf,r);
-			sprintf(buf,"<strong>Hostname/port:</strong> <tt>%s:%d</tt><br>\n",serv->server_hostname,serv->port);
+			ap_snprintf(buf, sizeof(buf), "<strong>Hostname/port:</strong> <tt>%s:%u</tt><br>\n",serv->server_hostname,serv->port);
 			rputs(buf,r);
-			sprintf(buf,"<strong>Daemons:</strong> <tt>start: %d &nbsp;&nbsp; min idle: %d &nbsp;&nbsp; max idle: %d &nbsp;&nbsp; max: %d</tt><br>\n",daemons_to_start,daemons_min_free,daemons_max_free,daemons_limit);
+			ap_snprintf(buf, sizeof(buf), "<strong>Daemons:</strong> <tt>start: %d &nbsp;&nbsp; min idle: %d &nbsp;&nbsp; max idle: %d &nbsp;&nbsp; max: %d</tt><br>\n",daemons_to_start,daemons_min_free,daemons_max_free,daemons_limit);
 			rputs(buf,r);
-			sprintf(buf,"<strong>Max Requests:</strong> <tt>per child: %d &nbsp;&nbsp; per connection: %d</tt><br>\n",max_requests_per_child,serv->keep_alive);
+			ap_snprintf(buf, sizeof(buf), "<strong>Max Requests:</strong> <tt>per child: %d &nbsp;&nbsp; keep alive: %s &nbsp;&nbsp; max per connection: %d</tt><br>\n",max_requests_per_child,serv->keep_alive ? "on":"off", serv->keep_alive_max);
 			rputs(buf,r);
-			sprintf(buf,"<strong>Timeouts:</strong> <tt>connection: %d &nbsp;&nbsp; keep-alive: %d</tt><br>",serv->timeout,serv->keep_alive_timeout);
+			ap_snprintf(buf, sizeof(buf), "<strong>Timeouts:</strong> <tt>connection: %d &nbsp;&nbsp; keep-alive: %d</tt><br>",serv->timeout,serv->keep_alive_timeout);
 			rputs(buf,r);
-			sprintf(buf,"<strong>Server Root:</strong> <tt>%s</tt><br>\n",server_root);
+			ap_snprintf(buf, sizeof(buf), "<strong>Server Root:</strong> <tt>%s</tt><br>\n",server_root);
 			rputs(buf,r);
-			sprintf(buf,"<strong>Config File:</strong> <tt>%s</tt><br>\n",server_confname);
+			ap_snprintf(buf, sizeof(buf), "<strong>Config File:</strong> <tt>%s</tt><br>\n",server_confname);
 			rputs(buf,r);
-			sprintf(buf,"<strong>PID File:</strong> <tt>%s</tt><br>\n",pid_fname);
+			ap_snprintf(buf, sizeof(buf), "<strong>PID File:</strong> <tt>%s</tt><br>\n",pid_fname);
 			rputs(buf,r);
-			sprintf(buf,"<strong>Scoreboard File:</strong> <tt>%s</tt><br>\n",scoreboard_fname);
+			ap_snprintf(buf, sizeof(buf), "<strong>Scoreboard File:</strong> <tt>%s</tt><br>\n",scoreboard_fname);
 			rputs(buf,r);
 		}
 		rputs("<hr><dl>",r);
-		for(modp = top_module, names=module_names; modp; modp = modp->next, names++) {
-			if(!r->args || !strcasecmp(*names,r->args)) {	
-				sprintf(buf,"<dt><a name=\"%s\"><strong>Module Name:</strong> <font size=+1><tt>%s</tt></a></font>\n",*names,*names);
+		for(modp = top_module; modp; modp = modp->next) {
+			if(!r->args || !strcasecmp(modp->name,r->args)) {	
+				ap_snprintf(buf, sizeof(buf), "<dt><a name=\"%s\"><strong>Module Name:</strong> <font size=+1><tt>%s</tt></a></font>\n",modp->name,modp->name);
 				rputs(buf,r);
 				rputs("<dt><strong>Content-types affected:</strong>",r);	
 				hand = modp->handlers;
 				if(hand) {
 					while(hand) {
 						if(hand->content_type) {
-							sprintf(buf," <tt>%s</tt>\n",hand->content_type);	
+							ap_snprintf(buf, sizeof(buf), " <tt>%s</tt>\n",hand->content_type);	
 							rputs(buf,r);
 						} else break;
 						hand++;
@@ -372,7 +390,7 @@ int display_info(request_rec *r) {
 				if(cmd) {
 					while(cmd) {
 						if(cmd->name) {
-							sprintf(buf,"<dd><tt>%s - <i>",mod_info_html_cmd_string(cmd->name));	
+							ap_snprintf(buf, sizeof(buf), "<dd><tt>%s - <i>",mod_info_html_cmd_string(cmd->name));	
 							rputs(buf,r);
 							if(cmd->errmsg) rputs(cmd->errmsg,r);
 							rputs("</i></tt>\n",r);
@@ -392,13 +410,14 @@ int display_info(request_rec *r) {
 		}
 		if(!modp && r->args && strcasecmp(r->args,"server")) rputs("<b>No such module</b>\n",r);
 	} else {
-		for(modp = top_module; modp; modp = modp->next, names++) {
-			rputs(*names,r);
+		for(modp = top_module; modp; modp = modp->next) {
+			rputs(modp->name,r);
 			if(modp->next) rputs("<br>",r);
 		}	
 	}	
 	rputs("</dl></body></html>\n",r);
 	/* Done, turn off timeout, close file and return */
+	kill_timeout(r);
 	return 0;
 }
 
@@ -422,5 +441,6 @@ module info_module = {
 	NULL,				/* check access */
 	NULL,				/* type_checker */
 	NULL,				/* fixups */
-	NULL				/* logger */
+	NULL,				/* logger */
+	NULL				/* header parser */
 };
