@@ -1,6 +1,5 @@
-
 /* ====================================================================
- * Copyright (c) 1995 The Apache Group.  All rights reserved.
+ * Copyright (c) 1995-1997 The Apache Group.  All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
@@ -51,25 +50,86 @@
  *
  */
 
-
 /*
  * This is module implements the TransferLog directive (same as the
- * common log module), and an additional directive, LogFormat.
+ * common log module), and additional directives, LogFormat and CustomLog.
  *
- * The argument to LogFormat is a string, which can include literal
- * characters copied into the log files, and '%' directives as follows:
  *
+ * Syntax:
+ *
+ *    TransferLog fn      Logs transfers to fn in standard log format, unless
+ *                        a custom format is set with LogFormat
+ *    LogFormat format    Set a log format from TransferLog files
+ *    CustomLog fn format
+ *                        Log to file fn with format given by the format
+ *                        argument
+ *
+ *    CookieLog fn        For backwards compatability with old Cookie
+ *                        logging module - now deprecated.
+ *
+ * There can be any number of TransferLog and CustomLog
+ * commands. Each request will be logged to _ALL_ the
+ * named files, in the appropriate format.
+ *
+ * If no TransferLog or CustomLog directive appears in a VirtualHost,
+ * the request will be logged to the log file(s) defined outside
+ * the virtual host section. If a TransferLog or CustomLog directive
+ * appears in the VirtualHost section, the log files defined outside
+ * the VirtualHost will _not_ be used. This makes this module compatable
+ * with the CLF and config log modules, where the use of TransferLog
+ * inside the VirtualHost section overrides its use outside.
+ * 
+ * Examples:
+ *
+ *    TransferLog    logs/access_log
+ *    <VirtualHost>
+ *    LogFormat      "... custom format ..."
+ *    TransferLog    log/virtual_only
+ *    CustomLog      log/virtual_useragents "%t %{user-agent}i"
+ *    </VirtualHost>
+ *
+ * This will log using CLF to access_log any requests handled by the
+ * main server, while any requests to the virtual host will be logged
+ * with the "... custom format..." to virtual_only _AND_ using
+ * the custom user-agent log to virtual_useragents.
+ *
+ * Note that the NCSA referer and user-agent logs are easily added with
+ * CustomLog:
+ *   CustomLog   logs/referer  "%{referer}i -> %U"
+ *   CustomLog   logs/agent    "%{user-agent}i"
+ *
+ * Except: no RefererIgnore functionality
+ *         logs '-' if no Referer or User-Agent instead of nothing
+ *
+ * But using this method allows much easier modification of the
+ * log format, e.g. to log hosts along with UA:
+ *   CustomLog   logs/referer "%{referer}i %U %h"
+ *
+ * The argument to LogFormat and CustomLog is a string, which can include
+ * literal characters copied into the log files, and '%' directives as
+ * follows:
+ *
+ * %...b:  bytes sent, excluding HTTP headers.
+ * %...{FOOBAR}e:  The contents of the environment variable FOOBAR
+ * %...f:  filename
  * %...h:  remote host
+ * %...{Foobar}i:  The contents of Foobar: header line(s) in the request
+ *                 sent to the client.
  * %...l:  remote logname (from identd, if supplied)
- * %...u:  remote user (from auth; may be bogus if return status (%s) is 401)
- * %...t:  time, in common log format time format
+ * %...{Foobar}n:  The contents of note "Foobar" from another module.
+ * %...{Foobar}o:  The contents of Foobar: header line(s) in the reply.
+ * %...p:  the port the request was served to
+ * %...P:  the process ID of the child that serviced the request.
  * %...r:  first line of request
  * %...s:  status.  For requests that got internally redirected, this
  *         is status of the *original* request --- %...>s for the last.
- * %...b:  bytes sent.
- * %...{Foobar}i:  The contents of Foobar: header line(s) in the request
- *                 sent to the client.
- * %...{Foobar}o:  The contents of Foobar: header line(s) in the reply.
+ * %...t:  time, in common log format time format
+ * %...{format}t:  The time, in the form given by format, which should
+ *                 be in strftime(3) format.
+ * %...T:  the time taken to serve the request, in seconds.
+ * %...u:  remote user (from auth; may be bogus if return status (%s) is 401)
+ * %...U:  the URL path requested.
+ * %...v:  the name of the server (i.e. which virtual host?)
  *
  * The '...' can be nothing at all (e.g. "%h %u %r %s %b"), or it can
  * indicate conditions for inclusion of the item (which will cause it
@@ -94,25 +154,9 @@
  * server.  If it doesn't have its own TransferLog, it writes to the
  * same descriptor (meaning the same process for "| ...").
  *
- * That means that you can do things like:
- *
- * <VirtualHost hosta.com>
- * LogFormat "hosta ..."
- * ...
- * </VirtualHost>
- *
- * <VirtualHost hosta.com>
- * LogFormat "hostb ..."
- * ...
- * </VirtualHost>
- *
- * ... to have different virtual servers write into the same log file,
- * but have some indication which host they came from, though a %v
- * directive may well be a better way to handle this.
- *
  * --- rst */
 
-#define DEFAULT_LOG_FORMAT "%h %l %u %t \"%r\" %s %b"
+#define DEFAULT_LOG_FORMAT "%h %l %u %t \"%r\" %>s %b"
 
 #include "httpd.h"
 #include "http_config.h"
@@ -128,6 +172,36 @@ static mode_t xfer_mode = ( S_IREAD | S_IWRITE );
 static mode_t xfer_mode = ( S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH );
 #endif
 
+/*
+ * multi_log_state is our per-(virtual)-server configuration. We store
+ * an array of the logs we are going to use, each of type config_log_state.
+ * If a default log format is given by LogFormat, store in default_format
+ * (backward compat. with mod_log_config). We also store a pointer to
+ * the logs specified for the main server for virtual servers, so that
+ * if this vhost has now logs defined, we can use the main server's
+ * logs instead.
+ *
+ * So, for the main server, config_logs contains a list of the log files
+ * and server_config_logs in empty. For a vhost, server_config_logs
+ * points to the same array as config_logs in the main server, and
+ * config_logs points to the array of logs defined inside this vhost,
+ * which might be empty.
+ */
+
+typedef struct {
+  array_header *default_format;
+  array_header *config_logs;    
+  array_header *server_config_logs;
+} multi_log_state;
+
+/*
+ * config_log_state holds the status of a single log file. fname cannot
+ * be NULL. format might be NULL, in which case the default_format from
+ * the multi_log_state should be used, or if that is NULL as well, use
+ * the CLF. log_fd is -1 before the log file is opened and set to a valid
+ * fd after it is opened.
+ */
+
 typedef struct {
     char *fname;
     array_header *format;
@@ -151,7 +225,7 @@ typedef struct {
 char *format_integer(pool *p, int i)
 {
     char dummy[40];
-    sprintf (dummy, "%d", i);
+    ap_snprintf (dummy, sizeof(dummy), "%d", i);
     return pstrdup (p, dummy);
 }
 
@@ -169,12 +243,24 @@ char *log_remote_host (request_rec *r, char *a)
 char *log_remote_logname(request_rec *r, char *a)
 {return (char *)get_remote_logname(r);}
 
-char *log_remote_user (request_rec *r, char *a)
-{ return r->connection->user; }
+char *log_remote_user (request_rec *r, char *a) {
+    char *rvalue = r->connection->user;
+
+    if (rvalue == NULL) {
+        rvalue = "-";
+    } else if (strlen (rvalue) == 0) {
+        rvalue = "\"\"";
+    }
+    return rvalue;
+}
 
 char *log_request_line (request_rec *r, char *a)
 { return r->the_request; }
 
+char *log_request_file (request_rec *r, char *a)
+{ return r->filename; }
+char *log_request_uri (request_rec *r, char *a)
+{ return r->uri; }
 char *log_status (request_rec *r, char *a)
 { return pfmt(r->pool, r->status); }
 
@@ -186,7 +272,7 @@ char *log_bytes_sent (request_rec *r, char *a)
 	long int bs;
 	char dummy[40];
 	bgetopt(r->connection->client, BO_BYTECT, &bs);
-	sprintf(dummy, "%ld", bs);
+	ap_snprintf(dummy, sizeof(dummy), "%ld", bs);
 	return pstrdup(r->pool, dummy);
     }
 }
@@ -197,32 +283,63 @@ char *log_header_in (request_rec *r, char *a)
 char *log_header_out (request_rec *r, char *a)
 {
     char *cp = table_get (r->headers_out, a);
+    if (!strcasecmp(a, "Content-type") && r->content_type)
+	cp = r->content_type;
     if (cp) return cp;
     return table_get (r->err_headers_out, a);
 }
 
+char *log_note (request_rec *r, char *a)
+{ return table_get (r->notes, a); }
 char *log_env_var (request_rec *r, char *a)
 { return table_get (r->subprocess_env, a); }
 
 char *log_request_time (request_rec *r, char *a)
 {
-    long timz;
+    int timz;
     struct tm *t;
-    char tstr[MAX_STRING_LEN],sign;
+    char tstr[MAX_STRING_LEN];
     
     t = get_gmtoff(&timz);
-    sign = (timz < 0 ? '-' : '+');
-    if(timz < 0) 
-        timz = -timz;
 
-    strftime(tstr,MAX_STRING_LEN,"[%d/%b/%Y:%H:%M:%S ",t);
+    if (a && *a) /* Custom format */
+	strftime(tstr, MAX_STRING_LEN, a, t);
+    else { /* CLF format */
+	char sign = (timz < 0 ? '-' : '+');
 
-    sprintf (tstr + strlen(tstr), "%c%02ld%02ld]",
-	     sign, timz/3600, timz%3600);
+	if(timz < 0) timz = -timz;
+
+	strftime(tstr,MAX_STRING_LEN,"[%d/%b/%Y:%H:%M:%S ",t);
+	ap_snprintf (tstr + strlen(tstr), sizeof(tstr)-strlen(tstr), 
+		"%c%.2d%.2d]", sign, timz/60, timz%60);
+    }
 
     return pstrdup (r->pool, tstr);
 }
 
+char *log_request_duration (request_rec *r, char *a) {
+    char duration[22];	/* Long enough for 2^64 */
+
+    ap_snprintf(duration, sizeof(duration), "%ld", time(NULL) - r->request_time);
+    return pstrdup(r->pool, duration);
+}
+
+char *log_virtual_host (request_rec *r, char *a) {
+    return pstrdup(r->pool, r->server->server_hostname);
+}
+
+char *log_server_port (request_rec *r, char *a) {
+    char portnum[22];
+
+    ap_snprintf(portnum, sizeof(portnum), "%u", r->server->port);
+    return pstrdup(r->pool, portnum);
+}
+
+char *log_child_pid (request_rec *r, char *a) {
+    char pidnum[22];
+    ap_snprintf(pidnum, sizeof(pidnum), "%ld", (long)getpid());
+    return pstrdup(r->pool, pidnum);
+}
 /*****************************************************************
  *
  * Parsing the log format string
@@ -237,12 +354,19 @@ struct log_item_list {
     { 'l', log_remote_logname, 0 },
     { 'u', log_remote_user, 0 },
     { 't', log_request_time, 0 },
+    { 'T', log_request_duration, 1 },
     { 'r', log_request_line, 1 },
+    { 'f', log_request_file, 0 },
+    { 'U', log_request_uri, 1 },
     { 's', log_status, 1 },
     { 'b', log_bytes_sent, 0 },
     { 'i', log_header_in, 0 },
     { 'o', log_header_out, 0 },
+    { 'n', log_note, 0 },
     { 'e', log_env_var, 0 },
+    { 'v', log_virtual_host, 0 },
+    { 'p', log_server_port, 0 },
+    { 'P', log_child_pid, 0 },
     { '\0' }
 };
 
@@ -257,7 +381,7 @@ struct log_item_list  *find_log_func (char k)
     return NULL;
 }
 
-char *log_format_substring (pool *p, char *start, char *end)
+char *log_format_substring (pool *p, const char *start, const char *end)
 {
     char *res = palloc (p, end - start + 1);
     strncpy (res, start, end - start);
@@ -265,9 +389,9 @@ char *log_format_substring (pool *p, char *start, char *end)
     return res;
 }
 
-char *parse_log_misc_string (pool *p, log_format_item *it, char **sa)
+char *parse_log_misc_string (pool *p, log_format_item *it, const char **sa)
 {
-    char *s = *sa;
+    const char *s = *sa;
     
     it->func = constant_item;
     it->conditions = NULL;
@@ -279,9 +403,9 @@ char *parse_log_misc_string (pool *p, log_format_item *it, char **sa)
     return NULL;
 }
 
-char *parse_log_item (pool *p, log_format_item *it, char **sa)
+char *parse_log_item (pool *p, log_format_item *it, const char **sa)
 {
-    char *s = *sa;
+    const char *s = *sa;
     if (*s != '%') return parse_log_misc_string (p, it, sa);
 
     ++s;
@@ -346,7 +470,7 @@ char *parse_log_item (pool *p, log_format_item *it, char **sa)
     return "Ran off end of LogFormat parsing args to some directive";
 }
 
-array_header *parse_log_string (pool *p, char *s, char **err)
+array_header *parse_log_string (pool *p, const char *s, const char **err)
 {
     array_header *a = make_array (p, 30, sizeof (log_format_item));
     char *res;
@@ -398,35 +522,38 @@ char *process_item(request_rec *r, request_rec *orig, log_format_item *item)
     return cp ? cp : "-";
 }
 
-int config_log_transaction(request_rec *r)
-{
-    config_log_state *cls = get_module_config (r->server->module_config,
-					       &config_log_module);
-  
-    array_header *strsa= make_array(r->pool, cls->format->nelts,sizeof(char*));
-    log_format_item *items = (log_format_item *)cls->format->elts;
+int config_log_transaction(request_rec *r, config_log_state *cls,
+			   array_header *default_format) {
+    array_header *strsa;
+    log_format_item *items;
     char *str, **strs, *s;
     request_rec *orig;
     int i;
     int len = 0;
+    array_header *format;
+
+    format = cls->format ? cls->format : default_format;
+
+    strsa= make_array(r->pool, format->nelts,sizeof(char*));
+    items = (log_format_item *)format->elts;
 
     orig = r;
     while (orig->prev) orig = orig->prev;
     while (r->next) r = r->next;
 
-    for (i = 0; i < cls->format->nelts; ++i)
-	*((char**)push_array (strsa)) = process_item (r, orig, &items[i]);
+    for (i = 0; i < format->nelts; ++i)
+        *((char**)push_array (strsa)) = process_item (r, orig, &items[i]);
 
     strs = (char **)strsa->elts;
     
-    for (i = 0; i < cls->format->nelts; ++i)
-	len += strlen (strs[i]);
+    for (i = 0; i < format->nelts; ++i)
+        len += strlen (strs[i]);
 
     str = palloc (r->pool, len + 1);
 
-    for (i = 0, s = str; i < cls->format->nelts; ++i) {
-	strcpy (s, strs[i]);
-	s += strlen (strs[i]);
+    for (i = 0, s = str; i < format->nelts; ++i) {
+        strcpy (s, strs[i]);
+        s += strlen (strs[i]);
     }
     
     write(cls->log_fd, str, strlen(str));
@@ -434,6 +561,33 @@ int config_log_transaction(request_rec *r)
     return OK;
 }
 
+int multi_log_transaction(request_rec *r)
+{
+    multi_log_state *mls = get_module_config (r->server->module_config,
+                                               &config_log_module);
+    config_log_state *clsarray;
+    int i;
+
+    if (mls->config_logs->nelts) {
+        clsarray = (config_log_state *)mls->config_logs->elts;
+        for (i = 0; i < mls->config_logs->nelts; ++i) {
+            config_log_state *cls = &clsarray[i];
+        
+            config_log_transaction(r, cls, mls->default_format);
+        }
+    }
+    else if (mls->server_config_logs) {
+        clsarray = (config_log_state *)mls->server_config_logs->elts;
+        for (i = 0; i < mls->server_config_logs->nelts; ++i) {
+            config_log_state *cls = &clsarray[i];
+        
+            config_log_transaction(r, cls, mls->default_format);
+        }
+    }
+
+    return OK;
+}
+
 /*****************************************************************
  *
  * Module glue...
@@ -441,40 +595,81 @@ int config_log_transaction(request_rec *r)
 
 void *make_config_log_state (pool *p, server_rec *s)
 {
-    config_log_state *cls =
-      (config_log_state *)palloc (p, sizeof (config_log_state));
-
-    cls->fname = NULL;
-    cls->format = NULL;
-    cls->log_fd = -1;
-
-    return (void *)cls;
+    multi_log_state *mls =
+	(multi_log_state *)palloc(p, sizeof (multi_log_state));
+    
+    mls->config_logs = 
+	make_array(p, 5, sizeof (config_log_state));
+    mls->default_format = NULL;
+    mls->server_config_logs = NULL;
+    
+    return mls;
 }
 
-char *set_config_log (cmd_parms *parms, void *dummy, char *arg)
+/*
+ * Use the merger to simply add a pointer from the vhost log state
+ * to the log of logs specified for the non-vhost configuration
+ */
+
+void *merge_config_log_state (pool *p, void *basev, void *addv)
 {
-    config_log_state *cls = get_module_config (parms->server->module_config,
-					       &config_log_module);
-  
-    cls->fname = arg;
-    return NULL;
+    multi_log_state *base = (multi_log_state *)basev;
+    multi_log_state *add = (multi_log_state *)addv;
+    
+    add->server_config_logs = base->config_logs;
+    if (!add->default_format)
+        add->default_format = base->default_format;
+    
+    return add;
 }
 
-char *log_format (cmd_parms *cmd, void *dummy, char *arg)
+const char *log_format (cmd_parms *cmd, void *dummy, char *arg)
 {
-    char *err_string = NULL;
-    config_log_state *cls = get_module_config (cmd->server->module_config,
+    const char *err_string = NULL;
+    multi_log_state *mls = get_module_config (cmd->server->module_config,
 					       &config_log_module);
   
-    cls->format = parse_log_string (cmd->pool, arg, &err_string);
+    mls->default_format = parse_log_string (cmd->pool, arg, &err_string);
     return err_string;
 }
 
+const char *add_custom_log(cmd_parms *cmd, void *dummy, char *fn, char *fmt)
+{
+    const char *err_string = NULL;
+    multi_log_state *mls = get_module_config (cmd->server->module_config,
+					      &config_log_module);
+    config_log_state *cls;
+
+    cls = (config_log_state*)push_array(mls->config_logs);
+    cls->fname = fn;
+    if (!fmt)
+	cls->format = NULL;
+    else
+	cls->format = parse_log_string (cmd->pool, fmt, &err_string);
+    cls->log_fd = -1;
+    
+    return err_string;
+}
+
+const char *set_transfer_log(cmd_parms *cmd, void *dummy, char *fn)
+{
+    return add_custom_log(cmd, dummy, fn, NULL);
+}
+
+const char *set_cookie_log(cmd_parms *cmd, void *dummy, char *fn)
+{
+    return add_custom_log(cmd, dummy, fn, "%{Cookie}n \"%r\" %t");
+}
+
 command_rec config_log_cmds[] = {
-{ "TransferLog", set_config_log, NULL, RSRC_CONF, TAKE1,
+{ "CustomLog", add_custom_log, NULL, RSRC_CONF, TAKE2,
+    "a file name and a custom log format string" },
+{ "TransferLog", set_transfer_log, NULL, RSRC_CONF, TAKE1,
     "the filename of the access log" },
 { "LogFormat", log_format, NULL, RSRC_CONF, TAKE1,
-      "a log format string (see docs)" },
+    "a log format string (see docs)" },
+{ "CookieLog", set_cookie_log, NULL, RSRC_CONF, TAKE1,
+    "the filename of the cookie log" },
 { NULL }
 };
 
@@ -487,71 +682,89 @@ void config_log_child (void *cmd)
     
     cleanup_for_exec();
     signal (SIGHUP, SIG_IGN);
+#ifdef __EMX__
+    /* For OS/2 we need to use a '/' */
+    execl (SHELL_PATH, SHELL_PATH, "/c", (char *)cmd, NULL);
+#else
     execl (SHELL_PATH, SHELL_PATH, "-c", (char *)cmd, NULL);
+#endif
+    perror ("exec");
     fprintf (stderr, "Exec of shell for logging failed!!!\n");
     exit (1);
 }
 
 config_log_state *open_config_log (server_rec *s, pool *p,
-				   config_log_state *defaults)
-{
-    config_log_state *cls = get_module_config (s->module_config,
-					       &config_log_module);
-  
+				   config_log_state *cls,
+				   array_header *default_format) {
     if (cls->log_fd > 0) return cls; /* virtual config shared w/main server */
-    
-    if (cls->format == NULL) {
-	char *dummy;
-	
-	if (defaults) cls->format = defaults->format;
-	else cls->format = parse_log_string (p, DEFAULT_LOG_FORMAT, &dummy);
-    }
 
-    if (cls->fname == NULL) {
-	if (defaults) {
-	    cls->log_fd = defaults->log_fd;
-	    return cls;
-	}
-	else cls->fname = DEFAULT_XFERLOG;
-    }
-    
     if (*cls->fname == '|') {
-	FILE *dummy;
-	
-	spawn_child(p, config_log_child, (void *)(cls->fname+1),
-		    kill_after_timeout, &dummy, NULL);
+        FILE *dummy;
+        
+        if (!spawn_child (p, config_log_child, (void *)(cls->fname+1),
+                    kill_after_timeout, &dummy, NULL)) {
+	    perror ("spawn_child");
+            fprintf (stderr, "Couldn't fork child for TransferLog process\n");
+            exit (1);
+        }
 
-	if (dummy == NULL) {
-	    fprintf (stderr, "Couldn't fork child for TransferLog process\n");
-	    exit (1);
-	}
-
-	cls->log_fd = fileno (dummy);
+        cls->log_fd = fileno (dummy);
     }
     else {
-	char *fname = server_root_relative (p, cls->fname);
-	if((cls->log_fd = popenf(p, fname, xfer_flags, xfer_mode)) < 0) {
-	    fprintf (stderr,
-		     "httpd: could not open transfer log file %s.\n", fname);
-	    perror("open");
-	    exit(1);
-	}
+        char *fname = server_root_relative (p, cls->fname);
+        if((cls->log_fd = popenf(p, fname, xfer_flags, xfer_mode)) < 0) {
+            perror("open");
+            fprintf (stderr,
+                     "httpd: could not open transfer log file %s.\n", fname);
+            exit(1);
+        }
     }
 
     return cls;
 }
 
+config_log_state *open_multi_logs (server_rec *s, pool *p)
+{
+    int i;
+    multi_log_state *mls = get_module_config(s->module_config,
+                                             &config_log_module);
+    config_log_state *clsarray;
+    const char *dummy;
+
+    if (!mls->default_format)
+      mls->default_format = parse_log_string (p, DEFAULT_LOG_FORMAT, &dummy);
+
+    if (mls->config_logs->nelts) {
+        clsarray = (config_log_state *)mls->config_logs->elts;
+        for (i = 0; i < mls->config_logs->nelts; ++i) {
+            config_log_state *cls = &clsarray[i];
+
+            cls = open_config_log(s, p, cls, mls->default_format);
+                }
+    }
+    else if (mls->server_config_logs) {
+        clsarray = (config_log_state *)mls->server_config_logs->elts;
+        for (i = 0; i < mls->server_config_logs->nelts; ++i) {
+            config_log_state *cls = &clsarray[i];
+
+            cls = open_config_log(s, p, cls, mls->default_format);
+        }
+    }
+
+    return NULL;
+}
+
 void init_config_log (server_rec *s, pool *p)
 {
     /* First, do "physical" server, which gets default log fd and format
      * for the virtual servers, if they don't override...
      */
     
-    config_log_state *default_conf = open_config_log (s, p, NULL);
+    open_multi_logs (s, p);
     
     /* Then, virtual servers */
     
-    for (s = s->next; s; s = s->next) open_config_log (s, p, default_conf);
+    for (s = s->next; s; s = s->next) open_multi_logs (s, p);
 }
 
 module config_log_module = {
@@ -560,7 +773,7 @@ module config_log_module = {
    NULL,			/* create per-dir config */
    NULL,			/* merge per-dir config */
    make_config_log_state,	/* server config */
-   NULL,			/* merge server config */
+   merge_config_log_state,     	/* merge server config */
    config_log_cmds,		/* command table */
    NULL,			/* handlers */
    NULL,			/* filename translation */
@@ -569,5 +782,6 @@ module config_log_module = {
    NULL,			/* check access */
    NULL,			/* type_checker */
    NULL,			/* fixups */
-   config_log_transaction	/* logger */
+   multi_log_transaction,	/* logger */
+   NULL				/* header parser */
 };
