@@ -1,6 +1,5 @@
-
 /* ====================================================================
- * Copyright (c) 1995 The Apache Group.  All rights reserved.
+ * Copyright (c) 1995-1997 The Apache Group.  All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
@@ -51,7 +50,6 @@
  *
  */
 
-
 /*
  * http_mime.c: Sends/gets MIME headers for requests
  * 
@@ -114,28 +112,31 @@ void *merge_mime_dir_configs (pool *p, void *basev, void *addv)
     return new;
 }
 
-char *add_type(cmd_parms *cmd, mime_dir_config *m, char *ct, char *ext)
+const char *add_type(cmd_parms *cmd, mime_dir_config *m, char *ct, char *ext)
 {
     if (*ext == '.') ++ext;
     table_set (m->forced_types, ext, ct);
     return NULL;
 }
 
-char *add_encoding(cmd_parms *cmd, mime_dir_config *m, char *enc, char *ext)
+const char *add_encoding(cmd_parms *cmd, mime_dir_config *m, char *enc,
+			 char *ext)
 {
     if (*ext == '.') ++ext;
     table_set (m->encoding_types, ext, enc);
     return NULL;
 }
 
-char *add_language(cmd_parms *cmd, mime_dir_config *m, char *lang, char *ext)
+const char *add_language(cmd_parms *cmd, mime_dir_config *m, char *lang,
+			 char *ext)
 {
     if (*ext == '.') ++ext;
     table_set (m->language_types, ext, lang);
     return NULL;
 }
 
-char *add_handler(cmd_parms *cmd, mime_dir_config *m, char *hdlr, char *ext)
+const char *add_handler(cmd_parms *cmd, mime_dir_config *m, char *hdlr,
+			char *ext)
 {
     if (*ext == '.') ++ext;
     table_set (m->handlers, ext, hdlr);
@@ -146,7 +147,7 @@ char *add_handler(cmd_parms *cmd, mime_dir_config *m, char *hdlr, char *ext)
  * the name of its config file, so...
  */
 
-char *set_types_config (cmd_parms *cmd, void *dummy, char *arg)
+const char *set_types_config (cmd_parms *cmd, void *dummy, char *arg)
 {
     set_module_config (cmd->server->module_config, &mime_module,
 		       pstrdup (cmd->pool, arg));
@@ -192,9 +193,9 @@ void init_mime (server_rec *s, pool *p)
     types_confname = server_root_relative (p, types_confname);
 
     if(!(f = fopen(types_confname,"r"))) {
+        perror("fopen");
         fprintf(stderr,"httpd: could not open mime types file %s\n",
                 types_confname);
-        perror("fopen");
         exit(1);
     }
 
@@ -202,7 +203,7 @@ void init_mime (server_rec *s, pool *p)
         hash_buckets[x] = make_table (p, 10);
 
     while(!(cfg_getline(l,MAX_STRING_LEN,f))) {
-        char *ll = l, *ct;
+        const char *ll = l, *ct;
       
         if(l[0] == '#') continue;
         ct = getword_conf (p, &ll);
@@ -218,7 +219,7 @@ void init_mime (server_rec *s, pool *p)
 
 int find_ct(request_rec *r)
 {
-    char *fn = strrchr(r->filename, '/');
+    const char *fn = strrchr(r->filename, '/');
     mime_dir_config *conf =
       (mime_dir_config *)get_module_config(r->per_dir_config, &mime_module);
     char *ext, *type, *orighandler = r->handler;
@@ -249,7 +250,13 @@ int find_ct(request_rec *r)
 
       /* Check for Content-Language */
       if ((type = table_get (conf->language_types, ext))) {
-	  r->content_language = type;
+	  char **new;
+
+	  r->content_language = type; /* back compat. only */
+	  if (!r->content_languages)
+	      r->content_languages = make_array (r->pool, 2, sizeof(char*));
+	  new = (char **)push_array (r->content_languages);
+	  *new = type;
 	  found = 1;
       }
 	
@@ -277,6 +284,7 @@ int find_ct(request_rec *r)
       if (!found) {
 	r->content_type = NULL;
 	r->content_language = NULL;
+	r->content_languages = NULL;
 	r->content_encoding = NULL;
 	r->handler = orighandler;
       }
@@ -311,5 +319,6 @@ module mime_module = {
    NULL,			/* check access */
    find_ct,			/* type_checker */
    NULL,			/* fixups */
-   NULL				/* logger */
+   NULL,			/* logger */
+   NULL				/* header parser */
 };
