@@ -1,5 +1,5 @@
 /* ====================================================================
- * Copyright (c) 1995, 1996 The Apache Group.  All rights reserved.
+ * Copyright (c) 1995-1997 The Apache Group.  All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
@@ -50,22 +50,29 @@
  *
  */
 
-/* Status Module.  Provide a way of getting at the internal Apache
- * status information without having to worry where the scoreboard is
- * held.
+/* Status Module.  Display lots of internal data about how Apache is
+ * performing and the state of all children processes.
  *
- * AddType application/x-httpd-status .status
- * You might like to do this in a .htaccess in a protected directory only
+ * To enable this, add the following lines into any config file:
  *
- * GET /.status - Returns pretty page for system admin user
- * GET /.status?refresh - Returns page with 1 second refresh
- * GET /.status?refresh=6 - Returns page with refresh every 6 seconds
- * GET /.status?auto - Returns page with data for automatic parsing
- * GET /.status?notable - Returns page with no table niceties
+ * <Location /server-status>
+ * SetHandler server-status
+ * </Location>
+ *
+ * You may want to protect this location by password or domain so no one
+ * else can look at it.  Then you can access the statistics with a URL like:
+ *
+ * http://your_server_name/server-status
+ *
+ * /server-status - Returns page using tables
+ * /server-status?notable - Returns page for browsers without table support
+ * /server-status?refresh - Returns page with 1 second refresh
+ * /server-status?refresh=6 - Returns page with refresh every 6 seconds
+ * /server-status?auto - Returns page with data for automatic parsing
  *
  * Mark Cox, mark@ukweb.com, November 1995
  *
- * 12.11.95 Initial version for telescope.org
+ * 12.11.95 Initial version for www.telescope.org
  * 13.3.96  Updated to remove rprintf's [Mark]
  * 18.3.96  Added CPU usage, process information, and tidied [Ben Laurie]
  * 18.3.96  Make extra Scoreboard variables #definable
@@ -78,8 +85,7 @@
  * 18.5.96  Adapted to use new rprintf() routine, incidentally fixing a missing
  *          piece in short reports [Ben Laurie]
  * 21.5.96  Additional Status codes (DNS and LOGGING only enabled if
-             extended STATUS is enabled) [George Burgyan/Jim J.]
- */
+             extended STATUS is enabled) [George Burgyan/Jim J.]  */
 
 #include "httpd.h"
 #include "http_config.h"
@@ -94,7 +100,7 @@
 #include <machine/param.h>
 #endif
 
-#define STATUS_MAXLINE		50
+#define STATUS_MAXLINE		64
 
 #define KBYTE			1024
 #define	MBYTE			1048576L
@@ -116,6 +122,16 @@ void format_byte_out(request_rec *r,unsigned long bytes)
 	rprintf(r,"%.1f GB",(float)bytes/GBYTE);
 }
 
+void format_kbyte_out(request_rec *r,unsigned long kbytes)
+{
+    if (kbytes < KBYTE)
+	rprintf(r,"%d kB",(int)kbytes);
+    else if (kbytes < MBYTE)
+	rprintf(r,"%.1f MB",(float)kbytes/KBYTE);
+    else
+	rprintf(r,"%.1f GB",(float)kbytes/MBYTE);
+}
+
 void show_time(request_rec *r,time_t tsecs)
 {
     long days,hrs,mins,secs;
@@ -186,6 +202,7 @@ int status_handler (request_rec *r)
     unsigned long my_lres,my_bytes,conn_bytes;
     unsigned short conn_lres;
     unsigned long bcount=0;
+    unsigned long kbcount=0;
 #ifdef NEXT
     float tick=HZ;
 #else
@@ -210,6 +227,7 @@ int status_handler (request_rec *r)
     status[SERVER_BUSY_KEEPALIVE]='K';
     status[SERVER_BUSY_LOG]='L';
     status[SERVER_BUSY_DNS]='D';
+    status[SERVER_GRACEFUL]='G';
 
     if (r->method_number != M_GET) return NOT_IMPLEMENTED;
     r->content_type = "text/html";
@@ -247,7 +265,6 @@ int status_handler (request_rec *r)
         }
     }
 
-    soft_timeout ("send status info", r);
     send_http_header(r);
 
     if (r->header_only) 
@@ -263,7 +280,8 @@ int status_handler (request_rec *r)
 	    ready++;
         else if (res == SERVER_BUSY_READ || res==SERVER_BUSY_WRITE || 
 		 res == SERVER_STARTING || res==SERVER_BUSY_KEEPALIVE ||
-		 res == SERVER_BUSY_LOG || res==SERVER_BUSY_DNS)
+		 res == SERVER_BUSY_LOG || res==SERVER_BUSY_DNS ||
+		 res == SERVER_GRACEFUL)
 	    busy++;
 #if defined(STATUS)
         lres = score_record.access_count;
@@ -277,17 +295,23 @@ int status_handler (request_rec *r)
 	    tcs+=score_record.times.tms_cstime;
             count+=lres;
 	    bcount+=bytes;
+	    if (bcount>=KBYTE) {
+	    	kbcount += (bcount >> 10);
+		bcount = bcount & 0x3ff;
+	    }
 	}
 #endif /* STATUS */
     }
 
     up_time=nowtime-restart_time;
 
+    hard_timeout("send status info", r);
+
     if (!short_report)
     {
-        rputs("<html><head><title>Apache Status</title></head><body>\n",r);
-        rputs("<h1>Apache Server Status for ",r);
-	rvputs(r,server->server_hostname,"</h1>\n\n",NULL);
+        rputs("<HTML><HEAD>\n<TITLE>Apache Status</TITLE>\n</HEAD><BODY>\n",r);
+        rputs("<H1>Apache Server Status for ",r);
+	rvputs(r,server->server_hostname,"</H1>\n\n",NULL);
 	rvputs(r,"Current Time: ",asctime(localtime(&nowtime)),"<br>\n",NULL);
 	rvputs(r,"Restart Time: ",asctime(localtime(&restart_time)),"<br>\n",
 	       NULL);
@@ -299,30 +323,37 @@ int status_handler (request_rec *r)
 #if defined(STATUS)
     if (short_report)
     {
-        rprintf(r,"Total Accesses: %lu\nTotal Bytes: %lu\n",count,bcount);
+        rprintf(r,"Total Accesses: %lu\nTotal kBytes: %lu\n",count,kbcount);
 
+#ifndef __EMX__
+    /* Allow for OS/2 not having CPU stats */
 	if(ts || tu || tcu || tcs)
 	    rprintf(r,"CPULoad: %g\n",(tu+ts+tcu+tcs)/tick/up_time*100.);
+#endif
 
 	rprintf(r,"Uptime: %ld\n",(long)(up_time));
 	if (up_time>0)
 	    rprintf(r,"ReqPerSec: %g\n",(float)count/(float)up_time);
 
 	if (up_time>0)
-	    rprintf(r,"BytesPerSec: %g\n",(float)bcount/(float)up_time);
+	    rprintf(r,"BytesPerSec: %g\n",KBYTE*(float)kbcount/(float)up_time);
 
 	if (count>0)
-	    rprintf(r,"BytesPerReq: %g\n",(float)bcount/(float)count);
+	    rprintf(r,"BytesPerReq: %g\n",KBYTE*(float)kbcount/(float)count);
     } else /* !short_report */
     {
 	rprintf(r,"Total accesses: %lu - Total Traffic: ", count);
-	format_byte_out(r,bcount);
+	format_kbyte_out(r,kbcount);
+
+#ifndef __EMX__
+	/* Allow for OS/2 not having CPU stats */
 	rputs("<br>\n",r);
         rprintf(r,"CPU Usage: u%g s%g cu%g cs%g",
 		tu/tick,ts/tick,tcu/tick,tcs/tick);
 
 	if(ts || tu || tcu || tcs)
 	    rprintf(r," - %.3g%% CPU load",(tu+ts+tcu+tcs)/tick/up_time*100.);
+#endif
 
 	rputs("<br>\n",r);
 
@@ -332,53 +363,54 @@ int status_handler (request_rec *r)
 
 	if (up_time>0)
 	{
-	    format_byte_out(r,(float)bcount/(float)up_time);
+	    format_byte_out(r,KBYTE*(float)kbcount/(float)up_time);
 	    rputs("/second - ",r);
 	}
 
 	if (count>0)
 	{
-	    format_byte_out(r,(float)bcount/(float)count);
+	    format_byte_out(r,KBYTE*(float)kbcount/(float)count);
 	    rputs("/request",r);
 	}
 
-	rputs("<p>\n",r);
+	rputs("<br>\n",r);
     } /* short_report */
 #endif /* STATUS */
 
-    /* send the scoreboard 'table' out */
+    if (!short_report)
+        rprintf(r,"\n%d requests currently being processed, %d idle servers\n"
+		,busy,ready);
+    else
+        rprintf(r,"BusyServers: %d\nIdleServers: %d\n",busy,ready);
 
-    rputs("Scoreboard: \n",r);
+    /* send the scoreboard 'table' out */
 
     if(!short_report)
 	rputs("<PRE>",r);
-
-    rputs("\n",r);
+    else
+        rputs("Scoreboard: ",r);
 
     for (i = 0; i<HARD_SERVER_LIMIT; ++i)
     {
 	rputc(stat_buffer[i], r);
-	if(i%STATUS_MAXLINE == (STATUS_MAXLINE - 1))
+	if((i%STATUS_MAXLINE == (STATUS_MAXLINE - 1))&&!short_report)
 	    rputs("\n",r);
     }
 
     if (short_report)
-    {
-        rprintf(r,"\nBusyServers: %d\nIdleServers: %d\n",busy,ready);
-    }
-    else 
-    {
+        rputs("\n",r);
+    else {
 	rputs("</PRE>\n",r);
-	rputs("Key:<br> \n",r);
-	rputs("\"<code>_</code>\" Waiting for Connection, \n",r);
-	rputs("\"<code>S</code>\" Starting up,<br> \n",r);
-	rputs("\"<code>R</code>\" Reading Request, \n",r);
-	rputs("\"<code>W</code>\" Sending Reply,<br> \n",r);
-	rputs("\"<code>K</code>\" Keepalive (read), \n",r);
-	rputs("\"<code>D</code>\" DNS Lookup, \n",r);
-	rputs("\"<code>L</code>\" Logging<p>\n",r);
-        rprintf(r,"\n%d requests currently being processed, %d idle servers\n"
-		,busy,ready);
+	rputs("Scoreboard Key: <br>\n",r);
+	rputs("\"<B><code>_</code></B>\" Waiting for Connection, \n",r);
+	rputs("\"<B><code>S</code></B>\" Starting up, \n",r);
+	rputs("\"<B><code>R</code></B>\" Reading Request,<BR>\n",r);
+	rputs("\"<B><code>W</code></B>\" Sending Reply, \n",r);
+	rputs("\"<B><code>K</code></B>\" Keepalive (read), \n",r);
+	rputs("\"<B><code>D</code></B>\" DNS Lookup,<BR>\n",r);
+	rputs("\"<B><code>L</code></B>\" Logging, \n",r);
+	rputs("\"<B><code>G</code></B>\" Gracefully finishing, \n",r);
+	rputs("\"<B><code>.</code></B>\" Open slot with no current process<P>\n",r);
     }
 
 #if defined(STATUS)
@@ -386,7 +418,12 @@ int status_handler (request_rec *r)
     	if(no_table_report)
             rputs("<p><hr><h2>Server Details</h2>\n\n",r);
 	else
-            rputs("<p>\n\n<table border=0><tr><th>Srv<th>PID<th>Acc<th>M<th>CPU\n<th>SS<th>Conn<th>Child<th>Slot<th>Host<th>Request</tr>\n\n",r);
+#ifdef __EMX__
+            /* Allow for OS/2 not having CPU stats */
+            rputs("<p>\n\n<table border=0><tr><th>Srv<th>PID<th>Acc<th>M\n<th>SS<th>Conn<th>Child<th>Slot<th>Host<th>VHost<th>Request</tr>\n\n",r);
+#else
+            rputs("<p>\n\n<table border=0><tr><th>Srv<th>PID<th>Acc<th>M<th>CPU\n<th>SS<th>Conn<th>Child<th>Slot<th>Host<th>VHost<th>Request</tr>\n\n",r);
+#endif
 
 
     for (i = 0; i<HARD_SERVER_LIMIT; ++i)
@@ -434,12 +471,24 @@ int status_handler (request_rec *r)
 		        case SERVER_DEAD:
 		            rputs("Dead",r);
 		            break;
+			case SERVER_GRACEFUL:
+			    rputs("Graceful",r);
+			    break;
+			default:
+			    rputs("?STATE?",r);
+			    break;
 		    }
+#ifdef __EMX__
+                    /* Allow for OS/2 not having CPU stats */
+                    rprintf(r,"]\n %s (",
+#else
+
 		    rprintf(r,"] u%g s%g cu%g cs%g\n %s (",
 			    score_record.times.tms_utime/tick,
 			    score_record.times.tms_stime/tick,
 			    score_record.times.tms_cutime/tick,
 			    score_record.times.tms_cstime/tick,
+#endif
 			    asctime(localtime(&score_record.last_used)));
 		    format_byte_out(r,conn_bytes);
 		    rputs("|",r);
@@ -448,7 +497,8 @@ int status_handler (request_rec *r)
 		    format_byte_out(r,bytes);
 		    rputs(")\n",r);
 		    rprintf(r," <i>%s {%s}</i><br>\n\n",
-			    score_record.client, score_record.request);
+			    score_record.client,
+			    escape_html(r->pool, score_record.request));
 		}
 		else /* !no_table_report */
 		{
@@ -481,18 +531,30 @@ int status_handler (request_rec *r)
 		        case SERVER_DEAD:
 		            rputs("<td>.",r);
 		            break;
+			case SERVER_GRACEFUL:
+			    rputs("<td>G",r);
+			    break;
+			default:
+			    rputs("<td>?",r);
+			    break;
 		    }
+#ifdef __EMX__
+	            /* Allow for OS/2 not having CPU stats */
+        	    rprintf(r,"\n<td>%.0f",
+#else
 		    rprintf(r,"\n<td>%.2f<td>%.0f",
 			    (score_record.times.tms_utime +
 			    score_record.times.tms_stime +
 			    score_record.times.tms_cutime +
 			    score_record.times.tms_cstime)/tick,
+#endif
 			    difftime(nowtime, score_record.last_used));
 		    rprintf(r,"<td>%-1.1f<td>%-2.2f<td>%-2.2f\n",
 			(float)conn_bytes/KBYTE, (float)my_bytes/MBYTE,
 			(float)bytes/MBYTE);
-		    rprintf(r,"<td>%s<td nowrap>%s</tr>\n\n",
-			score_record.client, score_record.request);
+		    rprintf(r,"<td>%s<td nowrap>%s<td nowrap>%s</tr>\n\n",
+			    score_record.client, score_record.vhost,
+			    escape_html(r->pool, score_record.request));
 		}	/* no_table_report */
 	    }		/* !short_report */
 	}		/* if (<active child>) */
@@ -500,6 +562,20 @@ int status_handler (request_rec *r)
 
     if (!(short_report || no_table_report))
     {
+#ifdef __EMX__
+	rputs("</table>\n \
+<hr> \
+<table>\n \
+<tr><th>Srv<td>Server number\n \
+<tr><th>PID<td>OS process ID\n \
+<tr><th>Acc<td>Number of accesses this connection / this child / this slot\n \
+<tr><th>M<td>Mode of operation\n \
+<tr><th>SS<td>Seconds since beginning of most recent request\n \
+<tr><th>Conn<td>Kilobytes transferred this connection\n \
+<tr><th>Child<td>Megabytes transferred this child\n \
+<tr><th>Slot<td>Total megabytes transferred this slot\n \
+</table>\n",r);
+#else
 	rputs("</table>\n \
 <hr> \
 <table>\n \
@@ -513,21 +589,22 @@ int status_handler (request_rec *r)
 <tr><th>Child<td>Megabytes transferred this child\n \
 <tr><th>Slot<td>Total megabytes transferred this slot\n \
 </table>\n",r);
+#endif
     }
 
 #else /* !defined(STATUS) */
 
-    rputs("<hr>To obtain a full report with current status information \n",r);
-    rputs("you need to recompile Apache adding the <code>-DSTATUS</code> \n",r);
-    rputs("directive on the <code>CFLAGS</code> line in the \n",r);
-    rputs("<code>Configuration</code> file.\n",r);
-    rputs("<code>DNS</code> and <code>LOGGING</code> status \n",r);
-    rputs("also requires the <code>-DSTATUS</code> directive. \n",r);
+    rputs("<hr>To obtain a full report with current status information and",r);
+    rputs(" DNS and LOGGING status codes \n",r);
+    rputs("you need to recompile Apache after adding the line <pre>",r);
+    rputs("Rule STATUS=yes</pre>into the file <code>Configuration</code>\n",r);
 
 #endif /* STATUS */
 
     if (!short_report)
-        rputs("</body></html>",r);
+        rputs("</BODY></HTML>\n",r);
+
+    kill_timeout(r);
     return 0;
 }
 
@@ -554,38 +631,6 @@ module status_module =
    NULL,			/* check access */
    NULL,			/* type_checker */
    NULL,			/* fixups */
-   NULL				/* logger */
+   NULL,			/* logger */
+   NULL				/* header parser */
 };
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
