@@ -1,6 +1,5 @@
-
 /* ====================================================================
- * Copyright (c) 1995 The Apache Group.  All rights reserved.
+ * Copyright (c) 1995-1997 The Apache Group.  All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
@@ -72,6 +71,7 @@
 #define SERVER_BUSY_KEEPALIVE 5 /* Waiting for more requests via keepalive */
 #define SERVER_BUSY_LOG 6       /* Logging the request */
 #define SERVER_BUSY_DNS 7       /* Looking up a hostname */
+#define SERVER_GRACEFUL 8	/* server is gracefully finishing request */
 
 typedef struct {
     pid_t pid;
@@ -87,8 +87,24 @@ typedef struct {
     time_t last_used;
     char client[32];	/* Keep 'em small... */
     char request[64];	/* We just want an idea... */
+    char vhost[32];     /* What virtual host is being accessed? */
 #endif
 } short_score;
 
+typedef struct
+    {
+    int exit_generation;	/* Set by the main process if a graceful
+				   restart is required */
+    } global_score;
+
+typedef struct
+    {
+    short_score servers[HARD_SERVER_LIMIT];
+    global_score global;
+    } scoreboard;
+
+#define SCOREBOARD_SIZE		sizeof(scoreboard)
+
 extern void sync_scoreboard_image(void);
 short_score get_scoreboard_info(int x);
+
