@@ -1,6 +1,5 @@
-
 /* ====================================================================
- * Copyright (c) 1995 The Apache Group.  All rights reserved.
+ * Copyright (c) 1995-1997 The Apache Group.  All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
@@ -51,58 +50,20 @@
  *
  */
 
-
 /*
- * str.c: string utility things
+ * util.c: string utility things
  * 
  * 3/21/93 Rob McCool
+ * 1995-96 Many changes by the Apache Group
  * 
  */
 
-
 #include "httpd.h"
 #include "http_conf_globals.h"	/* for user_id & group_id */
-#ifdef QNX
-#include <time.h>
-#endif
 
-#ifdef NOTDEF
-extern char** environ;
-
-/* taken from bdflush-1.5 for Linux source code */
-void inststr(char *dst[], int argc, char *src)
-{
-    if (strlen(src) <= strlen(dst[0]))
-    {
-        char *ptr;
-
-        for (ptr = dst[0]; *ptr; *(ptr++) = '\0');
-
-        strcpy(dst[0], src);
-    } else
-    {
-        /* stolen from the source to perl 4.036 (assigning to $0) */
-        char *ptr, *ptr2;
-        int count;
-        ptr = dst[0] + strlen(dst[0]);
-        for (count = 1; count < argc; count++) {
-            if (dst[count] == ptr + 1)
-                ptr += strlen(++ptr);
-        }
-        if (environ[0] == ptr + 1) {
-            for (count = 0; environ[count]; count++)
-                if (environ[count] == ptr + 1)
-                    ptr += strlen(++ptr);
-        }
-        count = 0;
-        for (ptr2 = dst[0]; ptr2 <= ptr; ptr2++) {
-            *ptr2 = '\0';
-            count++;
-        }
-        strncpy(dst[0], src, count);
-    }
-}
-#endif
+const char month_snames[12][4] = {
+    "Jan","Feb","Mar","Apr","May","Jun","Jul","Aug","Sep","Oct","Nov","Dec"
+};
 
 char *get_time() {
     time_t t;
@@ -114,7 +75,7 @@ char *get_time() {
     return (time_string);
 }
 
-char *ht_time(pool *p, time_t t, char *fmt, int gmt) {
+char *ht_time(pool *p, time_t t, const char *fmt, int gmt) {
     char ts[MAX_STRING_LEN];
     struct tm *tms;
 
@@ -126,41 +87,57 @@ char *ht_time(pool *p, time_t t, char *fmt, int gmt) {
 }
 
 char *gm_timestr_822(pool *p, time_t sec) {
-    return ht_time(p, sec, HTTP_TIME_FORMAT, 1);
+    static const char *const days[7]=
+       {"Sun","Mon", "Tue", "Wed", "Thu", "Fri", "Sat"};
+    char ts[50];
+    struct tm *tms;
+
+    tms = gmtime(&sec);
+
+/* RFC date format; as strftime '%a, %d %b %Y %T GMT' */
+    ap_snprintf(ts, sizeof(ts), 
+	    "%s, %.2d %s %d %.2d:%.2d:%.2d GMT", days[tms->tm_wday],
+	    tms->tm_mday, month_snames[tms->tm_mon], tms->tm_year + 1900,
+	    tms->tm_hour, tms->tm_min, tms->tm_sec);
+
+    return pstrdup (p, ts);
 }
 
 /* What a pain in the ass. */
-struct tm *get_gmtoff(long *tz) {
-    time_t tt;
+#if defined(HAVE_GMTOFF)
+struct tm *get_gmtoff(int *tz) {
+    time_t tt = time(NULL);
     struct tm *t;
 
-    tt = time(NULL);
     t = localtime(&tt);
-#if defined(HAS_GMTOFF)
-    *tz = t->tm_gmtoff;
-#elif !defined(NO_TIMEZONE)
-    *tz = - timezone;
-    if(t->tm_isdst)
-        *tz += 3600;
-#else
-  {
-    static struct tm loc_t;
-
-    loc_t = *t;   /* save it */
-    t = gmtime(&tt);
-    *tz = mktime(&loc_t) - mktime(t);
-    t = &loc_t; /* return pointer to saved time */
-  }
-#endif
+    *tz = (int) (t->tm_gmtoff / 60);
     return t;
 }
+#else
+struct tm *get_gmtoff(int *tz) {
+    time_t tt = time(NULL);
+    struct tm gmt;
+    struct tm *t;
+    int days, hours, minutes;
+
+    /* Assume we are never more than 24 hours away. */
+    gmt = *gmtime(&tt); /* remember gmtime/localtime return ptr to static */
+    t = localtime(&tt); /* buffer... so be careful */
+    days = t->tm_yday - gmt.tm_yday;
+    hours = ((days < -1 ? 24 : 1 < days ? -24 : days * 24)
+		 + t->tm_hour - gmt.tm_hour);
+    minutes = hours * 60 + t->tm_min - gmt.tm_min;
+    *tz = minutes;
+    return t;
+}
+#endif
 
 
 /* Match = 0, NoMatch = 1, Abort = -1 */
 /* Based loosely on sections of wildmat.c by Rich Salz
  * Hmmm... shouldn't this really go component by component?
  */
-int strcmp_match(char *str, char *exp) {
+int strcmp_match(const char *str, const char *exp) {
     int x,y;
 
     for(x=0,y=0;exp[y];++y,++x) {
@@ -183,7 +160,30 @@ int strcmp_match(char *str, char *exp) {
     return (str[x] != '\0');
 }
 
-int is_matchexp(char *str) {
+int strcasecmp_match(const char *str, const char *exp) {
+    int x,y;
+
+    for(x=0,y=0;exp[y];++y,++x) {
+        if((!str[x]) && (exp[y] != '*'))
+            return -1;
+        if(exp[y] == '*') {
+            while(exp[++y] == '*');
+            if(!exp[y])
+                return 0;
+            while(str[x]) {
+                int ret;
+                if((ret = strcasecmp_match(&str[x++],&exp[y])) != 1)
+                    return ret;
+            }
+            return -1;
+        } else 
+            if((exp[y] != '?') && (tolower(str[x]) != tolower(exp[y])))
+                return 1;
+    }
+    return (str[x] != '\0');
+}
+
+int is_matchexp(const char *str) {
     register int x;
 
     for(x=0;str[x];x++)
@@ -192,6 +192,84 @@ int is_matchexp(char *str) {
     return 0;
 }
 
+/* This function substitutes for $0-$9, filling in regular expression
+ * submatches. Pass it the same nmatch and pmatch arguments that you
+ * passed regexec(). pmatch should not be greater than the maximum number
+ * of subexpressions - i.e. one more than the re_nsub member of regex_t.
+ *
+ * input should be the string with the $-expressions, source should be the
+ * string that was matched against.
+ *
+ * It returns the substituted string, or NULL on error.
+ *
+ * Parts of this code are based on Henry Spencer's regsub(), from his
+ * AT&T V8 regexp package.
+ */
+
+char *pregsub(pool *p, const char *input, const char *source,
+	      size_t nmatch, regmatch_t pmatch[]) {
+    const char *src = input;
+    char *dest, *dst;
+    char c;
+    int no, len;
+
+    if (!source) return NULL;
+    if (!nmatch) return pstrdup(p, src);
+
+    /* First pass, find the size */
+
+    len = 0;
+
+    while ((c = *src++) != '\0') {
+	if (c == '&')
+	    no = 0;
+	else if (c == '$' && isdigit(*src))
+	    no = *src++ - '0';
+	else
+	    no = -1;
+	
+	if (no < 0) {   /* Ordinary character. */
+	    if (c == '\\' && (*src == '$' || *src == '&'))
+		c = *src++;
+	    len++;
+	} else if (no <= nmatch && pmatch[no].rm_so < pmatch[no].rm_eo) {
+	    len += pmatch[no].rm_eo - pmatch[no].rm_so;
+	}
+
+    }
+
+    dest = dst = pcalloc(p, len + 1);
+
+    /* Now actually fill in the string */
+
+    src = input;
+
+    while ((c = *src++) != '\0') {
+	if (c == '&')
+	    no = 0;
+	else if (c == '$' && isdigit(*src))
+	    no = *src++ - '0';
+	else
+	    no = -1;
+	
+	if (no < 0) {   /* Ordinary character. */
+	    if (c == '\\' && (*src == '$' || *src == '&'))
+		c = *src++;
+	    *dst++ = c;
+	} else if (no <= nmatch && pmatch[no].rm_so < pmatch[no].rm_eo) {
+	    len = pmatch[no].rm_eo - pmatch[no].rm_so;
+	    strncpy(dst, source + pmatch[no].rm_so, len);
+	    dst += len;
+	    if (*(dst-1) == '\0') /* strncpy hit NULL. */
+		return NULL;
+	}
+
+    }
+    *dst = '\0';
+    
+    return dest;
+}
+
 /*
  * Parse .. so we don't compromise security
  */
@@ -260,7 +338,7 @@ void no2slash(char *name) {
 	else x++;
 }
 
-char *make_dirstr(pool *p, char *s, int n) {
+char *make_dirstr(pool *p, const char *s, int n) {
     register int x,f;
     char *res;
 
@@ -281,7 +359,7 @@ char *make_dirstr(pool *p, char *s, int n) {
         return pstrcat (p, s, "/", NULL);
 }
 
-int count_dirs(char *path) {
+int count_dirs(const char *path) {
     register int x,n;
 
     for(x=0,n=0;path[x];x++)
@@ -290,17 +368,22 @@ int count_dirs(char *path) {
 }
 
 
-void chdir_file(char *file) {
+void chdir_file(const char *file) {
     int i;
 
     if((i = rind(file,'/')) == -1)
         return;
-    file[i] = '\0';
+    ((char *)file)[i] = '\0';
     chdir(file);
-    file[i] = '/';
+    ((char *)file)[i] = '/';
 }
 
-char *getword(pool* atrans, char **line, char stop) {
+char *getword_nc(pool* atrans, char **line, char stop)
+    {
+    return getword(atrans,(const char **)line,stop);
+    }
+
+char *getword(pool* atrans, const char **line, char stop) {
     int pos = ind(*line, stop);
     char *res;
 
@@ -321,18 +404,78 @@ char *getword(pool* atrans, char **line, char stop) {
     return res;
 }
 
+char *getword_white_nc(pool* atrans, char **line)
+{
+    return getword_white(atrans,(const char **)line);
+}
+
+char *getword_white(pool* atrans, const char **line) {
+    int pos = -1, x;
+    char *res;
+
+    for(x=0;(*line)[x];x++) {
+        if (isspace((*line)[x])) {
+          pos=x;
+          break;
+      }
+   }
+
+    if (pos == -1) {
+        res = pstrdup (atrans, *line);
+      *line += strlen (*line);
+      return res;
+    }
+
+    res = palloc(atrans, pos + 1);
+    strncpy (res, *line, pos);
+    res[pos] = '\0';
+
+    while (isspace((*line)[pos])) ++pos;
+
+    *line += pos;
+
+    return res;
+}
+
+char *getword_nulls_nc(pool* atrans, char **line, char stop)
+{
+    return getword_nulls(atrans,(const char **)line,stop);
+}
+
+char *getword_nulls(pool* atrans, const char **line, char stop) {
+    int pos = ind(*line, stop);
+    char *res;
+
+    if (pos == -1) {
+        res = pstrdup (atrans, *line);
+	*line += strlen (*line);
+	return res;
+    }
+  
+    res = palloc(atrans, pos + 1);
+    strncpy (res, *line, pos);
+    res[pos] = '\0';
+    
+    ++pos;
+    
+    *line += pos;
+    
+    return res;
+}
+
 /* Get a word, (new) config-file style --- quoted strings and backslashes
  * all honored
  */
 
-char *substring_conf (pool *p, char *start, int len)
+static char *substring_conf (pool *p, const char *start, int len, char quote)
 {
     char *result = palloc (p, len + 2);
     char *resp = result;
     int i;
 
     for (i = 0; i < len; ++i) {
-        if (start[i] == '\\') 
+        if (start[i] == '\\' && (start[i+1] == '/'
+				 || (quote && start[i+1] == quote)))
 	    *resp++ = start[++i];
 	else
 	    *resp++ = start[i];
@@ -342,8 +485,13 @@ char *substring_conf (pool *p, char *start, int len)
     return result;
 }
 
-char *getword_conf(pool* p, char **line) {
-    char *str = *line, *strend, *res;
+char *getword_conf_nc(pool* p, char **line) {
+    return getword_conf(p,(const char **)line);
+}
+
+char *getword_conf(pool* p, const char **line) {
+    const char *str = *line, *strend;
+    char *res;
     char quote;
 
     while (*str && isspace (*str))
@@ -357,19 +505,19 @@ char *getword_conf(pool* p, char **line) {
     if ((quote = *str) == '"' || quote == '\'') {
         strend = str + 1;
 	while (*strend && *strend != quote) {
-	    if (*strend == '\\' && strend[1]) strend += 2;
+	    if (*strend == '\\' && strend[1] && strend[1] == quote)
+		strend += 2;
 	    else ++strend;
 	}
-	res = substring_conf (p, str + 1, strend - str - 1);
+	res = substring_conf (p, str + 1, strend - str - 1, quote);
 
 	if (*strend == quote) ++strend;
     } else {
         strend = str;
 	while (*strend && !isspace (*strend))
-	    if (*strend == '\\' && strend[1]) strend += 2;
-	    else ++strend;
+	    ++strend;
 
-	res = substring_conf (p, str, strend - str);
+	res = substring_conf (p, str, strend - str, 0);
     }
 
     while (*strend && isspace(*strend)) ++ strend;
@@ -377,6 +525,9 @@ char *getword_conf(pool* p, char **line) {
     return res;
 }
 
+#ifdef UNDEF
+/* this function is dangerous, and superceded by getword_white, so don't use it
+ */
 void cfg_getword(char *word, char *line) {
     int x=0,y;
     
@@ -395,6 +546,7 @@ void cfg_getword(char *word, char *line) {
     while(line[x] && isspace(line[x])) ++x;
     for(y=0;(line[y] = line[x]);++x,++y);
 }
+#endif
 
 int
 cfg_getline(char *s, int n, FILE *f) {
@@ -406,6 +558,9 @@ cfg_getline(char *s, int n, FILE *f) {
         c = getc(f);
     } while (c == '\t' || c == ' ');
 
+    if(c == EOF)
+	return 1;
+
     while(1) {
         if((c == '\t') || (c == ' ')) {
             s[i++] = ' ';
@@ -419,7 +574,7 @@ cfg_getline(char *s, int n, FILE *f) {
             /* blast trailing whitespace */
             while(i && (s[i-1] == ' ')) --i;
             s[i] = '\0';
-            return (feof(f) ? 1 : 0);
+	    return 0;
         }
         s[i] = c;
         ++i;
@@ -427,7 +582,118 @@ cfg_getline(char *s, int n, FILE *f) {
     }
 }
 
-char *escape_shell_cmd(pool *p, char *s) {
+/* Retrieve a token, spacing over it and returning a pointer to
+ * the first non-white byte afterwards.  Note that these tokens
+ * are delimited by semis and commas; and can also be delimited
+ * by whitespace at the caller's option.
+ */
+
+char *get_token (pool *p, char **accept_line, int accept_white)
+{
+    char *ptr = *accept_line;
+    char *tok_start;
+    char *token;
+    int tok_len;
+  
+    /* Find first non-white byte */
+    
+    while (*ptr && isspace(*ptr))
+      ++ptr;
+
+    tok_start = ptr;
+    
+    /* find token end, skipping over quoted strings.
+     * (comments are already gone).
+     */
+    
+    while (*ptr && (accept_white || !isspace(*ptr))
+	   && *ptr != ';' && *ptr != ',')
+    {
+	if (*ptr++ == '"')
+	    while (*ptr)
+	        if (*ptr++ == '"') break;
+    }
+	  
+    tok_len = ptr - tok_start;
+    token = palloc (p, tok_len + 1);
+    strncpy (token, tok_start, tok_len);
+    token[tok_len] = '\0';
+    
+    /* Advance accept_line pointer to the next non-white byte */
+
+    while (*ptr && isspace(*ptr))
+      ++ptr;
+
+    *accept_line = ptr;
+    return token;
+}
+
+static char* tspecials = " \t()<>@,;:\\/[]?={}";
+
+/* Next HTTP token from a header line.  Warning --- destructive!
+ * Use only with a copy!
+ */
+
+static char *next_token (char **toks) {
+    char *cp = *toks;
+    char *ret;
+
+    while (*cp && (iscntrl (*cp) || strchr (tspecials, *cp))) {
+        if (*cp == '"')
+	  while (*cp && (*cp != '"')) ++cp;
+	else
+	  ++cp;
+    }
+
+    if (!*cp) ret = NULL;
+    else {
+        ret = cp;
+
+        while (*cp && !iscntrl(*cp) && !strchr (tspecials, *cp))
+            ++cp;
+
+        if (*cp) {
+            *toks = cp + 1;
+            *cp = '\0';
+	}
+        else *toks = cp;
+    }
+
+    return ret;
+}
+
+int find_token (pool *p, const char *line, const char *tok) {
+    char *ltok;
+    char *lcopy;
+
+    if (!line) return 0;
+
+    lcopy = pstrdup (p, line);
+    while ((ltok = next_token (&lcopy)))
+        if (!strcasecmp (ltok, tok))
+            return 1;
+
+    return 0;
+}
+
+int find_last_token (pool *p, const char *line, const char *tok)
+{
+    int llen, tlen, lidx;
+
+    if (!line) return 0;
+
+    llen = strlen(line);
+    tlen = strlen(tok);
+    lidx = llen - tlen;
+
+    if ((lidx < 0) ||
+        ((lidx > 0) && !(isspace(line[lidx-1]) || line[lidx-1] == ',')))
+        return 0;
+
+    return (strncasecmp(&line[lidx], tok, tlen) == 0);
+}
+
+char *escape_shell_cmd(pool *p, const char *s) {
     register int x,y,l;
     char *cmd;
 
@@ -436,6 +702,15 @@ char *escape_shell_cmd(pool *p, char *s) {
     strcpy (cmd, s);
     
     for(x=0;cmd[x];x++) {
+    
+#ifdef __EMX__
+        /* Don't allow '&' in parameters under OS/2. */
+        /* This can be used to send commands to the shell. */
+        if (cmd[x] == '&') {
+            cmd[x] = ' ';
+        }
+#endif
+
         if(ind("&;`'\"|*?~<>^()[]{}$\\\n",cmd[x]) != -1){
             for(y=l+1;y>x;y--)
                 cmd[y] = cmd[y-1];
@@ -460,7 +735,7 @@ void spacetoplus(char *str) {
     for(x=0;str[x];x++) if(str[x] == ' ') str[x] = '+';
 }
 
-char x2c(char *what) {
+static char x2c(const char *what) {
     register char digit;
 
     digit = ((what[0] >= 'A') ? ((what[0] & 0xdf) - 'A')+10 : (what[0] - '0'));
@@ -507,19 +782,25 @@ unescape_url(char *url) {
     else return OK;
 }
 
-char *construct_url(pool *p, char *uri, server_rec *s) {
-    char portnum[10];		/* Long enough.  Really! */
+char *construct_server(pool *p, const char *hostname, unsigned port) {
+    char portnum[22];		
+	/* Long enough, even if port > 16 bits for some reason */
   
-    if (s->port == 80) {
-        return pstrcat (p, "http://", s->server_hostname, uri, NULL);
-    } else {
-        sprintf (portnum, "%d", s->port);
-	return pstrcat (p, "http://", s->server_hostname, ":", portnum, uri,
-			NULL);
+    if (port == DEFAULT_PORT)
+	return (char *)hostname;
+    else {
+        ap_snprintf (portnum, sizeof(portnum), "%u", port);
+	return pstrcat (p, hostname, ":", portnum, NULL);
     }
 }
 
-#define c2x(what,where) sprintf(where,"%%%02x",what)
+char *construct_url(pool *p, const char *uri, const server_rec *s) {
+    return pstrcat (p, "http://",
+		    construct_server(p, s->server_hostname, s->port),
+		    uri, NULL);
+}
+
+#define c2x(what,where) sprintf(where,"%%%02x",(unsigned char)what)
 
 /*
 escape_path_segment() escapes a path segment, as defined in RFC 1808. This
@@ -531,6 +812,9 @@ prefixed with "./" (or the ':' escaped). In the case of Unix, this means
 leaving '/' alone, but otherwise doing what escape_path_segment() does. For
 efficiency reasons, we don't use escape_path_segment(), which is provided for
 reference. Again, RFC 1808 is where this stuff is defined.
+
+If partial is set, os_escape_path() assumes that the path will be appended to
+something with a '/' in it (and thus does not prefix "./").
 */
 
 char *escape_path_segment(pool *p, const char *segment) {
@@ -552,16 +836,20 @@ char *escape_path_segment(pool *p, const char *segment) {
     return copy;
 }
 
-char *os_escape_path(pool *p,const char *path) {
+char *os_escape_path(pool *p,const char *path,int partial) {
   char *copy=palloc(p,3*strlen(path)+3);
   char *s=copy;
-  int colon=ind(path,':');
-  int slash=ind(path,'/');
 
-  if(colon >= 0 && (colon < slash || slash < 0))
+  if(!partial)
     {
-      *s++='.';
-      *s++='/';
+      int colon=ind(path,':');
+      int slash=ind(path,'/');
+
+      if(colon >= 0 && (colon < slash || slash < 0))
+	{
+	  *s++='.';
+	  *s++='/';
+	}
     }
   for( ; *path ; ++path)
     {
@@ -579,22 +867,9 @@ char *os_escape_path(pool *p,const char *path) {
   return copy;
 }
 
-char *escape_uri(pool *p, char *uri) {
-    register int x,y;
-    char *copy = palloc (p, 3 * strlen (uri) + 1);
-            
-    for(x=0,y=0; uri[x]; x++,y++) {
-        if (ind (":% ?+&",(copy[y] = uri[x])) != -1) {
-            c2x(uri[x],&copy[y]);
-            y+=2;
-        }
-    }
-    copy[y] = '\0';
-    return copy;
-}
+/* escape_uri is now a macro for os_escape_path */
 
-char *
-escape_html(pool *p, const char *s)
+char *escape_html(pool *p, const char *s)
 {
     int i, j;
     char *x;
@@ -626,28 +901,7 @@ escape_html(pool *p, const char *s)
     return x;
 }
 
-#ifdef NOTDEF
-
-void escape_url(char *url) {
-    register int x,y;
-    register char digit;
-    char *copy;
-
-    copy = strdup(url);
-            
-    for(x=0,y=0;copy[x];x++,y++) {
-        if(ind("% ?+&",url[y] = copy[x]) != -1) {
-            c2x(copy[x],&url[y]);
-            y+=2;
-        }
-    }
-    url[y] = '\0';
-    free(copy);
-}
-
-#endif
-
-int is_directory(char *path) {
+int is_directory(const char *path) {
     struct stat finfo;
 
     if(stat(path,&finfo) == -1)
@@ -656,7 +910,7 @@ int is_directory(char *path) {
     return(S_ISDIR(finfo.st_mode));
 }
 
-char *make_full_path(pool *a, char *src1,char *src2) {
+char *make_full_path(pool *a, const char *src1, const char *src2) {
     register int x;
 
     x = strlen(src1);
@@ -666,30 +920,50 @@ char *make_full_path(pool *a, char *src1,char *src2) {
     else return pstrcat (a, src1, src2, NULL);
 }
 
-int is_url(char *u) {
+/*
+ * Check for an absoluteURI syntax (see section 3.2 in RFC2068).
+ */
+int is_url(const char *u) {
     register int x;
 
-    for(x=0;u[x] != ':';x++)
-        if((!u[x]) || (!isalpha(u[x])))
+    for (x = 0; u[x] != ':'; x++) {
+        if ((! u[x]) ||
+	    ((! isalpha(u[x])) && (! isdigit(u[x])) &&
+	     (u[x] != '+') && (u[x] != '-') && (u[x] != '.'))) {
             return 0;
+	}
+    }
 
-    if((u[x+1] == '/') && (u[x+2] == '/'))
-        return 1;
-    else return 0;
+    return (x ? 1 : 0);  /* If the first character is ':', it's broken, too */
 }
 
-int can_exec(struct stat *finfo) {
+int can_exec(const struct stat *finfo) {
+#ifdef MULTIPLE_GROUPS
+  int cnt;
+#endif
+#ifdef __EMX__
+    /* OS/2 dosen't have Users and Groups */
+    return 1;
+#else    
     if(user_id == finfo->st_uid)
         if(finfo->st_mode & S_IXUSR)
             return 1;
     if(group_id == finfo->st_gid)
         if(finfo->st_mode & S_IXGRP)
             return 1;
+#ifdef MULTIPLE_GROUPS
+    for(cnt=0; cnt < NGROUPS_MAX; cnt++) {
+        if(group_id_list[cnt] == finfo->st_gid)
+            if(finfo->st_mode & S_IXGRP)
+                return 1;
+    }
+#endif
     return (finfo->st_mode & S_IXOTH);
+#endif    
 }
 
 #ifdef NEED_STRDUP
-char *strdup (char *str)
+char *strdup (const char *str)
 {
   char *dup;
 
@@ -742,9 +1016,8 @@ int strncasecmp (const char *a, const char *b, int n)
 #ifdef NEED_INITGROUPS
 int initgroups(const char *name, gid_t basegid)
 {
-#ifdef QNX
-/* QNX does not appear to support supplementary groups.
-Ben <ben@algroup.co.uk> */
+#if defined(QNX) || defined(MPE)
+/* QNX and MPE do not appear to support supplementary groups. */
 	return 0;
 #else /* ndef QNX */
   gid_t groups[NGROUPS_MAX];
@@ -813,7 +1086,7 @@ void str_tolower(char *str) {
     }
 }
         
-uid_t uname2id(char *name) {
+uid_t uname2id(const char *name) {
     struct passwd *ent;
 
     if(name[0] == '#') 
@@ -823,10 +1096,10 @@ uid_t uname2id(char *name) {
         fprintf(stderr,"httpd: bad user name %s\n",name);
         exit(1);
     }
-    else return(ent->pw_uid);
+    return(ent->pw_uid);
 }
 
-gid_t gname2id(char *name) {
+gid_t gname2id(const char *name) {
     struct group *ent;
 
     if(name[0] == '#') 
@@ -836,7 +1109,7 @@ gid_t gname2id(char *name) {
         fprintf(stderr,"httpd: bad group name %s\n",name);
         exit(1);
     }
-    else return(ent->gr_gid);
+    return(ent->gr_gid);
 }
 
 #if 0
@@ -856,8 +1129,8 @@ struct in_addr get_local_addr(int sd) {
 
     len = sizeof(struct sockaddr);
     if(getsockname(sd,&addr,&len) < 0) {
-        fprintf (stderr, "Can't get local host address!\n");
 	perror ("getsockname");
+        fprintf (stderr, "Can't get local host address!\n");
 	exit(1);
     }
          
@@ -865,16 +1138,39 @@ struct in_addr get_local_addr(int sd) {
 }
 #endif
 
-unsigned long get_virthost_addr (char *w, int wild_allowed) {
+/*
+ * Parses a host of the form <address>[:port]
+ * :port is permitted if 'port' is not NULL
+ */
+unsigned long get_virthost_addr (const char *w, unsigned short *ports) {
     struct hostent *hep;
     unsigned long my_addr;
-    
-    if (wild_allowed && !strcmp(w, "*")) 
+    char *p;
+
+    p = strchr(w, ':');
+    if (ports != NULL)
+    {
+	*ports = 0;
+	if (p != NULL && strcmp(p+1, "*") != 0) *ports = atoi(p+1);
+    }
+
+    if (p != NULL) *p = '\0';
+    if (strcmp(w, "*") == 0)
+    {
+	if (p != NULL) *p = ':';
 	return htonl(INADDR_ANY);
+    }
 	
+#ifdef DGUX
+    my_addr = inet_network(w);
+#else
     my_addr = inet_addr(w);
-    if (my_addr != ((unsigned long) 0xffffffff))
+#endif
+    if (my_addr != INADDR_NONE)
+    {
+	if (p != NULL) *p = ':';
 	return my_addr;
+    }
 
     hep = gethostbyname(w);
 	    
@@ -890,101 +1186,11 @@ unsigned long get_virthost_addr (char *w, int wild_allowed) {
 	exit(1);
     }
 	    
+    if (p != NULL) *p = ':';
+
     return ((struct in_addr *)(hep->h_addr))->s_addr;
 }
 
-#ifdef DEBUG_ACCESS
-/* Host name and IP address used if getpeername() fails.
- * The most frequent cause of this is running the server in inetd-mode
- * with the tty as input and output, under the debugger.
- */
-#define NO_SUCH_HOST_NAME "volterra.ai.mit.edu"
-#define NO_SUCH_HOST_ADDR "128.52.39.173"
-#else
-#define NO_SUCH_HOST_NAME "UNKNOWN_HOST"
-#define NO_SUCH_HOST_ADDR "UNKNOWN_IP"
-#endif
-
-void get_remote_host(conn_rec *conn)
-{
-    struct sockaddr addr;
-    int len;
-    struct in_addr *iaddr;
-#if defined(MAXIMUM_DNS) || !defined(MINIMAL_DNS)
-    struct hostent *hptr;
-#endif
-
-    len = sizeof(struct sockaddr);
-
-    if ((getpeername(fileno(conn->client), &addr, &len)) < 0) {
-	conn->remote_name = conn->remote_host = NO_SUCH_HOST_NAME;
-	conn->remote_ip = NO_SUCH_HOST_ADDR;
-        return;
-    }
-
-    iaddr = &(((struct sockaddr_in *)&addr)->sin_addr);
-#ifndef MINIMAL_DNS
-    hptr = gethostbyaddr((char *)iaddr, sizeof(struct in_addr), AF_INET);
-    if(hptr) {
-        conn->remote_host = pstrdup(conn->pool, (void *)hptr->h_name);
-        conn->remote_name = conn->remote_host;
-        str_tolower (conn->remote_host);
-    }
-    else 
-#endif
-        conn->remote_host = NULL;
-
-#ifdef MAXIMUM_DNS
-    /* Grrr. Check THAT name to make sure it's really the name of the addr. */
-    /* Code from Harald Hanche-Olsen <hanche@imf.unit.no> */
-    if (conn->remote_host) {
-        char **haddr;
-
-        hptr = gethostbyname(conn->remote_host);
-        if (hptr) {
-            for(haddr=hptr->h_addr_list;*haddr;haddr++) {
-                if(((struct in_addr *)(*haddr))->s_addr == iaddr->s_addr)
-                    break;
-            }
-        }
-        if((!hptr) || (!(*haddr)))
-	    conn->remote_host = conn->remote_name = NULL;
-    }
-#endif
-    conn->remote_ip = pstrdup (conn->pool, inet_ntoa(*iaddr));
-    if(!conn->remote_host){ 	
-        conn->remote_name = conn->remote_ip;
-    }
-    if (!conn->remote_name) {
-	conn->remote_name = "UNKNOWN_HOST";
-    }
-}
-
-#ifdef NOTDEF    
-    
-char *get_remote_logname(FILE *fd) {
-    int len;
-    char *result;
-#if defined(NEXT) || defined(BSD4_4) || defined(SOLARIS2) || defined(LINUX)
-    struct sockaddr sa_server, sa_client;
-#else
-    struct sockaddr_in sa_server,sa_client;
-#endif
-
-    len = sizeof(sa_client);
-    if(getpeername(fileno(stdout),&sa_client,&len) != -1) {
-        len = sizeof(sa_server);
-        if(getsockname(fileno(stdout),&sa_server,&len) == -1)
-            result = "unknown";
-        else
-            result = rfc931((struct sockaddr_in *) & sa_client,
-                                    (struct sockaddr_in *) & sa_server);
-    }
-    else result = "unknown";
-
-    return result; /* robm=pinhead */
-}
-#endif    
 
 static char *find_fqdn(pool *a, struct hostent *p) {
     int x;
@@ -996,17 +1202,24 @@ static char *find_fqdn(pool *a, struct hostent *p) {
                 return pstrdup(a, p->h_aliases[x]);
         }
         return NULL;
-    } else return pstrdup(a, (void *)p->h_name);
+    }
+    return pstrdup(a, (void *)p->h_name);
 }
 
 char *get_local_host(pool *a)
 {
-    char str[128];
-    int len = 128;
+#ifndef MAXHOSTNAMELEN
+#define MAXHOSTNAMELEN 256
+#endif
+    char str[MAXHOSTNAMELEN+1];
     char *server_hostname;
-
     struct hostent *p;
-    gethostname(str, len);
+
+    if( gethostname( str, sizeof( str ) - 1 ) != 0 ) {
+	perror( "Unable to gethostname" );
+	exit(1);
+    }
+    str[MAXHOSTNAMELEN] = '\0';
     if((!(p=gethostbyname(str))) || (!(server_hostname = find_fqdn(a, p)))) {
         fprintf(stderr,"httpd: cannot determine local host name.\n");
 	fprintf(stderr,"Use ServerName to set it manually.\n");
@@ -1031,7 +1244,7 @@ const int pr2six[256]={
     64,64,64,64,64,64,64,64,64,64,64,64,64
 };
 
-char *uudecode(pool *p, char *bufcoded) {
+char *uudecode(pool *p, const char *bufcoded) {
     int nbytesdecoded;
     register unsigned char *bufin;
     register char *bufplain;
@@ -1076,6 +1289,32 @@ char *uudecode(pool *p, char *bufcoded) {
     return bufplain;
 }
 
+#ifdef __EMX__
+void os2pathname(char *path) {
+    char newpath[MAX_STRING_LEN];
+    int loop;
+    int offset;
+
+    offset = 0;
+    for (loop=0; loop < (strlen(path) + 1) && loop < sizeof(newpath)-1; loop++) {
+        if (path[loop] == '/') {
+            newpath[offset] = '\\';
+            /*
+            offset = offset + 1;
+            newpath[offset] = '\\';
+            */
+        } else
+            newpath[offset] = path[loop];
+        offset = offset + 1;
+    };
+    /* Debugging code */
+    /* fprintf(stderr, "%s \n", newpath); */
+
+    strcpy(path, newpath);
+};
+#endif
+
+
 #ifdef NEED_STRERROR
 char *
 strerror (int err) {
